@@ -9,6 +9,7 @@
 #include <X11/Xmu/CharSet.h>
 #include <X11/Xew/XewInit.h>
 #include <X11/Xew/BasicP.h>
+#include "PrintP.h"
 #include <errno.h>
 #include <ctype.h>
 #ifdef SH_MEM
@@ -28,9 +29,9 @@ extern int atoi( /* No header for this in SYSV? */
 #endif
 
 #include <fcntl.h>	/* Is this standard header? check! --msa */
+#include "Color.h"
 #include "BasicColor.h"
 
-
 /*
 ** X11R4 doesn't seem to have XmuCvtStringToCursor. Register this
 ** converter only for post X11R4 releases. The resource 'cursor'
@@ -64,6 +65,8 @@ static XtActionsRec actions[] =
 	    {"leave-window",	LeaveWindow},
     };
 
+static XeFrameDimension init_FrameDimension;
+
 #define offset(field) XtOffsetOf(XeBasicRec, basic.field)
 static XtResource resources[] =
     {
@@ -73,17 +76,20 @@ static XtResource resources[] =
 		 XtOffsetOf(XeBasicRec, core.border_width),
 		 XtRImmediate, (XtPointer)0},
 	{XtNcursor, XtCCursor, XtRCursor, sizeof(Cursor),
-		 offset(cursor), XtRImmediate, (XtPointer) None},
+		 offset(cursor), XtRImmediate, (XtPointer)None},
 	{XtNforeground, XtCForeground, XtRPixel, sizeof(Pixel),
 		 offset(foreground_pixel), XtRString, XtDefaultForeground},
 	{XtNcolormapUse, XtCColormapUse, XtRXeColormapUse,
-		 sizeof(XeColormapUse), offset(colormap_use),
+		 sizeof(XeColormapUse), offset(imaging.colormap_use),
 		 XtRImmediate, (XtPointer)XeColormapUse_SHARED},
 	{XtNcolormapInstall, XtCColormapInstall, XtRBoolean,
 		 sizeof(Boolean),offset(colormap_install),XtRImmediate,False},
 	{XtNcolorMode, XtCColorMode, XtRXeColorMode,
-		 sizeof(XeColorMode), offset(color_mode),
+		 sizeof(XeColorMode), offset(imaging.color_mode),
 		 XtRImmediate, (XtPointer)XeColorMode_COLOR},
+	{XtNcolorQuantize, XtCColorQuantize, XtRXeColorQuantize,
+		 sizeof(XeColorQuantize), offset(imaging.color_quantize),
+		 XtRImmediate, (XtPointer)XeColorQuantize_FAST},
 	{XtNcontentFile, XtCContentFile, XtRString, sizeof(String),
 		 offset(content_file), XtRImmediate, (XtPointer)NULL},
 	{XtNcontentFilter, XtCContentFilter, XtRString, sizeof(String),
@@ -101,14 +107,20 @@ static XtResource resources[] =
 		 offset(content_length), XtRImmediate, (long)0},
 	{XtNcontentLoaded, XtCContentLoaded, XtRBoolean, sizeof(Boolean),
 		 offset(content_loaded), XtRImmediate, False},
+        {XtNdither, XtCDither, XtRXeDither, sizeof(XeDither),
+		 offset(imaging.dither),
+		 XtRImmediate, (XtPointer)XeDither_FS4 },
 	{XtNexportCallback, XtCExportCallback, XtRCallback, sizeof(XtPointer),
 		 offset(export_callbacks), XtRCallback, (XtPointer)NULL},
 	{XtNexposeCallback, XtCExposeCallback, XtRCallback, sizeof(XtPointer),
 		 offset(expose_callbacks), XtRCallback, (XtPointer)NULL},
+	{XtNhorizontal, XtCFrameDimension, XeRFrameDimension,
+		 sizeof(XeFrameDimension), offset(dimension[FRAME_HORIZONTAL]),
+		 XeRFrameDimension, (XtPointer)&init_FrameDimension},
 	{XtNimportCallback, XtCImportCallback, XtRCallback, sizeof(XtPointer),
 		 offset(import_callbacks), XtRCallback, (XtPointer)NULL},
 	{XtNmaxColors, XtCMaxColors, XtRInt, sizeof(int),
-		 offset(max_colors), XtRImmediate, (XtPointer)65535},
+		 offset(imaging.max_colors), XtRImmediate, (XtPointer)65535},
 	{XtNmirrorImage, XtCMirrorImage, XtRBoolean, sizeof(Boolean),
 		 offset(mirror_image), XtRImmediate, False},
 	{XtNnotifyCallback, XtCNotifyCallback, XtRCallback, sizeof(XtPointer),
@@ -122,9 +134,20 @@ static XtResource resources[] =
 	{XtNscaling, XtCScaling, XtRXeScaling, sizeof(XeScaling),
 		 offset(scaling), XtRImmediate, (XtPointer)XeScaling_WIDTH},
 	{XtNuseShm, XtCUseShm, XtRBoolean, sizeof(Boolean),
-		 offset(use_shm), XtRImmediate, False},
+		 offset(imaging.use_shm), XtRImmediate, False},
+#ifndef USING_MOTIF_122
+	{XtNuserData, XtCUserData, XtRPointer, sizeof(XtPointer),
+		 offset(user_data), XtRImmediate, 0},
+#endif
+	{XtNvertical, XtCFrameDimension, XeRFrameDimension,
+		 sizeof(XeFrameDimension), offset(dimension[FRAME_VERTICAL]),
+		 XeRFrameDimension, (XtPointer)&init_FrameDimension},
 	{XtNvisual, XtCVisual, XtRVisual,
 		 sizeof(Visual *), offset(visual), XtRImmediate, 0},
+	{XtNxTranslation, XtCXTranslation, XtRInt, sizeof(int),
+		 offset(x_translation), XtRImmediate, 0},
+	{XtNyTranslation, XtCYTranslation, XtRInt, sizeof(int),
+		 offset(y_translation), XtRImmediate, 0},
     };
 #undef offset
 
@@ -132,6 +155,7 @@ static void Initialize(), Realize(), Destroy(), Redisplay();
 static void ClassPartInitialize(), ClassInitialize();
 static Boolean SetValues();
 static XtGeometryResult GeometryManager();
+static void PrintInitialize(), PrintDestroy();
 
 #ifdef USING_MOTIF_122
 static XmNavigability WidgetNavigable() ;
@@ -211,35 +235,40 @@ XeBasicClassRec xeBasicClassRec = {
 #endif
   },
   { /* composite class fields */
-    /* geometry_manager		*/   GeometryManager,
-    /* change_managed		*/   XtInheritChangeManaged,
-    /* insert_child		*/   XtInheritInsertChild,
-    /* delete_child		*/   XtInheritDeleteChild,
-    /* extension		*/   NULL
+    /* geometry_manager		*/	GeometryManager,
+    /* change_managed		*/	XtInheritChangeManaged,
+    /* insert_child		*/	XtInheritInsertChild,
+    /* delete_child		*/	XtInheritDeleteChild,
+    /* extension		*/	NULL
   },
   { /* constraint class fields */
-    /* subresources		*/   NULL,
-    /* subresource_count	*/   0,
-    /* constraint_size		*/   0,
-    /* initialize		*/   NULL,
-    /* destroy			*/   NULL,
-    /* set_values		*/   NULL,
-    /* extension		*/   NULL
+    /* subresources		*/	NULL,
+    /* subresource_count	*/	0,
+    /* constraint_size		*/	0,
+    /* initialize		*/	NULL,
+    /* destroy			*/	NULL,
+    /* set_values		*/	NULL,
+    /* extension		*/	NULL
    },
 #ifdef USING_MOTIF_122
-    {  /* manager class record */
-       /* translations */                 XtInheritTranslations,
-       /* syn_resources */                NULL,
-       /* num_syn_resources */            0,
-       /* syn_constraint_resources */     NULL,
-       /* num_syn_constraint_resources */ 0,
-       /* parent_process */               XmInheritParentProcess,
-       /* extension */                    NULL  
-    },
+   { /* manager class record	*/
+     /* translations		*/	XtInheritTranslations,
+     /* syn_resources		*/	NULL,
+     /* num_syn_resources	*/	0,
+     /* syn_constraint_resources */	NULL,
+     /* num_syn_constraint_resources */	0,
+     /* parent_process		*/	XmInheritParentProcess,
+     /* extension		*/	NULL  
+   },
 #endif
    { /* basic fields */
-    /* not used			*/	0
-  }
+     /* print_initialize	*/	PrintInitialize,
+     /* print			*/	_XeDefaultPrint,
+     /* print_cancel		*/	_XeDefaultPrintCancel,
+     /* print_finish		*/	PrintDestroy,
+     /* print_constraint_initialize */	NULL,
+     /* print_constraint_destroy */	NULL
+   }
 };
 
 WidgetClass xeBasicWidgetClass = (WidgetClass)&xeBasicClassRec;
@@ -385,26 +414,78 @@ XtPointer *converter_data;
 			break;
 		    }
 	    }
-/* For some reason, the CvtDone macro doesn't work!??? gcc problem? */
-#if 0
 	CvtDone(int, e);
-#else
-	if (toVal->addr != NULL)
-	    {
-		if (toVal->size < sizeof(XeScaling))
-		    {
-			toVal->size = sizeof(XeScaling);
-			return False;
-		    }
-		*(int*)(toVal->addr) = e;
-	    }
-	else
-		toVal->addr = (XPointer)&e;
-	toVal->size = sizeof(XeScaling);
-	return True;
-#endif
     }
 
+/*
+** XeCvtStringToPixel
+**	is a modified version of XtCvtStringPixel that uses Xew Color
+**	library instead of raw XAllocNamed color.
+*/
+static Boolean XeCvtStringToPixel
+	(dpy, args, num_args, fromVal, toVal, closure_ret)
+Display* dpy;
+XrmValuePtr args;
+Cardinal *num_args;
+XrmValuePtr fromVal;
+XrmValuePtr toVal;
+XtPointer *closure_ret;
+    {
+	String str = (String)fromVal->addr;
+	XColor color;
+	XeSample red, green, blue;
+	Screen *screen;
+	Colormap colormap;
+	Status status;
+	String params[1];
+	Cardinal num_params=1;
+
+	if (*num_args != 2)
+	    {
+		XtAppWarningMsg(XtDisplayToApplicationContext(dpy),
+				"wrongParameters","XeCvtStringToPixel",
+				"XewError",
+	"String to pixel conversion needs screen and colormap arguments",
+				(String *)NULL, (Cardinal *)NULL);
+		return False;
+	    }
+	screen = *((Screen **) args[0].addr);
+	colormap = *((Colormap *) args[1].addr);
+
+	if (XmuCompareISOLatin1(str, XtDefaultBackground) == 0)
+	    {
+		*closure_ret = False;
+		CvtDone(Pixel, WhitePixelOfScreen(screen));
+	    }
+	if (XmuCompareISOLatin1(str, XtDefaultForeground) == 0)
+	    {
+		*closure_ret = False;
+		CvtDone(Pixel, BlackPixelOfScreen(screen));
+	    }
+	status = XParseColor(DisplayOfScreen(screen), colormap,
+			     (char*)str, &color);
+	if (status == 0)
+	    {
+		params[0] = str;
+		XtAppWarningMsg(XtDisplayToApplicationContext(dpy),
+				"badColor", "cvtStringToPixel", "XewError",
+				"Cannot parse color \"%s\" definition",
+				params, &num_params);
+		*closure_ret = False;
+		return False;
+	    }
+	red = XToSample(color.red);
+	green = XToSample(color.green);
+	blue = XToSample(color.blue);
+	(void)XeAllocColors
+		(DisplayOfScreen(screen), RootWindowOfScreen(screen), colormap,
+		 XeColormapUse_SHAREDOWN, 1, 1,
+		 &red, &green, &blue, &color.pixel);
+	*closure_ret = (char*)True;
+	CvtDone(Pixel, color.pixel);
+    }
+
+#ifdef not_used_but_left_as_reference_for_future_need
 
 typedef enum
     {
@@ -500,7 +581,7 @@ int mms;
 		    }
 	return d;
     }
-
+#endif
 
 /*
 ** XeCvtStringToHorizontalValue
@@ -518,23 +599,7 @@ XtPointer *converter_data;
 	char *s = (char *) fromVal->addr;
 
 	d = atoi(s); /* Dummy Thing Here */
-#if 0
 	CvtDone(int, d);
-#else
-	if (toVal->addr != NULL)
-	    {
-		if (toVal->size < sizeof(int))
-		    {
-			toVal->size = sizeof(int);
-			return False;
-		    }
-		*(int*)(toVal->addr) = d;
-	    }
-	else
-		toVal->addr = (XPointer)&d;
-	toVal->size = sizeof(int);
-	return True;
-#endif
     }
 
 static void ClassInitialize()
@@ -564,6 +629,19 @@ static void ClassInitialize()
 		}
 	    };
 #endif
+	static XtConvertArgRec convertPixelArg[] =
+	    {
+		{
+		    XtWidgetBaseOffset,
+		    (XtPointer) XtOffsetOf(WidgetRec, core.screen),
+		    sizeof(Screen *),
+		},
+		{
+		    XtWidgetBaseOffset,
+		    (XtPointer) XtOffsetOf(WidgetRec, core.colormap),
+		    sizeof(Colormap),
+		}
+	    };
 	static XtConvertArgRec convertContentFormatArg[] =
 	    {
 		{
@@ -622,6 +700,14 @@ static void ClassInitialize()
 			   convertArg, XtNumber(convertArg), 
 			   XtCacheByDisplay, NULL);
 #endif
+	/* Override Xt String to Pixel with own converter!! */
+
+	XtSetTypeConverter(XtRString, XtRPixel,
+			   XeCvtStringToPixel,
+			   convertPixelArg,
+			   XtNumber(convertPixelArg),
+			   XtCacheNone,
+			   (XtDestructor)NULL);
 	XtSetTypeConverter(XtRString, XtRXeColormapUse,
 			   XeCvtStringToEnum,
 			   convertColormapUseArg,
@@ -668,23 +754,20 @@ Widget request, new;
     {
 	XeBasicWidget w = (XeBasicWidget)new;
 
-	w->basic.num_colors = 0;
-	w->basic.num_allocated = 0;
-	w->basic.alloc = NULL;
 	w->basic.cmap = 0;
 #ifdef SH_MEM
-	if (w->basic.use_shm && XShmQueryExtension(XtDisplay(w)))
+	if (w->basic.imaging.use_shm && XShmQueryExtension(XtDisplay(w)))
 	    {
 		int major, minor, stat;
 		Bool pixs;
 
 		stat = XShmQueryVersion(XtDisplay(w), &major, &minor, &pixs);
-		w->basic.use_shm = (stat != 0);
+		w->basic.imaging.use_shm = (stat != 0);
 		if (stat && pixs)
-			w->basic.use_shm = 2;
+			w->basic.imaging.use_shm = 2;
 	    }
 	else
-		w->basic.use_shm = False;
+		w->basic.imaging.use_shm = False;
 #endif
 	/*
 	** Make local copies of some resource values
@@ -839,6 +922,14 @@ Dimension *hScaled;	/* New scaled height of rectangle in pixels */
 static void ClassPartInitialize(class)
 WidgetClass class;
     {
+	XeBasicWidgetClass c = (XeBasicWidgetClass)class;
+	XeBasicWidgetClass super =
+		(XeBasicWidgetClass)c->core_class.superclass;
+
+	if (c->basic_class.print == XeInheritPrint)
+		c->basic_class.print = super->basic_class.print;
+	if (c->basic_class.print_cancel == XeInheritPrintCancel)
+		c->basic_class.print_cancel = super->basic_class.print_cancel;
     }
 
 /*
@@ -884,6 +975,67 @@ Region r;
 			   (XtPointer)&data);
     }
 
+/*
+**	****************
+**	Printing support
+**	****************
+*/
+static void PrintInitialize(w, c, pp)
+Widget w, c;
+XePrintProcess pp;
+    {
+	XeBasicWidget clone = (XeBasicWidget)c;
+	XeBasicWidget bw = (XeBasicWidget)w;
+	XeSample r, g, b;
+	/*
+	** Override the imaging attributes in the clone structure
+	*/
+	clone->basic.imaging = pp->printer->imaging;
+	/*
+	** Special with colormode: if the widget color_mode specifies
+	** less than printing color mode, use the widget one.
+	*/
+	if (bw->basic.imaging.color_mode != XeColorMode_NONE &&
+	    bw->basic.imaging.color_mode < clone->basic.imaging.color_mode)
+		clone->basic.imaging.color_mode = bw->basic.imaging.color_mode;
+	/*
+	** ..a bit dubious juggling with max_colors, but this makes it
+	** easier to use XePrinter structure.. most sensible defaults
+	** are produced by zeroing the structure.. -- msa
+	*/
+	if (clone->basic.imaging.max_colors <= 0)
+		clone->basic.imaging.max_colors =
+			((XeBasicWidget)w)->basic.imaging.max_colors;
+	clone->basic.colormap_install = False;
+
+	/*
+	** Transfer foreground_pixel. Upon entering PrintInitialize, the
+	** pixel still has the original value. Retrieve the colors using
+	** the original widget and allocate a new pixel matching the
+	** colors from the printer.
+	*/
+	if (pp->printer->background)
+		XeQueryColors(XtDisplay(w), w->core.colormap, 1,
+			      &clone->basic.foreground_pixel, &r, &g, &b);
+	else
+		r = g = b = 0; /* 'black' */
+
+	(void)XeAllocColors
+		(XtDisplay(c), XtWindow(c), c->core.colormap,
+		 XeColormapUse_SHARED, 1, 1, &r, &g, &b,
+		 &clone->basic.foreground_pixel);
+    }
+
+static void PrintDestroy(w, c, pp)
+Widget w, c;
+XePrintProcess pp;
+    {
+	XeBasicWidget clone = (XeBasicWidget)c;
+
+	XeFreeColors(XtDisplay(c), clone->core.colormap,
+		     1, &clone->basic.foreground_pixel);
+    }
+
 static Boolean SetValues(current, request, new, args, num_args)
 Widget current, request, new;
 ArgList args;
@@ -947,7 +1099,10 @@ Cardinal *num_args;
 	if (new_cursor && XtIsRealized(new))
 		XDefineCursor(XtDisplay(new), XtWindow(new),
 			      s_new->basic.cursor);
-	return s_old->basic.foreground_pixel != s_new->basic.foreground_pixel;
+	return
+		s_old->basic.foreground_pixel!=s_new->basic.foreground_pixel ||
+		s_old->basic.x_translation != s_new->basic.x_translation ||
+		s_old->basic.y_translation != s_new->basic.y_translation;
     }
 
 /*
@@ -995,16 +1150,11 @@ Widget w;
 
 	XtFree(rw->basic.content_file);
 	XtFree(rw->basic.content_filter);
-	
-	_XeFreeColors(rw);
     }
 
 /*
 ** Widget needs to follow cursor to change the colourmaps if private
 ** colourmaps are installed.
-**
-** A private colormap is indicated by non-zero value in basic.cmap
-** field (which is actually the saved old default colormap).
 */
 static void EnterWindow(w, e, s, n)
 Widget w;
@@ -1016,8 +1166,8 @@ Cardinal *n;
 	XeBasicWidget r = (XeBasicWidget)w;
 	Widget wlst[2];
 	
-	if (r->basic.cmap == 0)
-		return;	/* No private colormap, nothing to do */
+	if (!XtParent(r) || r->core.colormap == XtParent(r)->core.colormap)
+		return;	/* No private colormap, or same as parent */
 	if (r->basic.colormap_install)
 	    {
 		XInstallColormap(XtDisplay(w), r->core.colormap);
@@ -1038,8 +1188,8 @@ Cardinal *n;
     {
 	XeBasicWidget r = (XeBasicWidget)w;
 
-	if (r->basic.cmap == 0)
-		return;	/* No private colormap, nothing to do */
+	if (!XtParent(r) || r->core.colormap == XtParent(r)->core.colormap)
+		return; /* No private colormap, or same as parent */
 	if (r->basic.colormap_install)
 	    {
 		XUninstallColormap(XtDisplay(w), r->core.colormap);
@@ -1241,8 +1391,12 @@ XeDataFlow *df;
 		else if (obytes > 0)
 		    {
 			df->length -= obytes;
-			/* memmove(df->buffer, df->buffer + obytes, df->length); -- fix later, msa*/
+#if HAVE_MEMMOVE
+			memmove(df->buffer, df->buffer + obytes, df->length);
+#else
+			/* No memmove, take a risk with memcpy.. */
 			memcpy(df->buffer, df->buffer + obytes, df->length);
+#endif
 			return 1;
 		    }
 		else
