@@ -1,22 +1,23 @@
 #ifndef BasicColor_h
 #define BasicColor_h
 /*
-** Copyright 1992, 1993, 1994 by Markku Savela and
+** Copyright 1992-1995 by Markku Savela and
 **	Technical Research Centre of Finland
 */
 #include <X11/Xew/BasicP.h>
+#include "ImageTools.h"
 
 _XFUNCPROTOBEGIN
 void _XeFreeColors
 	(
 #if NeedFunctionPrototypes
-	 XeBasicWidget
+	 XeImage *
 #endif
 	 );
 Colormap _XeResetColormap
 	(
 #if NeedFunctionPrototypes
-	 XeBasicWidget,
+	 XeImage *,
 	 int
 #endif
 	 );
@@ -33,7 +34,7 @@ Colormap _XeResetColormap
 int _XeAllocColors
 	(
 #if NeedFunctionPrototypes
-	 XeBasicWidget,		/* Widget */
+	 XeImage *,		/* Imaging context */
 	 int,			/* Size of the input RGB color maps */
 	 unsigned char *,	/* Red map */
 	 unsigned char *,	/* Green map */
@@ -41,7 +42,8 @@ int _XeAllocColors
 	 unsigned char *,	/* Image data base address */
 	 int,			/* Image width */
 	 int,			/* Image height */
-	 int			/* Image scanline length (bytes) */
+	 int,			/* Image scanline length (bytes) */
+	 int			/* True = request sorting of colors */
 #endif
 	 );
 _XFUNCPROTOEND
