@@ -1,5 +1,5 @@
 /*
-** Copyright 1994, 1995 by Markku Savela and
+** Copyright 1994-1995 by Markku Savela and
 **	Technical Research Centre of Finland
 */
 
@@ -46,9 +46,6 @@ static XtResource resources[] =
 	{XtNbottomShadowContrast, XtCBottomShadowContrast, XtRInt, sizeof(int),
 		 offset(shadow_contrast[FRAME_BOTTOM]),
 		 XtRImmediate, (XtPointer)40},
-	{XtNhorizontal, XtCFrameDimension, XeRFrameDimension,
-		 sizeof(XeFrameDimension), offset(dimension[FRAME_HORIZONTAL]),
-		 XeRFrameDimension, (XtPointer)&init_FrameDimension},
 	{XtNinvertShadow, XtCInvertShadow, XtRBoolean, sizeof(Boolean),
 		 offset(invert_shadow), XtRImmediate, (XtPointer)False},
 	{XtNlayoutPath, XtCLayoutPath, XtRInt, sizeof(int),
@@ -70,9 +67,6 @@ static XtResource resources[] =
 	{XtNtrailingBorder, XtCFrameBorder, XeRFrameBorder,
 		 sizeof(XeFrameBorder), offset(border[FRAME_TRAILING]),
 		 XeRFrameBorder, (XtPointer)&init_FrameBorder},
-	{XtNvertical, XtCFrameDimension, XeRFrameDimension,
-		 sizeof(XeFrameDimension), offset(dimension[FRAME_VERTICAL]),
-		 XeRFrameDimension, (XtPointer)&init_FrameDimension},
     };
 
 #undef offset
@@ -145,6 +139,11 @@ static void ChangeManaged(), InsertChild(), DeleteChild();
 static Boolean SetValues(), FrameSetValues();
 static XtGeometryResult QueryGeometry();
 
+static void PrintInitialize();
+static void PrintCancel(), PrintDestroy();
+static XtPointer Print();
+static void PrintConstraintInitialize(), PrintConstraintDestroy();
+
 #ifdef USING_MOTIF_122
 static XmNavigability WidgetNavigable() ;
 
@@ -181,72 +180,77 @@ XmBaseClassExtRec _XeFramebaseClassExtRec = {
 
 XeFrameClassRec xeFrameClassRec = {
   { /* core class fields */
-    /* superclass		*/   (WidgetClass) SuperClass,
-    /* class name		*/   "XeFrame",
-    /* size			*/   sizeof(XeFrameRec),
-    /* class_initialize		*/   ClassInitialize,
-    /* class_part init		*/   NULL,
-    /* class_inited		*/   FALSE,
-    /* initialize		*/   Initialize,
-    /* initialize_hook		*/   NULL,
-    /* realize			*/   Realize,
-    /* actions			*/   actions,
-    /* num_actions		*/   XtNumber(actions),
-    /* resources		*/   resources,
-    /* resource_count		*/   XtNumber(resources),
-    /* xrm_class		*/   NULLQUARK,
-    /* compress_motion		*/   TRUE,
-    /* compress_exposure	*/   TRUE,
-    /* compress_enterleave	*/   TRUE,
-    /* visible_interest		*/   FALSE,
-    /* destroy			*/   Destroy,
-    /* resize			*/   Resize,
-    /* expose			*/   Redisplay,
-    /* set_values		*/   SetValues,
-    /* set_values_hook		*/   NULL,
-    /* set_values_almost	*/   XtInheritSetValuesAlmost,
-    /* get_values_hook		*/   NULL,
-    /* accept_focus		*/   NULL,
-    /* version			*/   XtVersion,
-    /* callback_private		*/   NULL,
-    /* tm_table			*/   XtInheritTranslations,
-    /* query_geometry		*/   QueryGeometry,
-    /* display_accelerator	*/   XtInheritDisplayAccelerator,
+    /* superclass		*/	(WidgetClass) SuperClass,
+    /* class name		*/	"XeFrame",
+    /* size			*/	sizeof(XeFrameRec),
+    /* class_initialize		*/	ClassInitialize,
+    /* class_part init		*/	NULL,
+    /* class_inited		*/	FALSE,
+    /* initialize		*/	Initialize,
+    /* initialize_hook		*/	NULL,
+    /* realize			*/	Realize,
+    /* actions			*/	actions,
+    /* num_actions		*/	XtNumber(actions),
+    /* resources		*/	resources,
+    /* resource_count		*/	XtNumber(resources),
+    /* xrm_class		*/	NULLQUARK,
+    /* compress_motion		*/	TRUE,
+    /* compress_exposure	*/	TRUE,
+    /* compress_enterleave	*/	TRUE,
+    /* visible_interest		*/	FALSE,
+    /* destroy			*/	Destroy,
+    /* resize			*/	Resize,
+    /* expose			*/	Redisplay,
+    /* set_values		*/	SetValues,
+    /* set_values_hook		*/	NULL,
+    /* set_values_almost	*/	XtInheritSetValuesAlmost,
+    /* get_values_hook		*/	NULL,
+    /* accept_focus		*/	NULL,
+    /* version			*/	XtVersion,
+    /* callback_private		*/	NULL,
+    /* tm_table			*/	XtInheritTranslations,
+    /* query_geometry		*/	QueryGeometry,
+    /* display_accelerator	*/	XtInheritDisplayAccelerator,
 #ifdef USING_MOTIF_122
-    /* extension		*/   (XtPointer)&_XeFramebaseClassExtRec,
+    /* extension		*/	(XtPointer)&_XeFramebaseClassExtRec,
 #else
-    /* extension		*/   NULL,
+    /* extension		*/	NULL,
 #endif
   },
   { /* composite class fields */
-    /* geometry_manager		*/   GeometryManager,
-    /* change_managed		*/   ChangeManaged,
-    /* insert_child		*/   InsertChild,
-    /* delete_child		*/   DeleteChild,
-    /* extension		*/   NULL
+    /* geometry_manager		*/	GeometryManager,
+    /* change_managed		*/	ChangeManaged,
+    /* insert_child		*/	InsertChild,
+    /* delete_child		*/	DeleteChild,
+    /* extension		*/	NULL
   },
   { /* constraint class fields */
-    /* subresources		*/   subresources,
-    /* subresource_count	*/   XtNumber(subresources),
-    /* constraint_size		*/   sizeof(XeFrameConstraintsRec),
-    /* initialize		*/   _XeFrameInitialize,
-    /* destroy			*/   _XeFrameDestroy,
-    /* set_values		*/   FrameSetValues,
-    /* extension		*/   NULL
-   },
+    /* subresources		*/	subresources,
+    /* subresource_count	*/	XtNumber(subresources),
+    /* constraint_size		*/	sizeof(XeFrameConstraintsRec),
+    /* initialize		*/	_XeFrameInitialize,
+    /* destroy			*/	_XeFrameDestroy,
+    /* set_values		*/	FrameSetValues,
+    /* extension		*/	NULL
+  },
 #ifdef USING_MOTIF_122
-    {  /* manager class record */
-       /* translations */                 XtInheritTranslations,
-       /* syn_resources */                NULL,
-       /* num_syn_resources */            0,
-       /* syn_constraint_resources */     NULL,
-       /* num_syn_constraint_resources */ 0,
-       /* parent_process */               XmInheritParentProcess,
-       /* extension */                    NULL  
-    },
+  { /* manager class record	*/
+    /* translations		*/	XtInheritTranslations,
+    /* syn_resources		*/	NULL,
+    /* num_syn_resources	*/	0,
+    /* syn_constraint_resources */	NULL,
+    /* num_syn_constraint_resources */	0,
+    /* parent_process		*/	XmInheritParentProcess,
+    /* extension		*/	NULL  
+  },
 #endif
-  { /* XeBasic fields */
-    /* not used			*/	0
+  { /* basic fields */
+    /* print_initialize		*/	PrintInitialize,
+    /* print			*/	Print,
+    /* print_cancel		*/	PrintCancel,
+    /* print_destroy		*/	PrintDestroy,
+    /* print_constraint_initialize */	PrintConstraintInitialize,
+    /* print_constraint_destroy */	PrintConstraintDestroy
   },
   { /* XeFrame fields */
     /* not used			*/	0
@@ -305,7 +309,7 @@ typedef struct XeFrameLayoutInfoRec
 	LayoutStack normal;	/* Children with normal layout positioning */
 	LayoutStack reverse;	/* Children with reverse layout positioning */
 	LayoutStack fixed;	/* Children with fixed layout positioning */
-	LayoutChild *maximize;	/* Maximize this child in layout direction */
+	int maximize;		/* "Maximize" children in layout direction */
 	int num_children;	/* Number of children in this structure */
 	LayoutChild list[1];	/* Computed layout information */
     } LayoutInfo;
@@ -371,15 +375,15 @@ int top, path, invert;
 			+ child->leading;
 		if (invert)
 		    {
-			child->position[path] = top;
+			child->position[path] = top + child->leading;
 			top += space;
 		    }
 		else
 		    {
 			top -= space;
-			child->position[path] = top;
+			child->position[path] = top + child->trailing;
 		    }
-		child->position[path] += child->trailing - child->b;
+		child->position[path] -= child->b;
 		child->position[1-path] =
 			Alignment(info, child, path, invert);
 		child->located = True;
@@ -432,18 +436,30 @@ int *w_out, *h_out;
     }
 
 #if NeedFunctionPrototypes
-static void MaximizeOneChild(LayoutInfo *, LayoutChild *, int, int);
+static void MaximizeChildren(LayoutInfo *, int, int, int);
 #endif
-static void MaximizeOneChild(info, child, spare, path)
+static void MaximizeChildren(info, num, spare, path)
 LayoutInfo *info;
-LayoutChild *child;
-int spare, path;
+int num, spare, path;
     {
-	int old;
+	int adjust = spare / num;
+	int extra = spare - adjust * num;
+	int i;
 
-	old = child->dimension[path];
-	child->dimension[path] += spare;
-	child->head->allocation += child->dimension[path] - old;
+	for (i = 0; i < info->num_children; i++)
+	    {
+		XeFrameConstraints fc = (XeFrameConstraints)
+			info->list[i].child->core.constraints;
+
+		if (fc->frame.dimension[path].value == XeFrameDimension_MAX)
+		    {
+			int twiddle = (extra > 0);
+
+			info->list[i].dimension[path] += adjust + twiddle;
+			info->list[i].head->allocation += adjust + twiddle;
+			extra -= twiddle;
+		    }
+	    }
     }
 
 #if NeedFunctionPrototypes
@@ -455,7 +471,7 @@ XeFrameConstraints fc;
 LayoutStack *stack;
 int path, reverse;
     {
-	int separation;
+	int separation, room;
 	int frame_size[2], padding[2];
 	int b_left, b_right, b_leading, b_trailing;
 	XeFrameDimension dimension[2];
@@ -549,17 +565,16 @@ int path, reverse;
 	padding[1-path] = stack->info->b_left	/* Frame left border */
 		+ stack->info->b_right	/* Frame right border */
 		+ child->left + child->right;	/* Child left/right offsets */
-	if (frame_size[1-path] > 0 && (dimension[1-path].value == 0 ||
-			  dimension[1-path].value >
-			  (frame_size[1-path] - padding[1-path])))
-		dimension[1-path].value = frame_size[1-path] - padding[1-path];
+	room = frame_size[1-path] - padding[1-path];
+	if (room > 0 &&
+	    (dimension[1-path].value == 0 || dimension[1-path].value > room))
+		dimension[1-path].value = room;
 	/*
-	** Allow one child to act like rubber in layout path direction. Mark
-	** this child for later re-evaluation.
+	** If child requests Max in layout path direction, bump the counter.
+	** This child will receive more space later...
 	*/
-	if (stack->info->maximize == NULL &&
-	    dimension[path].value == XeFrameDimension_MAX)
-		stack->info->maximize = child;
+	if (dimension[path].value == XeFrameDimension_MAX)
+		stack->info->maximize += 1;
 	GetGeometry(child, dimension[0].value, dimension[1].value,
 		    &dimension[0].value, &dimension[1].value);
 	/*
@@ -570,13 +585,13 @@ int path, reverse;
 	** alone can exceed the fixed dimension in layout path direction).
 	** --msa
 	*/
-	if (frame_size[1-path] > 0 && dimension[1-path].value >
-	    (frame_size[1-path] - padding[1-path]))
-		dimension[1-path].value = frame_size[1-path] - padding[1-path];
+	room = frame_size[1-path] - padding[1-path];
+	if (room > 0 && dimension[1-path].value > room)
+		dimension[1-path].value = room;
 	padding[path] = child->trailing + child->leading;
-	if (frame_size[path] > padding[path] && dimension[path].value >
-	    (frame_size[path] - padding[path]))
-		dimension[path].value = frame_size[path] - padding[path];
+	room = frame_size[path] - padding[path];
+	if (room > 0 && dimension[path].value > room)
+		dimension[path].value = room;
 	/*
 	** Further limit dimensions between (min,max) if dimension is not
 	** explicitly  specified, and if min,max have been specified.
@@ -661,8 +676,10 @@ FrameSize *size;
 		** flag a request for the original layout function to redo
 		** everything when it gets to it. (Just hope it doesn't go
 		** into loop... --msa)
+		** [ ...well, it seems to get into loop, redo setting
+		**   commented out for now... -- msa ]
 		*/
-		fw->frame.layout_info->redo = True;
+		/* fw->frame.layout_info->redo = True; */
 		return;
 	    }
 	/*
@@ -680,8 +697,8 @@ FrameSize *size;
 	    }
 	else
 	    {
-		dimension[0] = fw->frame.dimension[0];
-		dimension[1] = fw->frame.dimension[1];
+		dimension[0] = fw->basic.dimension[0];
+		dimension[1] = fw->basic.dimension[1];
 	    }
     redo_layout:
 	info = fw->frame.layout_info = (LayoutInfo *)XtCalloc
@@ -873,7 +890,7 @@ FrameSize *size;
 		    }
 	    }
 	if (info->maximize)
-		MaximizeOneChild(info, info->maximize,
+		MaximizeChildren(info, info->maximize,
 				 info->dimension[path].value
 				 - info->b_leading - glue - info->b_trailing
 				 - info->normal.allocation
@@ -994,14 +1011,66 @@ FrameSize *size;
 	if (mode)
 		for (i = 0; i < info->num_children; ++i)
 		    {
+			int width, height;
+			Widget cw;
+			
 			child = &info->list[i];
-			if (XtIsManaged(child->child))
+			cw = child->child;
+
+			if (!XtIsManaged(cw))
+				continue;
+			width = MAX(child->dimension[0], 1);
+			height = MAX(child->dimension[1], 1);
+			if (fw->frame.printing)
+			    {
+				/*
+				** If printing is set, assume we are really
+				** working with the cloned tree. As currently
+				** each cloned widget really has the same
+				** window, doing reconfigure to it is not
+				** the best thing... (especially, when
+				** there might not be any real window!!)
+				*/
+				cw->core.x = child->position[0];
+				cw->core.y = child->position[1];
+				cw->core.border_width = child->b;
+				if (cw->core.width != width ||
+				    cw->core.height != height)
+				    {
+					cw->core.width = width;
+					cw->core.height = height;
+					/*
+					** Call resize method directly if size
+					** changed (and if the widget is from
+					** Xew, cannot call unknown Resize
+					** safely for printing...)
+					*/
+#ifndef USING_MOTIF_122
+					/*
+					** MOTIF has some 'ResizeWrapper'
+					** around the resize method, and this
+					** seems require 'per display
+					** information', which 'printer
+					** display' does not have! Cannot
+					** do anything else, but skip the
+					** resize totally for MOTIF subclassed
+					** Xew...
+					*/
+					if (cw->core.widget_class->core_class.resize &&
+					    XtIsSubclass(cw, xeBasicWidgetClass))
+						(*cw->core.widget_class->
+						 core_class.resize)
+							((Widget)cw);
+#endif
+				    }
+			    }
+			else
 				XtConfigureWidget
-					(child->child,
+					(cw,
 					 child->position[0],
 					 child->position[1],
-					 MAX(child->dimension[0], 1),
-					 MAX(child->dimension[1], 1),
+					 width,
+					 height,
 					 child->b);
 		    }
 	/*
@@ -1114,6 +1183,7 @@ Widget request, init;
 		(init, &fw->frame.line,
 		 fw->frame.border, XtNumber(fw->frame.border),
 		 fw->frame.shadow_contrast);
+	fw->frame.printing = False;
     }
 
 static void Realize(w, valueMask, attributes)
@@ -1124,6 +1194,13 @@ XSetWindowAttributes *attributes;
 	XeFrameWidget fw = (XeFrameWidget) w;
 
 	FrameLayout(fw,(LayoutChild *)NULL,LAYOUT_AND_COMMIT,(FrameSize*)NULL);
+	/*
+	** Ad hoc safety kludge: enforce minimum size 1 for both dimensions!
+	*/
+	if (!fw->core.width)
+		fw->core.width = 1;
+	if (!fw->core.height)
+		fw->core.height = 1;
 	(*SuperClass->core_class.realize) (w, valueMask, attributes);
     }
 
@@ -1306,14 +1383,14 @@ XtWidgetGeometry *preferred;
 		size.width.min = size.width.max = 0;
 	    }
 	else
-		size.width = fw->frame.dimension[0];
+		size.width = fw->basic.dimension[0];
 	if (rq.request_mode & CWHeight)
 	    {
 		size.height.value = rq.height;
 		size.height.min = size.height.max = 0;
 	    }
 	else
-		size.height = fw->frame.dimension[1];
+		size.height = fw->basic.dimension[1];
 	size.border_width = (rq.request_mode & CWBorderWidth)
 		? rq.border_width : fw->core.border_width;
 	FrameLayout(fw, (LayoutChild *)NULL, LAYOUT_QUERY_ONLY, &size);
@@ -1330,6 +1407,84 @@ XtWidgetGeometry *preferred;
 	else
 		return XtGeometryAlmost;
     }
+/*
+**	****************
+**	Printing support
+**	****************
+**	As frames are often just large empty space, it makes some sense
+**	to have printing methods for them. The default print method with
+**	the heavy XGetImage/XPutImage processing can be avoided.
+*/
+static void PrintInitialize(w, c, pp)
+Widget w, c;
+XePrintProcess *pp;
+    {
+	XeFrameWidget fw = (XeFrameWidget)w;
+	XeFrameWidget clone = (XeFrameWidget)c;
+	FrameSize size;
+
+	Initialize(w, c);
+	clone->frame.printing = fw->frame.printing = True;
+	size.width.value = clone->core.width;
+	size.width.max = size.width.min = 0;
+	size.height.value = clone->core.height;
+	size.height.max = size.height.min = 0;
+	FrameLayout(clone, (LayoutChild *)NULL, LAYOUT_AND_COMMIT, &size);
+	if (!XtParent(clone))
+	    {
+		clone->core.width = size.width.value;
+		clone->core.height = size.height.value;
+	    }
+    }
+
+static XtPointer Print(w, c, x, y, pp)
+Widget w, c;
+int x, y;
+XePrintProcess pp;
+    {
+	XeFrameWidget clone = (XeFrameWidget)c;
+
+	clone->basic.x_translation = x;
+	clone->basic.y_translation = y;
+	Redisplay(c, (XEvent *)NULL, (Region)0);
+	return NULL;
+    }
+
+static void PrintCancel(w, clone, pp, client_data)
+Widget w, clone;
+XePrintProcess pp;
+XtPointer client_data;
+    {
+    }
+
+
+static void PrintDestroy(w, c, pp)
+Widget w, c;
+XePrintProcess pp;
+    {
+	XeFrameWidget fw = (XeFrameWidget)w;
+
+	fw->frame.printing = False;
+	Destroy(c);
+    }
+
+static void PrintConstraintInitialize(w, c, pp)
+Widget w, c;
+XePrintProcess pp;
+    {
+	_XeFrameInitialize(w, c, (ArgList)0, (Cardinal)0);
+    }
+
+static void PrintConstraintDestroy(w, c, pp)
+Widget w, c;
+XePrintProcess pp;
+    {
+	_XeFrameDestroy(c);
+    }
+
+/*
+**	**************
+*/
 
 static Boolean CheckBorder(old, set)
 XeFrameBorderP *old, *set;
@@ -1378,7 +1533,7 @@ Widget old, request, set;
 			 nw->frame.shadow_contrast);
 	changed |= cw->frame.invert_shadow != nw->frame.invert_shadow;
 	layout= CheckBorder(cw->frame.border, nw->frame.border) ||
-		CheckDimension(cw->frame.dimension, nw->frame.dimension) ||
+		CheckDimension(cw->basic.dimension, nw->basic.dimension) ||
 		cw->frame.layout_path != nw->frame.layout_path;
 	if (layout)
 		FrameLayout(nw,(LayoutChild *)NULL,LAYOUT_AND_COMMIT,
