@@ -30,7 +30,11 @@ extern int atoi( /* No header for this in SYSV */
 #include <X11/Xew/BasicP.h>
 #include <X11/Xew/FrameP.h>
 #include "FrameSupport.h"
+#include "Color.h"
 
+
+#undef MAX
+#undef MIN
 #define MAX(a,b) ((a) > (b) ? (a) : (b))
 #define MIN(a,b) ((a) < (b) ? (a) : (b))
 
@@ -344,9 +348,9 @@ XtPointer *converter_data;
 		    }
 		else
 		    {
-			border.color[0] = color.red >> 8;
-			border.color[1] = color.green >> 8;
-			border.color[2] = color.blue >> 8;
+			border.color[0] = XToSample(color.red);
+			border.color[1] = XToSample(color.green);
+			border.color[2] = XToSample(color.blue);
 			border.color[3] = 1;
 		    }
 	    }
@@ -356,10 +360,11 @@ XtPointer *converter_data;
 			color.pixel = bg;
 		else
 			color.pixel = border_pixel;
-		XQueryColors(DisplayOfScreen(screen), colormap, &color, 1);
-		border.color[0] = color.red >> 8;
-		border.color[1] = color.green >> 8;
-		border.color[2] = color.blue >> 8;
+		XeQueryColors(DisplayOfScreen(screen), colormap,
+			      1, &color.pixel,
+			      &border.color[0],
+			      &border.color[1],
+			      &border.color[2]);
 		border.color[3] = 1;
 	    }
 	CvtDone(XeFrameBorder, border);
@@ -515,7 +520,7 @@ XeFrameBorderP *b;
     {
 	if (b->pixel != FRAME_NO_PIXEL)
 	    {
-		XFreeColors(XtDisplay(w), w->core.colormap, &b->pixel,1,0L);
+		XeFreeColors(XtDisplay(w), w->core.colormap, 1, &b->pixel);
 		b->pixel = FRAME_NO_PIXEL;
 	    }
     }
@@ -581,6 +586,7 @@ XeFrameBorderP *border;
 int edge, *shadow_contrast;
     {
 	XColor color;
+	XeSample r, g, b;
 
 	color.red = border->value.color[0] * 257;
 	color.green = border->value.color[1] * 257;
@@ -599,15 +605,12 @@ int edge, *shadow_contrast;
 		border->pixel = FRAME_NO_PIXEL;
 		return;
 	    }
-	if (XAllocColor(XtDisplay(w), w->core.colormap, &color))
-		border->pixel = color.pixel;
-	else
-	    {
-		XeWidgetWarningMsg(w, "colorAllocation",
-				   "Failed allocating frame border color.",
-				   (String *)NULL, 0);
-		border->pixel = FRAME_NO_PIXEL;
-	    }
+	r = XToSample(color.red);
+	g = XToSample(color.green);
+	b = XToSample(color.blue);
+	(void)XeAllocColors(XtDisplay(w), XtWindow(w), w->core.colormap,
+			    XeColormapUse_SHARED,
+			    1, 1, &r, &g, &b, &border->pixel);
     }
 
 #if NeedFunctionPrototypes
@@ -642,8 +645,10 @@ int *old_c, *set_c;
 ** _XeFrameInitialize
 **	initialize XeFrameContraints record
 */
-void _XeFrameInitialize(request, init)
+void _XeFrameInitialize(request, init, args, num_args)
 Widget request, init;
+ArgList args;
+Cardinal *num_args;
     {
 	XeFrameConstraints fc = (XeFrameConstraints)init->core.constraints;
 	Widget w = XtParent(init);
@@ -686,8 +691,11 @@ Widget child;
 /*
 ** _XeDrawBorderRectangle
 **
-**	(This is unprecise.. should draw the lines as rectangles with
-**	stipple pattern to implement dots and dashes when needed --msa )
+**	Uses XDrawLines only when dash patterns are requested. XDrawline
+**	is unprecise... should draw the lines as rectangles with
+**	stipple pattern to implement dots and dashes when needed --msa
+**
+**	Could also do better joining of corners for differing edges...
 */
 void _XeDrawBorderRectangle
 	(widget,gc,line,x,y,w,h,left,right,top,bottom,invert)
@@ -698,7 +706,19 @@ int x, y, w, h;
 XeFrameBorderP *left, *right, *top, *bottom;
 int invert;	/* Invert top/bottom shadow, if non-zero */
     {
-	if (top->value.type == XeBorderType_SHADOW)
+	XeBorderType same_type;
+	Window window = XtWindow(widget);
+	Display *display = XtDisplay(widget);
+
+	if (top->value.type == right->value.type &&
+	    top->value.type == left->value.type &&
+	    top->value.type == bottom->value.type)
+		same_type = top->value.type;
+	else
+		/* 'INVISIBLE' to marks the case where border types differ */
+		same_type = XeBorderType_INVISIBLE;
+	
+	if (same_type == XeBorderType_SHADOW)
 	    {
 		XPoint pt[6];
 
@@ -715,8 +735,7 @@ int invert;	/* Invert top/bottom shadow, if non-zero */
 		pt[4].y = pt[3].y;
 		pt[5].x = pt[4].x;
 		pt[5].y = pt[0].y - bottom->value.width;
-		XFillPolygon(XtDisplay(widget), XtWindow(widget),
-			     gc, pt, XtNumber(pt),
+		XFillPolygon(display, window, gc, pt, XtNumber(pt),
 			     Complex, CoordModeOrigin);
 		SetupGC(widget, gc, line, invert ? top : bottom);
 		pt[0].x = x;
@@ -731,12 +750,28 @@ int invert;	/* Invert top/bottom shadow, if non-zero */
 		pt[4].y = pt[0].y - bottom->value.width;
 		pt[5].x = x + left->value.width;
 		pt[5].y = pt[4].y;
-		XFillPolygon(XtDisplay(widget), XtWindow(widget),
-			     gc, pt, XtNumber(pt),
+		XFillPolygon(display, window, gc, pt, XtNumber(pt),
 			     Complex, CoordModeOrigin);
 		return;
 	    }
 	/*
+	** Make a special case, if all edges have the same parameters
+	*/
+	if (same_type &&
+	    top->value.width == right->value.width &&
+	    top->pixel == right->pixel &&
+
+	    top->value.width == bottom->value.width &&
+	    top->pixel == bottom->pixel &&
+
+	    top->value.width == left->value.width &&
+	    top->pixel == left->pixel)
+	    {
+		SetupGC(widget, gc, line, top);
+		XDrawRectangle(display, window, gc, x, y, w, h);
+		return;
+	    }
+	/*
 	** Attempt to avoid bad joins at corners by drawing the lines
 	** clockwise, each corner has at least one starting line, if
 	** all edges have borders (sort of hack... does not work always )
@@ -744,36 +779,59 @@ int invert;	/* Invert top/bottom shadow, if non-zero */
 	if (top->value.type)
 	    {
 		SetupGC(widget, gc, line, top);
-		XDrawLine(XtDisplay(widget), XtWindow(widget), gc,
-			  x,	y + top->value.width / 2,
-			  x + w,y + top->value.width / 2);
+		if (top->value.type == XeBorderType_SOLID)
+			XFillRectangle(display, window, gc,
+				       x, y, w, top->value.width);
+		else
+			XDrawLine(display, window, gc,
+				  x,	y + top->value.width / 2,
+				  x + w,y + top->value.width / 2);
 	    }
 	if (right->value.type)
 	    {
 		SetupGC(widget, gc, line, right);
-		XDrawLine(XtDisplay(widget), XtWindow(widget), gc,
-			  x + w - (right->value.width+1)/2, y,
-			  x + w - (right->value.width+1)/2, y + h);
+		if (right->value.type == XeBorderType_SOLID)
+			XFillRectangle(display, window, gc,
+				       x + w - right->value.width, y,
+				       right->value.width, h);
+
+		else
+			XDrawLine(display, window, gc,
+				  x + w - (right->value.width+1)/2, y,
+				  x + w - (right->value.width+1)/2, y + h);
 	    }
 	if (bottom->value.type)
 	    {
 		SetupGC(widget, gc, line, bottom);
-		XDrawLine(XtDisplay(widget), XtWindow(widget), gc,
-			  x + w,y + h - (bottom->value.width + 1) / 2,
-			  x,	y + h - (bottom->value.width + 1) / 2);
+		if (bottom->value.type == XeBorderType_SOLID)
+			XFillRectangle(display, window, gc,
+				       x, y + h - bottom->value.width,
+				       w, bottom->value.width);
+		else
+			XDrawLine(display, window, gc,
+				  x + w,y + h - (bottom->value.width + 1) / 2,
+				  x,	y + h - (bottom->value.width + 1) / 2);
 	    }
 	if (left->value.type)
 	    {
 		SetupGC(widget, gc, line, left);
-		XDrawLine(XtDisplay(widget), XtWindow(widget), gc,
-			  x + left->value.width / 2, y + h,
-			  x + left->value.width / 2, y);
+		if (left->value.type == XeBorderType_SOLID)
+			XFillRectangle(display, window, gc,
+				       x, y, left->value.width, h);
+		else
+			XDrawLine(XtDisplay(widget), XtWindow(widget), gc,
+				  x + left->value.width / 2, y + h,
+				  x + left->value.width / 2, y);
 	    }
     }
 
 /*
 ** _XeDrawOwnBorder
 **	Draw Own Frame border inside the main window.
+**
+** *NOTE*
+**	This function implicitly assumes that the widget is Xew widget
+**	(e.g. the widget is a subclass of XeBasic!!!!)
 */
 void _XeDrawOwnBorder(widget, gc, line, left, right, top, bottom, invert)
 Widget widget;
@@ -786,15 +844,19 @@ int invert;
 
 	w = widget->core.width - left->value.space - right->value.space;
 	h = widget->core.height - top->value.space - bottom->value.space;
-	x = left->value.space;
-	y = top->value.space;
+	x = ((XeBasicWidget)widget)->basic.x_translation + left->value.space;
+	y = ((XeBasicWidget)widget)->basic.y_translation + top->value.space;
 	_XeDrawBorderRectangle
 		(widget,gc,line,x,y,w,h, left,right,top,bottom,invert);
     }
 
 /*
 ** _XeDrawChildBorder
-**	Draw Child Frame border around the child widget
+**	Draw Child Frame border around the child widget.
+**
+** *NOTE*
+**	This function implicitly assumes that the parent is Xew widget
+**	(e.g. parent is a subclass of XeBasic!!!!)
 */
 void _XeDrawChildBorder(widget, gc, line, child, left, right,top,bottom,invert)
 Widget widget;
@@ -811,8 +873,10 @@ int invert;
 		+ right->value.space + right->value.width;
 	h = bw2 + child->core.height + top->value.space + top->value.width
 		+ bottom->value.space + bottom->value.width;
-	x = child->core.x - left->value.space - left->value.width;
-	y = child->core.y - top->value.space - top->value.width;
+	x = ((XeBasicWidget)widget)->basic.x_translation + child->core.x
+		- left->value.space - left->value.width;
+	y = ((XeBasicWidget)widget)->basic.y_translation + child->core.y
+		- top->value.space - top->value.width;
 	_XeDrawBorderRectangle
 		(widget,gc,line,x,y,w,h, left,right,top,bottom,invert);
     }
@@ -1006,6 +1070,8 @@ static int LimitDimension(d, limit)
 int d;
 XeFrameDimension *limit;
     {
+	if (limit->value > 0)
+		return limit->value;
 	if (limit->min > 0 && d < limit->min)
 		d = limit->min;
 	if (limit->max > 0 && d > limit->max)
@@ -1019,8 +1085,8 @@ XeFrameDimension *limit;
 **	account the constraint resources.
 **
 ** *NOTE*
-**	The width and height parameters are only to be used when the
-**	corresponding constraint dimension requests 'max'.
+**	The input width and height parameters are only to be used when
+**	the corresponding constraint dimension requests 'max'.
 */
 void _XeFrameQueryExtents(w, width, height, e)
 Widget w;
@@ -1058,31 +1124,65 @@ XeFrameExtents *e;
 	     (value = width) > 0))
 	    {
 		intended.request_mode |= CWWidth;
-		intended.width = value;
+		intended.width = width = value;
 	    }
+	else
+		intended.width = width = 0;
 	if ((value = fc->frame.dimension[1].value) > 0 ||
 	    (fc->frame.dimension[1].value == XeFrameDimension_MAX &&
 	     (value = height) > 0))
 	    {
 		intended.request_mode |= CWHeight;
-		intended.height= value;
+		intended.height = height = value;
 	    }
+	else
+		intended.height = height = 0;
 	preferred = intended;
 	if (((CWWidth|CWHeight) & intended.request_mode) != (CWWidth|CWHeight))
+	    {
+		/*
+		** Width or height is still open, ask the widget for
+		** it's preferred size on that dimension
+		*/
 		XtQueryGeometry(w, &intended, &preferred);
-	preferred.width =
-		LimitDimension(preferred.width, &fc->frame.dimension[0]);
-	preferred.height =
-		LimitDimension(preferred.height, &fc->frame.dimension[1]);
+		intended.request_mode = CWWidth | CWHeight;
+		if (width == 0)
+			intended.width = LimitDimension
+				(preferred.width, &fc->frame.dimension[0]);
+		else
+			intended.width = width;
+		if (height == 0)
+			intended.height = LimitDimension
+				(preferred.height, &fc->frame.dimension[1]);
+		else
+			intended.height = height;
+		/*
+		** Re-query dimensions, if min/max constraints changed
+		** the preferred values.
+		*/
+		if ((!width && intended.width != preferred.width) ||
+		    (!height && intended.height != preferred.height))
+		    {
+			XtQueryGeometry(w, &intended, &preferred);
+			if (width == 0)
+				intended.width = LimitDimension
+					(preferred.width,
+					 &fc->frame.dimension[0]);
+			if (height == 0)
+				intended.height = LimitDimension
+					(preferred.height,
+					 &fc->frame.dimension[1]);
+		    }
+	    }
 	/*
 	** ...just an quick and dirty solution to the occasional "Widget
 	** has zero width and/or height" error...
 	*/
-	if (preferred.width == 0)
-		preferred.width = 1;
-	if (preferred.height == 0)
-		preferred.height = 1;
-	_XeFrameExtents(w, preferred.width, preferred.height, e);
+	if (intended.width == 0)
+		intended.width = 1;
+	if (intended.height == 0)
+		intended.height = 1;
+	_XeFrameExtents(w, intended.width, intended.height, e);
     }
 
 
