@@ -6,6 +6,10 @@
 **
 ** Some of the ideas and code snippets in this module have been taken
 ** from the Berkeley MPEG decoder.
+**
+** This module needs a serious redesign! The image processing "pipeline
+** functions" use the input tables in hard to understand way, not clean
+** general logic... (hard to add gamma so that it would apply generally)
 */
 #include <stdio.h>
 #if SYSV_INCLUDES
@@ -30,6 +34,7 @@ void free();
 #include <X11/StringDefs.h>
 #include <X11/Xew/VideoP.h>
 #include "BasicColor.h"
+#include "Color.h"
 #include "ImageTools.h"
 #include "Raster24to8.h"
 
@@ -311,7 +316,7 @@ while (0)
 
 /*
 ** PROPAGATE_FS4_ERROR
-**	compute Floyd-Sternberg error terms forward. This macro
+**	compute Floyd-Steinberg error terms forward. This macro
 **	assumes the following variables exist:
 **
 **	e	error value to propagate (new forward value left in)
@@ -377,8 +382,8 @@ int size;
 **	* This filter ignores the src->inc field!! *
 **	********************************************
 */
-static void ScaleDitherChannel1to1(widget, in, val, out, dst, src)
-XeBasicWidget widget;
+static void ScaleDitherChannel1to1(image, in, val, out, dst, src)
+XeImage *image;
 unsigned char *in, *val, *out;
 XeImageChannel *dst, *src;
     {
@@ -497,8 +502,8 @@ XeImageChannel *dst, *src;
 **	********************************************
 **
 */
-static void ScaleDitherChannel8to1(widget, in, val, out, dst, src)
-XeBasicWidget widget;
+static void ScaleDitherChannel8to1(image, in, val, out, dst, src)
+XeImage *image;
 unsigned char *in, *val, *out;
 XeImageChannel *dst, *src;
     {
@@ -643,8 +648,8 @@ XeImageChannel *dst, *src;
 	    }
     }
 
-static void ScaleDitherChannel8to8(widget, in, val, out, dst, src)
-XeBasicWidget widget;
+static void ScaleDitherChannel8to8(image, in, val, out, dst, src)
+XeImage *image;
 unsigned char *in, *val, *out;
 XeImageChannel *dst, *src;
     {
@@ -759,8 +764,8 @@ XeImageChannel *dst, *src;
     }
 
 
-static void ScaleMapChannel8to8(widget, in, val, out, dst, src)
-XeBasicWidget widget;
+static void ScaleMapChannel8to8(image, in, val, out, dst, src)
+XeImage *image;
 unsigned char *in, *val, *out;
 XeImageChannel *dst, *src;
     {
@@ -858,8 +863,8 @@ XeImageChannel *dst, *src;
 	    }
     }
 
-static void ScaleMapChannel8to16(widget, in, val, out8, dst, src)
-XeBasicWidget widget;
+static void ScaleMapChannel8to16(image, in, val, out8, dst, src)
+XeImage *image;
 unsigned char *in, *val, *out8;
 XeImageChannel *dst, *src;
     {
@@ -958,8 +963,8 @@ XeImageChannel *dst, *src;
 	    }
     }
 
-static void ScaleDitherChannel8to16(widget, in, val, out8, dst, src)
-XeBasicWidget widget;
+static void ScaleDitherChannel8to16(image, in, val, out8, dst, src)
+XeImage *image;
 unsigned char *in, *val, *out8;
 XeImageChannel *dst, *src;
     {
@@ -1075,8 +1080,8 @@ XeImageChannel *dst, *src;
     }
 
 
-static void ScaleMapChannel8to32(widget, in, val, out8, dst, src)
-XeBasicWidget widget;
+static void ScaleMapChannel8to32(image, in, val, out8, dst, src)
+XeImage *image;
 unsigned char *in, *val, *out8;
 XeImageChannel *dst, *src;
     {
@@ -1179,8 +1184,8 @@ XeImageChannel *dst, *src;
 ** ScaleDropPixels8to8
 **	Copy and scale from 8 bit to 8 bit by simple pixel dropping.
 */
-static void ScaleDropPixels8to8(widget, in, val, out, dst, src)
-XeBasicWidget widget;
+static void ScaleDropPixels8to8(image, in, val, out, dst, src)
+XeImage *image;
 unsigned char *in, *val, *out;
 XeImageChannel *dst, *src;
     {
@@ -1205,7 +1210,7 @@ XeImageChannel *dst, *src;
 
 			for (dCol = 0; dCol < dstW; dCol++)
 			    {
-				*dst_ptr += out[in[*src_ptr]];
+				*dst_ptr = out[in[*src_ptr]];
 				dst_ptr += dstinc;
 				src_ptr += srcinc;
 			    }
@@ -1227,7 +1232,7 @@ XeImageChannel *dst, *src;
 			    {
 				for ( ; dCol < dstW; dCol++)
 				    {
-					*dst_ptr += out[in[*src_ptr]];
+					*dst_ptr = out[in[*src_ptr]];
 					dst_ptr += dstinc;
 					src_ptr += srcinc;
 				    }
@@ -1244,7 +1249,7 @@ XeImageChannel *dst, *src;
 					while ((Psum += dstW) < srcW);
 					do
 					    {
-						*dst_ptr += v;
+						*dst_ptr = v;
 						dst_ptr += dstinc;
 						if (++dCol == dstW)
 							goto end_row;
@@ -1268,8 +1273,8 @@ XeImageChannel *dst, *src;
 **	THIS WORKS ONLY FOR "UNTOUCHED" CHANNEL DESCRIPTORS
 **	(NO MIRROR/ROTATION JUGGLING DONE!!!)
 */
-static void ZeroChannel(widget, in, val, out, dst, src)
-XeBasicWidget widget;
+static void ZeroChannel(image, in, val, out, dst, src)
+XeImage *image;
 unsigned char *in, *val, *out;
 XeImageChannel *dst, *src;
     {
@@ -1283,13 +1288,13 @@ XeImageChannel *dst, *src;
 **
 **	The r, g and b tables must be at least 256 bytes long!
 */
-static void AllocateColors8(widget, r, g, b, dst, src)
-XeBasicWidget widget;
+static void AllocateColors8(image, r, g, b, dst, src)
+XeImage *image;
 unsigned char *r, *g, *b;
 XeImageChannel *dst, *src;
     {
-	(void)_XeAllocColors(widget, 256, r, g, b,
-			     src->addr, src->w, src->h, src->line);
+	(void)_XeAllocColors(image, 256, r, g, b,
+			     src->addr, src->w, src->h, src->line, True);
     }
 
 /*
@@ -1299,12 +1304,12 @@ XeImageChannel *dst, *src;
 **
 **	The r, g and b tables must be at least 256 bytes long!
 */
-static void ColorMapping24to8H(widget, r, g, b, dst, src)
-XeBasicWidget widget;
+static void ColorMapping24to8H(image, r, g, b, dst, src)
+XeImage *image;
 unsigned char *r, *g, *b;
 XeImageChannel *dst, *src;
     {
-	int ncolors = widget->basic.max_colors;
+	int ncolors = image->imaging.max_colors;
 	
 	if (ncolors > 256)
 		ncolors = 256;
@@ -1312,8 +1317,8 @@ XeImageChannel *dst, *src;
 			    ncolors,		/* Max colors to use */
 			    r, g, b,		/* Colormaps to be filled */
 			    XeColorQuantize_HECKBERT);
-	(void)_XeAllocColors(widget, ncolors, r, g, b,
-			     dst->addr, dst->w, dst->h, dst->line);
+	(void)_XeAllocColors(image, ncolors, r, g, b,
+			     dst->addr, dst->w, dst->h, dst->line, True);
     }
 
 /*
@@ -1323,12 +1328,12 @@ XeImageChannel *dst, *src;
 **
 **	The r, g and b tables must be at least 256 bytes long!
 */
-static void ColorMapping24to8P(widget, r, g, b, dst, src)
-XeBasicWidget widget;
+static void ColorMapping24to8P(image, r, g, b, dst, src)
+XeImage *image;
 unsigned char *r, *g, *b;
 XeImageChannel *dst, *src;
     {
-	int ncolors = widget->basic.max_colors;
+	int ncolors = image->imaging.max_colors;
 	
 	if (ncolors > 256)
 		ncolors = 256;
@@ -1336,8 +1341,8 @@ XeImageChannel *dst, *src;
 			    ncolors,		/* Max colors to use */
 			    r, g, b,		/* Colormaps to be filled */
 			    XeColorQuantize_PPMQUANT);
-	(void)_XeAllocColors(widget, ncolors, r, g, b,
-			     dst->addr, dst->w, dst->h, dst->line);
+	(void)_XeAllocColors(image, ncolors, r, g, b,
+			     dst->addr, dst->w, dst->h, dst->line, True);
     }
 
 /*
@@ -1426,8 +1431,8 @@ int dstW, dstH;
 	    }
     }
 
-static void ScaleConvertRGBto8(widget, in, val, out, dst, src)
-XeBasicWidget widget;
+static void ScaleConvertRGBto8(image, in, val, out, dst, src)
+XeImage *image;
 unsigned char *in, *val, *out;
 XeImageChannel *dst, *src;
     {
@@ -1476,8 +1481,8 @@ XeImageChannel *dst, *src;
 	    }
     }
 
-static void ScaleCopyRGBto24(widget, in, val, out, dst, src)
-XeBasicWidget widget;
+static void ScaleCopyRGBto24(image, in, val, out, dst, src)
+XeImage *image;
 unsigned char *in, *val, *out;
 XeImageChannel *dst, *src;
     {
@@ -1579,8 +1584,8 @@ static void ycc_rgb_init()
 **	Convert from YCbCr to RGB space. The destination is assumed
 **	to have 24 bytes per pixel, and in format RGBRGBRGB...
 */
-static void ScaleConvertYCbCrTo24(widget, in, val, out, dst, src)
-XeBasicWidget widget;
+static void ScaleConvertYCbCrTo24(image, in, val, out, dst, src)
+XeImage *image;
 unsigned char *in, *val, *out;
 XeImageChannel *dst, *src;
     {
@@ -1679,8 +1684,8 @@ while (0)
 #undef END_FS4_ERROR
 #define END_FS4_ERROR(i) *err_ptr = e2[i]
 
-static void ScaleDitherXYZto8(widget, in3, val3, out, dst, src)
-XeBasicWidget widget;
+static void ScaleDitherXYZto8(image, in3, val3, out, dst, src)
+XeImage *image;
 unsigned char *in3, *val3, *out;
 XeImageChannel *dst, *src;
     {
@@ -1732,8 +1737,8 @@ XeImageChannel *dst, *src;
 				CLAMP(e[1]);
 				e[2] += w[2] + err_ptr[5];
 				CLAMP(e[2]);
-				i = out[in[0][e[0]]+in[1][e[1]]+in[2][e[2]]];
-				*dst_ptr = i;
+				i = in[0][e[0]]+in[1][e[1]]+in[2][e[2]];
+				*dst_ptr = out[i];
 				dst_ptr += dstinc;
 				e[0] -= val[0][i];
 				e[1] -= val[1][i];
@@ -1763,8 +1768,8 @@ XeImageChannel *dst, *src;
 **	A complex tranformation of 3 input channels to single
 **	output channel in one phase.
 */
-static void ScaleMapXYZto8(widget, in3, val3, out, dst, src)
-XeBasicWidget widget;
+static void ScaleMapXYZto8(image, in3, val3, out, dst, src)
+XeImage *image;
 unsigned char *in3, *val3, *out;
 XeImageChannel *dst, *src;
     {
@@ -1821,8 +1826,76 @@ XeImageChannel *dst, *src;
     }
 
 
-static void ScaleDitherYCCto8(widget, in3, val3, out, dst, src)
-XeBasicWidget widget;
+/*
+** ScaleMapXYZto16
+**	A complex tranformation of 3 input channels to single
+**	output channel in one phase.
+*/
+static void ScaleMapXYZto16(image, in3, val3, out8, dst, src)
+XeImage *image;
+unsigned char *in3, *val3, *out8;
+XeImageChannel *dst, *src;
+    {
+	int dstW = dst->w;
+	int dstH = dst->h;
+	int dstinc = dst->inc;
+	int srcH[3];
+	unsigned char *srcA[3];
+	unsigned char *dstA = dst->addr;
+	unsigned char *in[3], *val[3];
+	INT16 *out[3];
+	int dRow, Qsum[3], *pW, *wEnd, i;
+
+	pW = GetWorkSpace(dstW * 3 * sizeof(int));
+	wEnd = &pW[3*dstW];
+
+	for (i = 0; i < 3; ++i)
+	    {
+		srcH[i] = src[i].h;
+		srcA[i] = src[i].addr;
+		Qsum[i] = 0;
+		in[i] = in3;
+		val[i] = in3 + 256;
+		out[i] = (INT16 *)(in3 + 512); /* temp hack!! */
+		in3 += 4 * 256;
+	    }
+	for (dRow = 0; dRow < dstH; )
+	    {
+		CollapsRows(Qsum, pW, srcA, src, dstW, dstH);
+		/*
+		** Copy collected data to the destination
+		*/
+		do
+		    {
+			unsigned char *dst_ptr = dstA;
+			int *w;
+
+			for (w = pW; w < wEnd; w += 3)
+			    {
+				INT16 pixel;
+
+				pixel = out[0][in[0][w[0]]] +
+					out[1][in[1][w[1]]] +
+					out[2][in[2][w[2]]];
+				*((INT16 *)dst_ptr) = pixel;
+				dst_ptr += dstinc;
+			    }
+			dstA += dst->line;
+			dRow += 1;
+			Qsum[0] -= srcH[0];
+			Qsum[1] -= srcH[1];
+			Qsum[2] -= srcH[2];
+		    }
+		while (Qsum[0] >= srcH[0] &&
+		       Qsum[1] >= srcH[1] &&
+		       Qsum[2] >= srcH[2] &&
+		       dRow < dstH);
+	    }
+    }
+
+
+static void ScaleDitherYCCto8(image, in3, val3, out, dst, src)
+XeImage *image;
 unsigned char *in3, *val3, *out;
 XeImageChannel *dst, *src;
     {
@@ -1890,8 +1963,8 @@ XeImageChannel *dst, *src;
 				CLAMP(e[0]);
 				CLAMP(e[1]);
 				CLAMP(e[2]);
-				i = out[in[0][e[0]]+in[1][e[1]]+in[2][e[2]]];
-				*dst_ptr = i;
+				i = in[0][e[0]]+in[1][e[1]]+in[2][e[2]];
+				*dst_ptr = out[i];
 				dst_ptr += dstinc;
 				e[0] -= val[0][i];
 				e[1] -= val[1][i];
@@ -1921,8 +1994,8 @@ XeImageChannel *dst, *src;
 **	A complex tranformation of 3 input channels to single
 **	output channel in one phase.
 */
-static void ScaleMapYCCto8(widget, in3, val3, out, dst, src)
-XeBasicWidget widget;
+static void ScaleMapYCCto8(image, in3, val3, out, dst, src)
+XeImage *image;
 unsigned char *in3, *val3, *out;
 XeImageChannel *dst, *src;
     {
@@ -1997,6 +2070,90 @@ XeImageChannel *dst, *src;
 	    }
     }
 
+/*
+** ScaleMapYCCto16
+**	A complex tranformation of 3 input channels to single
+**	output channel in one phase.
+*/
+static void ScaleMapYCCto16(image, in3, val3, out8, dst, src)
+XeImage *image;
+unsigned char *in3, *val3, *out8;
+XeImageChannel *dst, *src;
+    {
+	int dstW = dst->w;
+	int dstH = dst->h;
+	int dstinc = dst->inc;
+	int srcH[3];
+	unsigned char *srcA[3];
+	unsigned char *dstA = dst->addr;
+	unsigned char *in[3], *val[3];
+	INT16 *out[3];
+	int dRow, Qsum[3], *pW, *wEnd, i;
+
+	static int init;
+
+	if (!init)
+	    {
+		init = 1;
+		ycc_rgb_init();
+	    }
+	pW = GetWorkSpace(dstW * 3 * sizeof(int));
+	wEnd = &pW[3*dstW];
+	for (i = 0; i < 3; ++i)
+	    {
+		srcH[i] = src[i].h;
+		srcA[i] = src[i].addr;
+		Qsum[i] = 0;
+		in[i] = in3;
+		val[i] = in3 + 256;
+		out[i] = (INT16 *)(in3 + 512);
+		in3 += 4 * 256;
+	    }
+	for (dRow = 0; dRow < dstH; )
+	    {
+		CollapsRows(Qsum, pW, srcA, src, dstW, dstH);
+		/*
+		** Copy collected data to the destination
+		*/
+		do
+		    {
+			unsigned char *dst_ptr = dstA;
+			int *w;
+			
+			for (w = pW; w < wEnd; w += 3)
+			    {
+				int r, g, b;
+				int Y = w[0];
+				int Cb = w[1];
+				int Cr = w[2];
+				INT16 pixel;
+
+				r = Y + Cr_r_tab[Cr];
+				g = Y + (int)RIGHT_SHIFT
+					(Cb_g_tab[Cb] + Cr_g_tab[Cr],
+					 SCALEBITS);
+				b = Y + Cb_b_tab[Cb];
+				CLAMP(r);
+				CLAMP(g);
+				CLAMP(b);
+				pixel = out[0][in[0][r]] +
+					out[1][in[1][g]] +
+					out[2][in[2][b]];
+				*((INT16 *)dst_ptr) = pixel;
+				dst_ptr += dstinc;
+			    }
+			dstA += dst->line;
+			dRow += 1;
+			Qsum[0] -= srcH[0];
+			Qsum[1] -= srcH[1];
+			Qsum[2] -= srcH[2];
+		    }
+		while (Qsum[0] >= srcH[0] &&
+		       Qsum[1] >= srcH[1] &&
+		       Qsum[2] >= srcH[2] &&
+		       dRow < dstH);
+	    }
+    }
 
 /*
 ** MakeImageChannel8
@@ -2004,8 +2161,8 @@ XeImageChannel *dst, *src;
 **	using the width and height of the source channel. The channel is
 **	initialized to NULL content.
 */
-static void MakeImageChannel8(widget, in, val, out, dst, src)
-XeBasicWidget widget;
+static void MakeImageChannel8(image, in, val, out, dst, src)
+XeImage *image;
 unsigned char *in, *val, *out;
 XeImageChannel *dst, *src;
     {
@@ -2022,8 +2179,8 @@ XeImageChannel *dst, *src;
 **	using the width and height of the source channel. The channel is
 **	initialized to NULL content.
 */
-static void MakeImageChannel24(widget, in, val, out, dst, src)
-XeBasicWidget widget;
+static void MakeImageChannel24(image, in, val, out, dst, src)
+XeImage *image;
 unsigned char *in, *val, *out;
 XeImageChannel *dst, *src;
     {
@@ -2040,8 +2197,8 @@ XeImageChannel *dst, *src;
 **
 **	DO NOT CALL THIS FUNCTION FOR "PERMANENT" CHANNELS!
 */
-static void FreeImageChannel(widget, in, val, out, dst, src)
-XeBasicWidget widget;
+static void FreeImageChannel(image, in, val, out, dst, src)
+XeImage *image;
 unsigned char *in, *val, *out;
 XeImageChannel *dst, *src;
     {
@@ -2055,11 +2212,9 @@ XeImageChannel *dst, *src;
 **
 **	Only RGB and YCbCr colorspaces now!
 */
-static void InitDisplay_COLORCUBE(w, im, src, dither, quant)
-XeBasicWidget w;
+static void InitDisplay_COLORCUBE(im, src)
 XeImage *im;
 XeRawImage *src;
-int dither, quant;
     {
 	/*
 	** Build the mapping tables along the following principles:
@@ -2118,12 +2273,12 @@ int dither, quant;
 		{ScaleMapYCCto8,	0,	3*256,	6*256,	0,1,	},
 		{NULL,			0,	0,	0,	0,0,	},
 	    };
-	char buf[200];
-	int ncolors, got;
+
+	int ncolors;
 	int i, j, f, k, a;
 	unsigned char *in, *val, *out, *Y[3], *r, *g, *b;
 	int C, R[3];
-	int N = w->basic.max_colors > 256 ? 256 : w->basic.max_colors;
+	int N = im->imaging.max_colors > 256 ? 256 : im->imaging.max_colors;
 	int c[3];
 #if 0
 #define ADJ(i,R) (XeSample_RANGE/(2*(R))) /* Used by Berkeley MPEG decoder */
@@ -2157,16 +2312,19 @@ int dither, quant;
 			a = ADJ(k,R[j]);
 			Y[j][i] = (k * XeSample_RANGE) / R[j] + a;
 		    }
+	im->num_pixels = ncolors;
+	im->pixels = (Pixel *)XtMalloc(ncolors * sizeof(Pixel *));
+	im->colormap = XeAllocColors /* FIX! --msa */
+		(im->display, DefaultRootWindow(im->display),
+		 im->colormap,
+		 im->imaging.colormap_use, im->imaging.max_colors, ncolors,
+		 r, g, b, im->pixels);
+	/* This is a temp solution, should really use 'Pixel' as output
+	   array type (not 8 bit index!!) */
+
 	for (i = 0; i < ncolors; i++)
-		out[i] = i;
-	got = _XeAllocColors(w,ncolors, r,g,b, out,ncolors,1,ncolors);
-	if (got < ncolors)
-	    {
-		sprintf(buf,"Warning, got %d exact colors, out of %d requested.",
-			got, ncolors);
-		XeWidgetWarningMsg((Widget)w,"noteImageToolsColorFail",
-				   buf, (String *)NULL, 0);
-	    }
+		out[i] = im->pixels[i];
+
 	for (i = 0; i < XeSample_RANGE; i++)
 		for (f = 1, j = 3; --j >= 0; f *= R[j])
 		    {
@@ -2175,9 +2333,11 @@ int dither, quant;
 			in[256*j+i] = k * f;
 		    }
 	if (src->color_space == XeImageSpace_YCbCr)
-		im->image_op = (dither == XeDither_FS4)?dither_ycc_8:map_ycc_8;
+		im->image_op = (im->imaging.dither == XeDither_FS4)
+			? dither_ycc_8 : map_ycc_8;
 	else
-		im->image_op = (dither == XeDither_FS4)?dither_8:map_8;
+		im->image_op = (im->imaging.dither == XeDither_FS4)
+			? dither_8 : map_8;
     }
 
 /*
@@ -2187,11 +2347,9 @@ int dither, quant;
 **
 **	Only RGB and YCbCr colorspaces now!
 */
-static void InitDisplay_COLORQUANT(w, im, src, dither, quant)
-XeBasicWidget w;
+static void InitDisplay_COLORQUANT(im, src)
 XeImage *im;
 XeRawImage *src;
-int dither, quant;
     {
 	/*
 	** Build the mapping tables along the following principles:
@@ -2244,7 +2402,7 @@ int dither, quant;
 	for (i = 0; i < 256; i++)
 		map[i] = i;	/* Indentity mapping for now.. */
 
-	if (quant == XeColorQuantize_HECKBERT)
+	if (im->imaging.color_quantize == XeColorQuantize_HECKBERT)
 		if (src->color_space == XeImageSpace_YCbCr)
 			im->image_op = ycbcr_8H;
 		else
@@ -2263,11 +2421,9 @@ int dither, quant;
 **	in grayscale. Allocate a range of gray shades and initialize
 **	mapping tables.
 */
-static void InitDisplay_PSEUDOCOLOR_GRAY(w, im, raw, dither, quant)
-XeBasicWidget w;
+static void InitDisplay_PSEUDOCOLOR_GRAY(im, raw)
 XeImage *im;
 XeRawImage *raw;
-int dither, quant;
     {
 	static XeImageOp dither_1[] = 
 	    {
@@ -2310,7 +2466,7 @@ int dither, quant;
 		{FreeImageChannel,	0,	0,	0,	-1,0,},
 		{NULL,			0,	0,	0,	0,0,},
 	    };
-	int ncolors = w->basic.max_colors > 256 ? 256 : w->basic.max_colors;
+	int ncolors = im->imaging.max_colors>256 ? 256:im->imaging.max_colors;
 	int i;
 	unsigned char r[256], g[256], b[256], *map, *ir, *ig, *ib;
 	int mapsize;
@@ -2330,7 +2486,7 @@ int dither, quant;
 	    }
 	for (i = 0; i < ncolors; i++)
 		r[i] = g[i] = b[i] = ((i * 255) / (ncolors-1));
-	switch (raw->class)
+	switch (raw->image_class)
 	    {
 	    default:
 	    case XeImageClass_BILEVEL:
@@ -2346,13 +2502,14 @@ int dither, quant;
 	    case XeImageClass_FULLCOLOR:
 		if (raw->color_space != XeImageSpace_YCbCr)
 		    {
-			im->image_op = (dither == XeDither_FS4) ?
+			im->image_op = (im->imaging.dither == XeDither_FS4) ?
 				dither_24 : map_24;
 			break;
 		    }
 		/* --- On YCbCr, fall trhough to GRAYSCALE -- */
 	    case XeImageClass_GRAYSCALE:
-		im->image_op = (dither == XeDither_FS4) ? dither_8 : map_8;
+		im->image_op = (im->imaging.dither == XeDither_FS4)
+			? dither_8 : map_8;
 		break;
 	    case XeImageClass_PALETTE:
 		mapsize = 1 << raw->bits_per_sample;
@@ -2365,26 +2522,46 @@ int dither, quant;
 		else
 			for (i = 0; i < mapsize; i++)
 				map[i] = MONO(ir[i], ig[i], ib[i]);
-		im->image_op = (dither == XeDither_FS4) ? dither_8 : map_8;
+		im->image_op = (im->imaging.dither == XeDither_FS4)
+			? dither_8 : map_8;
 		break;
 	    }
-	_XeAllocColors((XeBasicWidget)w,ncolors,r,g,b,map+512,256,1,256);
+
+	im->num_pixels = ncolors;
+	im->pixels = (Pixel *)XtMalloc(ncolors * sizeof(Pixel *));
+	im->colormap = XeAllocColors /* FIX! --msa */
+		(im->display, DefaultRootWindow(im->display),
+		 im->colormap,
+		 im->imaging.colormap_use, im->imaging.max_colors, ncolors,
+		 r, g, b, im->pixels);
+	for (i = 0; i < 256; ++i)
+	    {
+		int p = map[512+i];
+
+		/*
+		  Put in the intensity value of the pixel actually
+		  assigned (replace the original requested value)
+		  */
+		map[256+i] = MONO(r[p], g[p], b[p]);
+
+		/* Ugh!.. look into this later!! Here it just assumes that
+		   all significant bits in the pixel fit into 8bits!! -- msa
+		   */
+		map[512+i] = im->pixels[p];
+	    }
     }
 
 /*
 **	Initialize Pseudocolor display for PALETTE images to be shown
 **	in COLOR.
 */
-static void InitDisplay_PALETTE_COLOR(w, im, raw, dither, quant)
-XeBasicWidget w;
+static void InitDisplay_PALETTE_COLOR(im, raw)
 XeImage *im;
 XeRawImage *raw;
-int dither, quant;
     {
 	static XeImageOp map_8[] = 
 	    {
 		/*   Op			in	val	out	dst,src */
-		{ZeroChannel,		0,	0,	0,	0,0,},
 		{ScaleDropPixels8to8,	0,	1*256,	2*256,	0,1,},
 		{AllocateColors8,	3*256,	4*256,	5*256,	0,0,},
 		{NULL,			0,	0,	0,	0,0,},
@@ -2408,8 +2585,8 @@ int dither, quant;
 
 /*
 ** InitDisplay_TRUECOLOR_COLOR
-**	Initialize TrueColor or DirectColor display for FULLCOLOR
-**	images to be shown in color.
+**	Initialize TrueColor or DirectColor display for any images
+**	depth > 1 to be shown in color.
 */
 static int highbit(ul)
 unsigned long ul;
@@ -2422,11 +2599,9 @@ unsigned long ul;
 	return i;
     }
 
-static void InitDisplay_TRUECOLOR_COLOR(w, im, raw, dither, quant)
-XeBasicWidget w;
+static void InitDisplay_TRUECOLOR_COLOR(im, raw)
 XeImage *im;
 XeRawImage *raw;
-int dither, quant;
     {
 	static XeImageOp true_dither_8[] = 
 	    {
@@ -2473,6 +2648,7 @@ int dither, quant;
 		{ScaleDitherChannel8to16,8*256,	9*256,	10*256,	0,3,},
 		{NULL,			0,	0,	0,	0,0,},
 	    };
+#if 0
 	static XeImageOp true_map_16[] = 
 	    {
 		/*   Op			in	val	out	dst,src */
@@ -2482,6 +2658,32 @@ int dither, quant;
 		{ScaleMapChannel8to16,	8*256,	9*256,	10*256,	0,3,},
 		{NULL,			0,	0,	0,	0,0,},
 	    };
+#else
+	static XeImageOp true_map_16[] = 
+	    {
+		/*   Op			in	val	out	dst,src */
+		{ScaleMapXYZto16,	0,	0,	0,	0,1,},
+		{NULL,			0,	0,	0,	0,0,},
+	    };
+	static XeImageOp ycc_map_16[] = 
+	    {
+		/*   Op			in	val	out	dst,src */
+		{ScaleMapYCCto16,	0,	0,	0,	0,1,},
+		{NULL,			0,	0,	0,	0,0,},
+	    };
+
+	static XeImageOp gray_map_16[] =
+	    {
+		{MakeImageChannel8,	0,	0,	0,	-1,0,},
+		{ZeroChannel,		0,	0,	0,	0,0,},
+		{ScaleConvertRGBto8,	0,	0,	0,	-1,1,},
+		{ScaleMapChannel8to16,	0,	1*256,	2*256,	0,-1,},
+		{ScaleMapChannel8to16,	4*256,	5*256,	6*256,	0,-1,},
+		{ScaleMapChannel8to16,	8*256,	9*256,	10*256,	0,-1,},
+		{FreeImageChannel,	0,	0,	0,	-1,0,},
+		{NULL,			0,	0,	0,	0,0,},
+	    };
+#endif
 	static XeImageOp palette_dither_16[] = 
 	    {
 		/*   Op			in	val	out	dst,src */
@@ -2520,14 +2722,36 @@ int dither, quant;
 	    };
 
 	static XeImageOp
-		*palette_map[4] =
-		    {palette_map_8,palette_map_16,0,palette_map_32},
-		*palette_dither[4] =
-		    {palette_dither_8,palette_dither_16,0,palette_map_32},
-		*true_map[4] = {true_map_8,true_map_16,0,true_map_32},
-		*true_dither[4] = {true_dither_8,true_dither_16,0,true_map_32};
+		*palette_map[4][4] =
+		    {
+			{palette_map_8,palette_map_16,0,palette_map_32},
+			{palette_map_8,palette_map_16,0,palette_map_32},
+			{palette_map_8,palette_map_16,0,palette_map_32},
+			{palette_map_8,palette_map_16,0,palette_map_32},
+		    },
+		*palette_dither[4][4] =
+		    {
+			 {palette_dither_8,palette_dither_16,0,palette_map_32},
+			 {palette_dither_8,palette_dither_16,0,palette_map_32},
+			 {palette_dither_8,palette_dither_16,0,palette_map_32},
+			 {palette_dither_8,palette_dither_16,0,palette_map_32},
+		    },
+		*true_map[4][4] =
+		    {
+			{true_map_8,true_map_16,0,true_map_32},
+			{true_map_8,ycc_map_16,0,true_map_32},
+			{true_map_8,gray_map_16,0,true_map_32},
+			{true_map_8,gray_map_16,0,true_map_32},
+		    },
+		*true_dither[4][4] =
+		    {
+			{true_dither_8,true_dither_16,0,true_map_32},
+			{true_dither_8,ycc_map_16,0,true_map_32},
+			{true_dither_8,gray_map_16,0,true_map_32},
+			{true_dither_8,gray_map_16,0,true_map_32},
+		    };
 	
-	int i, rs, gs, bs, out_width;
+	int i, rs, gs, bs, out_width, c;
 	unsigned long rm = im->image->red_mask;
 	unsigned long gm = im->image->green_mask;
 	unsigned long bm = im->image->blue_mask;
@@ -2620,7 +2844,9 @@ int dither, quant;
 		    }
 		break;
 	    }
-	if (raw->class == XeImageClass_PALETTE)
+	c = (raw->color_space == XeImageSpace_YCbCr) +
+		((im->imaging.color_mode == XeColorMode_GRAY) << 1);
+	if (raw->image_class == XeImageClass_PALETTE)
 	    {
 		unsigned char *r, *g, *b;
 		int mapsize = 1 << raw->bits_per_sample;
@@ -2634,15 +2860,30 @@ int dither, quant;
 			g_in[i] = g[i];
 			b_in[i] = b[i];
 		    }
-		im->image_op = (dither == XeDither_FS4) ?
-			palette_dither[out_width-1] : palette_map[out_width-1];
+		im->image_op = (im->imaging.dither == XeDither_FS4) ?
+			palette_dither[c][out_width-1] :
+				palette_map[c][out_width-1];
+	    }
+	else if (raw->image_class == XeImageClass_GRAYSCALE)
+	    {
+		int mapsize = 1 << raw->bits_per_sample;
+
+		for (i = 0; i < mapsize; i++)
+		    {
+			r_in[i] = i;
+			g_in[i] = i;
+			b_in[i] = i;
+		    }
+		im->image_op = (im->imaging.dither == XeDither_FS4) ?
+			palette_dither[c][out_width-1] :
+				palette_map[c][out_width-1];
 	    }
 	else
 	    {
 		for (i = 0; i < XeSample_RANGE; i++)
 			r_in[i] = g_in[i] = b_in[i] = i;
-		im->image_op = (dither == XeDither_FS4) ?
-			true_dither[out_width-1] : true_map[out_width-1];
+		im->image_op = (im->imaging.dither == XeDither_FS4) ?
+			true_dither[c][out_width-1] : true_map[c][out_width-1];
 	    }
     }
 
@@ -2651,11 +2892,9 @@ int dither, quant;
 **	Initialize for displaying the image as bilevel dithered image.
 **	Any visual, any image class.
 */
-static void InitDisplay_ANY_MONO(w, im, raw, dither, quant)
-XeBasicWidget w;
+static void InitDisplay_ANY_MONO(im, raw)
 XeImage *im;
 XeRawImage *raw;
-int dither, quant;
     {
 	static XeImageOp dither_1[] = 
 	    {
@@ -2686,7 +2925,7 @@ int dither, quant;
 	unsigned char *map, *r, *g, *b;
 	int mapsize;
 
-	switch (raw->class)
+	switch (raw->image_class)
 	    {
 	    default:
 	    case XeImageClass_BILEVEL:
@@ -2740,107 +2979,98 @@ int dither, quant;
 			map[i] = fs_gamcr[255-i];
     }
 
-static void InitDisplay_GRAYSCALE(w, im, raw, dither, quant)
-XeBasicWidget w;
+static void InitDisplay_GRAYSCALE(im, raw)
 XeImage *im;
 XeRawImage *raw;
-int dither, quant;
     {
-	switch (w->basic.visual->class)
+	switch (im->visual->class)
 	    {
 	    case TrueColor:
 	    case DirectColor:
 		/* Should do something different?? */
-
+		InitDisplay_TRUECOLOR_COLOR(im, raw);
+		break;
 	    case StaticColor:
 	    case GrayScale:
 	    case StaticGray:
 	    case PseudoColor:
 	    default:
-		InitDisplay_PSEUDOCOLOR_GRAY(w, im, raw, dither, quant);
+		InitDisplay_PSEUDOCOLOR_GRAY(im, raw);
 		break;
 	    }
     }
 
-static void InitImage_GRAYSCALE(w, im, raw, mode, dither, quant)
-XeBasicWidget w;
+static void InitImage_GRAYSCALE(im, raw)
 XeImage *im;
 XeRawImage *raw;
-XeColorMode mode;
-int dither, quant;
     {
-	if (mode == XeColorMode_MONO)
-		InitDisplay_ANY_MONO(w, im, raw, dither, quant);
+	if (im->imaging.color_mode == XeColorMode_MONO)
+		InitDisplay_ANY_MONO(im, raw);
 	else
-		InitDisplay_GRAYSCALE(w, im, raw, dither, quant);
+		InitDisplay_GRAYSCALE(im, raw);
     }
 
-static void InitImage_PALETTE(w, im, raw, mode, dither, quant)
-XeBasicWidget w;
+static void InitImage_PALETTE(im, raw)
 XeImage *im;
 XeRawImage *raw;
-XeColorMode mode;
-int dither, quant;
     {
-	switch (mode)
+	switch (im->imaging.color_mode)
 	    {
 	    case XeColorMode_MONO:
 	    default:
-		InitDisplay_ANY_MONO(w, im, raw, dither, quant);
+		InitDisplay_ANY_MONO(im, raw);
 		break;
 	    case XeColorMode_GRAY:
-		InitDisplay_GRAYSCALE(w, im, raw, dither, quant);
+		InitDisplay_GRAYSCALE(im, raw);
 		break;
 	    case XeColorMode_COLOR:
-		switch (w->basic.visual->class)
+		switch (im->visual->class)
 		    {
 		    case TrueColor:
 		    case DirectColor:
-			InitDisplay_TRUECOLOR_COLOR(w, im, raw, dither, quant);
+			InitDisplay_TRUECOLOR_COLOR(im, raw);
 			break;
 		    case StaticColor:
 		    case GrayScale:
 		    case StaticGray:
 		    case PseudoColor:
 		    default:
-			InitDisplay_PALETTE_COLOR(w, im, raw, dither, quant);
+			InitDisplay_PALETTE_COLOR(im, raw);
 			break;
 		    }
 		break;
 	    }
     }
 
-static void InitImage_FULLCOLOR(w, im, raw, mode, dither, quant)
-XeBasicWidget w;
+static void InitImage_FULLCOLOR(im, raw)
 XeImage *im;
 XeRawImage *raw;
-XeColorMode mode;
-int dither, quant;
     {
-	switch (mode)
+	switch (im->imaging.color_mode)
 	    {
 	    default:
 	    case XeColorMode_MONO:
-		InitDisplay_ANY_MONO(w, im, raw, dither, quant);
+		InitDisplay_ANY_MONO(im, raw);
 		break;
 	    case XeColorMode_GRAY:
-		InitDisplay_GRAYSCALE(w, im, raw, dither, quant);
+		InitDisplay_GRAYSCALE(im, raw);
 		break;
 	    case XeColorMode_COLOR:
-		switch (w->basic.visual->class)
+		switch (im->visual->class)
 		    {
 		    case TrueColor:
 		    case DirectColor:
-			InitDisplay_TRUECOLOR_COLOR(w, im, raw, dither, quant);
+			InitDisplay_TRUECOLOR_COLOR(im, raw);
 			break;
 		    default:
 		    case StaticColor:
 		    case PseudoColor:
-			if (quant == XeColorQuantize_FAST ||
-			    quant == XeColorQuantize_JPEG)
-				InitDisplay_COLORCUBE(w,im,raw,dither,quant);
+			if (im->imaging.color_quantize == XeColorQuantize_FAST
+			    ||
+			    im->imaging.color_quantize == XeColorQuantize_JPEG)
+				InitDisplay_COLORCUBE(im, raw);
 			else
-				InitDisplay_COLORQUANT(w,im,raw,dither,quant);
+				InitDisplay_COLORQUANT(im, raw);
 			break;
 		    }
 		break;
@@ -2848,65 +3078,53 @@ int dither, quant;
     }
 /*
 ** _XeCreateImage
-**	Create Xew Image structure for displaying the width*height
-**	(logical) source image on the X Window. (width/height are
-**	not real image size, they just determine the scaling
-**	and final image shape on the screen).
+**	Create Xew Image structure (image pipeline) for converting the
+**	raw image data into display format.
 */
-XeImage *_XeCreateImage(w, raw, dither, quant, width, height)
-XeBasicWidget w;
+XeImage *_XeCreateImage(display, window, raw, imaging, width, height)
+Display *display;
+Window window;
 XeRawImage *raw;
-int dither, quant, width, height;
+XeImaging *imaging;
+int width, height;
     {
 	static int init;
 	XImage *image;
 	XeImage *im;
 	Dimension dw, dh;
 	int format, depth, bitmap_pad;
-	XeColorMode mode;
+	XWindowAttributes attr;
 
 	if (!init)
 	    {
 		InitFSDither();
 		init = 1;
 	    }
-	if (w->basic.visual == NULL)
-		w->basic.visual = XeGetVisual((Widget)w);
-	_XeBasicScaling((XeBasicWidget)w, width, height,
-			w->core.width, w->core.height, &dw, &dh);
-	/*
-	** Request resize?
-	*/
-	if (w->basic.resize && (w->core.width != dw || w->core.height != dh))
-	    {
-		Dimension rw = dw;
-		Dimension rh = dh;
-		int do_almost = 1;
-		
-		while (XtMakeResizeRequest((Widget)w, rw, rh, &rw, &rh)
-		       == XtGeometryAlmost && do_almost)
-			do_almost = 0;
-		/*
-		** Recompute new scaled size
-		*/
-		_XeBasicScaling((XeBasicWidget)w, width, height,
-				rw, rh, &dw, &dh);
-	    }
+
+	XGetWindowAttributes(display, window, &attr);
+
+	dw = width;
+	dh = height;
+
     retry_after_out_of_memory:
 	im = (XeImage *)XtCalloc(1, sizeof(XeImage));
+	im->imaging = *imaging;
+	im->display = display;
+	im->colormap = attr.colormap;
+	im->visual = attr.visual;
 	/*
 	** Constrain the requested color_mode by display capabilities and
 	** choose XImage parameters according to the result. (If NONE, try
 	** for COLOR by default).
 	*/
-	if ((mode = w->basic.color_mode) == XeColorMode_NONE)
-		mode = XeColorMode_COLOR;
-	switch (w->basic.visual->class)
+	if (im->imaging.color_mode == XeColorMode_NONE)
+		im->imaging.color_mode = XeColorMode_COLOR;
+	switch (im->visual->class)
 	    {
 	    case GrayScale:
 	    case StaticGray:
-		if (mode == XeColorMode_COLOR)
-			mode = XeColorMode_GRAY;
+		if (im->imaging.color_mode == XeColorMode_COLOR)
+			im->imaging.color_mode = XeColorMode_GRAY;
 		break;
 	    case TrueColor:
 	    case StaticColor:
@@ -2914,19 +3132,21 @@ int dither, quant, width, height;
 	    case DirectColor:
 		break;
 	    default:
-		mode = XeColorMode_MONO;
+		im->imaging.color_mode = XeColorMode_MONO;
 	    }
-	if (w->core.depth == 1 || w->basic.max_colors < 2 ||
-	    raw->class == XeImageClass_BILEVEL || mode == XeColorMode_MONO)
+	if (attr.depth == 1 ||
+	    im->imaging.max_colors < 2 ||
+	    raw->image_class == XeImageClass_BILEVEL ||
+	    im->imaging.color_mode == XeColorMode_MONO)
 	    {
 		depth = 1;
 		bitmap_pad = 32;
 		format = XYBitmap;
-		mode = XeColorMode_MONO;
+		im->imaging.color_mode = XeColorMode_MONO;
 	    }
 	else
 	    {
-		depth = w->core.depth;
+		depth = attr.depth;
 		bitmap_pad = depth > 8 ? 32 : 8;
 		format = ZPixmap;
 	    }
@@ -2935,19 +3155,19 @@ int dither, quant, width, height;
 		
 #ifdef SH_MEM
 		im->shared = False;
-		if (w->basic.use_shm == 2)
+		if (im->imaging.use_shm == 2)
 		    {
-			int server = XShmPixmapFormat(XtDisplay(w));
+			int server = XShmPixmapFormat(display);
 			if (server != format)
-				w->basic.use_shm = 1;
+				im->imaging.use_shm = 1;
 		    }
-		if (w->basic.use_shm)
+		if (im->imaging.use_shm)
 		    {
 			im->shm.shmid = -1;
 			im->shm.shmaddr = (char *)-1;
 			im->image = image = XShmCreateImage
-				(XtDisplay(w),
-				 w->basic.visual,	/* visual */
+				(display,
+				 attr.visual,		/* visual */
 				 depth,			/* depth */
 				 format,		/* format */
 				 (char *)NULL,		/* Bitmap */
@@ -2969,9 +3189,9 @@ int dither, quant, width, height;
 			im->shm.readOnly = False;
 			if (image->data == NULL)
 				goto try_non_shared;
-			if (!XShmAttach(XtDisplay(w), &im->shm))
+			if (!XShmAttach(display, &im->shm))
 				goto try_non_shared;
-			XSync(XtDisplay(w), 0);
+			XSync(display, 0);
 			/*
 			** To prevent unwanted accumulation of shared segments
 			** due abnormal program exits, delete the segment
@@ -2979,10 +3199,10 @@ int dither, quant, width, height;
 			** from all processess.
 			*/
 			shmctl(im->shm.shmid, IPC_RMID, 0);
-			if (w->basic.use_shm == 2)
+			if (im->imaging.use_shm == 2)
 				im->pixmap = XShmCreatePixmap
-					(XtDisplay(w),
-					 XtWindow(w),
+					(display,
+					 window,
 					 (char *)im->shm.shmaddr,
 					 &im->shm,
 					 (unsigned int)dw,
@@ -3004,15 +3224,17 @@ int dither, quant, width, height;
 			im->shared = False;
 			im->shm.shmid = -1;
 			im->shm.shmaddr = (char *)-1;
+#if 0
 			XeWidgetWarningMsg
-				((Widget)w, "noteImageToolsNoShm",
+				(w, "noteImageToolsNoShm",
 				 "Shared memory not usable",
 				 (String *)NULL, (Cardinal)0);
+#endif
 		    }
 #endif
 		im->image = image = XCreateImage
-			(XtDisplay(w),
-			 w->basic.visual,	/* visual */
+			(display,
+			 attr.visual,		/* visual */
 			 depth,			/* depth */
 			 format,		/* format */
 			 0,			/* Offset */
@@ -3027,10 +3249,12 @@ int dither, quant, width, height;
 		    {
 			XFree((char *)image);
 			XtFree((char *)im);
-			XeWidgetWarningMsg((Widget)w,"noteImageToolsOutMemory",
+#if 0
+			XeWidgetWarningMsg(w,"noteImageToolsOutMemory",
 					   "Out of memory, reducing the size",
 					   (String *)NULL, 0);
-			XFlush(XtDisplay(w));
+#endif
+			XFlush(display);
 			/*
 			** This is a bit gross, but things are getting
 			** desperate.. even dirty goto's will do.. :-)
@@ -3052,20 +3276,37 @@ int dither, quant, width, height;
 		image->byte_order = MSBFirst;
 		image->bitmap_unit = 8;
 	    }
-	switch (raw->class)
+	else if (format == ZPixmap && image->bitmap_unit > 8)
+	    {
+		/*
+		** ..for now, if the host byte order and display byte order
+		** are different, just let the Xlib deal with it and use
+		** the host byte order in XImage (may not be the most
+		** efficient way to solve this... )
+		*/
+		union
+		    {
+			long l;
+			char c[sizeof(long)];
+		    } u;
+		u.l = 1;
+		image->byte_order =
+			(u.c[sizeof(long)-1] == 1) ? MSBFirst : LSBFirst;
+	    }
+	switch (raw->image_class)
 	    {
 	    default:
 	    case XeImageClass_BILEVEL:
-		InitDisplay_ANY_MONO(w, im, raw, dither, quant);
+		InitDisplay_ANY_MONO(im, raw);
 		break;
 	    case XeImageClass_GRAYSCALE:
-		InitImage_GRAYSCALE(w, im, raw, mode, dither, quant);
+		InitImage_GRAYSCALE(im, raw);
 		break;
 	    case XeImageClass_PALETTE:
-		InitImage_PALETTE(w, im, raw, mode, dither, quant);
+		InitImage_PALETTE(im, raw);
 		break;
 	    case XeImageClass_FULLCOLOR:
-		InitImage_FULLCOLOR(w, im, raw, mode, dither, quant);
+		InitImage_FULLCOLOR(im, raw);
 		break;
 	    }
 	return im;
@@ -3125,8 +3366,7 @@ XeImageChannel *src, *dst;
 **	Build the actuall XImage from the raw input image using the
 **	control structures.
 */
-void _XeBuildImage(w, im, n, data)
-XeBasicWidget w;	/* Widget */
+void _XeBuildImage(im, n, data)
 XeImage *im;		/* Xew Image Control Structure */
 int n;			/* Number of entries in data */
 XeImageChannel *data;	/* Raw Image data channels */
@@ -3141,14 +3381,15 @@ XeImageChannel *data;	/* Raw Image data channels */
 			continue;	/* Ref to non-existent data, skip op */
 		dst = (op->dst < 0) ? &temp[-op->dst - 1] : &data[op->dst];
 		src = (op->src < 0) ? &temp[-op->src - 1] : &data[op->src];
-		(*op->op)(w, &map[op->in_offset], &map[op->val_offset],
+		(*op->op)(im, &map[op->in_offset], &map[op->val_offset],
 			  &map[op->out_offset], dst, src);
 	    }
     }
 
 
-void _XePutImage(w, gc, im, src_x, src_y, dst_x, dst_y, width, height)
-XeBasicWidget w;
+void _XePutImage(dsp, dst, gc, im, src_x, src_y, dst_x, dst_y, width, height)
+Display *dsp;
+Drawable dst;
 GC gc;
 XeImage *im;
 int src_x, src_y;
@@ -3169,19 +3410,18 @@ unsigned int height;
 #ifdef SH_MEM
 	if (im->shared)
 	    {
-		(void)XShmPutImage(XtDisplay(w), XtWindow(w),
+		(void)XShmPutImage(dsp, dst,
 				   gc, im->image,
 				   src_x, src_y, dst_x, dst_y,
 				   width, height, False);
 		return;
 	    }
 #endif
-	XPutImage(XtDisplay(w), XtWindow(w), gc, im->image,
+	XPutImage(dsp, dst, gc, im->image,
 		  src_x, src_y, dst_x, dst_y, width, height);
     }
 
-void _XeDestroyImage(w, im)
-XeBasicWidget w;
+void _XeDestroyImage(im)
 XeImage *im;
     {
 	if (im == NULL)
@@ -3192,13 +3432,13 @@ XeImage *im;
 #ifdef SH_MEM
 	if (im->shared)
 	    {
-		XShmDetach(XtDisplay(w), &im->shm);
+		XShmDetach(im->display, &im->shm);
 		shmdt(im->shm.shmaddr);
 		/* Note: shared memory id is already deleted. no need
 		** to call "shmctl(im->shm.shmid, IPC_RMID, 0)"
 		*/
 		if (im->pixmap)
-			XFreePixmap(XtDisplay(w), im->pixmap);
+			XFreePixmap(im->display, im->pixmap);
 	    }
 #endif
 	if (im->image)
@@ -3206,7 +3446,16 @@ XeImage *im;
 	/*
 	** Release old colors if any allocated
 	*/
-	_XeFreeColors((XeBasicWidget)w);
+	if (im->num_pixels)
+	    {
+		/* "New style" color handling */
+		/* Problem! Should restore default colormap to the
+		   window/widget, if the private colormap was released here!
+		   */
+		XeFreeColors(im->display, im->colormap,
+			     im->num_pixels, im->pixels);
+		XtFree((char *)im->pixels);
+	    }
 	XtFree((char *)im->map);
 	XtFree((char *)im);
     }
