@@ -50,10 +50,15 @@ extern void bcopy();
 #if ANSI_INCLUDES
 #	include <stdlib.h>
 #	include <string.h>
+#endif
 /*
-** We have 'memmove' with ansi includes
+** If don't have memmove, but need overlapping copy.. must use 'bcopy' and
+** just assume it exists...
 */
+#if HAVE_MEMMOVE
 #	define bcopy(a,b,c) memmove(b,a,c)
+#else
+extern void bcopy();
 #endif
 #include <stdio.h>
 #include <assert.h>
@@ -593,17 +598,17 @@ int r1, g1, b1;
 			ptr->entries[ptr->num_ents][1] = 0;
 			++ptr->num_ents;
 			
-			tmp = *rp - r1;
+			tmp = (int)*rp - r1;
 			if (tmp < (MAX_COLOR/C_LEN/2))
 				tmp = MAX_COLOR/C_LEN-1 - tmp;
 			dist = (tmp*tmp * R2FACT);
 			
-			tmp = *gp - g1;
+			tmp = (int)*gp - g1;
 			if (tmp < (MAX_COLOR/C_LEN/2))
 				tmp = MAX_COLOR/C_LEN-1 - tmp;
 			dist += (tmp*tmp * G2FACT);
 			
-			tmp = *bp - b1;
+			tmp = (int)*bp - b1;
 			if (tmp < (MAX_COLOR/C_LEN/2))
 				tmp = MAX_COLOR/C_LEN-1 - tmp;
 			dist += (tmp*tmp * B2FACT);
@@ -624,14 +629,14 @@ int r1, g1, b1;
 			
 			dist = 0;
 			
-			if ((tmp = r1 - *rp)>0 ||
-			    (tmp = *rp - (r1 + MAX_COLOR/C_LEN-1)) > 0 )
+			if ((tmp = r1 - (int)*rp)>0 ||
+			    (tmp = (int)*rp - (r1 + MAX_COLOR/C_LEN-1)) > 0 )
 				dist += (tmp*tmp * R2FACT);
-			if((tmp = g1 - *gp)>0 ||
-			   (tmp = *gp - (g1 + MAX_COLOR/C_LEN-1)) > 0 )
+			if((tmp = g1 - (int)*gp)>0 ||
+			   (tmp = (int)*gp - (g1 + MAX_COLOR/C_LEN-1)) > 0 )
 				dist += (tmp*tmp * G2FACT);
-			if((tmp = b1 - *bp)>0 ||
-			   (tmp = *bp - (b1 + MAX_COLOR/C_LEN-1)) > 0 )
+			if((tmp = b1 - (int)*bp)>0 ||
+			   (tmp = (int)*bp - (b1 + MAX_COLOR/C_LEN-1)) > 0 )
 				dist += (tmp*tmp * B2FACT);
 			if( dist < mindist )
 			    {
