@@ -33,8 +33,6 @@ typedef struct
 	unsigned int clip_width, clip_height;
 	unsigned int raster_width, raster_height;
 	Dimension offset_x, offset_y;
-	XeColorQuantize color_quantize;
-	XeDither dither;
 
 	/* private state */
 
