@@ -1,5 +1,5 @@
 /*
-** Copyright 1992, 1993, 1994, 1995 by Markku Savela and
+** Copyright 1992-1995 by Markku Savela and
 **	Technical Research Centre of Finland
 */
 #include <stdio.h>
@@ -26,15 +26,19 @@ extern int atoi( /* No header for this in SYSV */
 #include <X11/Xmu/CharSet.h>
 #include <X11/Xew/XewInit.h>
 #include <X11/Xew/TextP.h>
+
 #include "TextImport.h"
 #include "TextLayout.h"
 #include "TextFont.h"
 #include "FrameSupport.h"
+#include "Color.h"
 
 #define TEXT_MIN_WIDTH 1
 #define TEXT_MIN_HEIGHT 1
-#define TEXT_SET_WIDTH(t) (int)\
-	((t)->text.column_width > 0 ? (t)->text.column_width : (t)->core.width)
+
+#ifndef INT_MAX
+#	define INT_MAX (~((unsigned)0) >> 1)
+#endif
 
 static void
 	Highlight(), Unhighlight();
@@ -50,6 +54,7 @@ static XeFrameDimension init_FrameDimension;
 static XeFrameBorder init_FrameBorder;
 static XeFrameAlignment init_FrameAlignment = XeFrameAlignment_ABOVE;
 static XeFrameBorder framed_FrameBorder = {XeBorderType_SOLID,1,1};
+static int init_Columns = 1;
 
 #define offset(field) XtOffsetOf(XeTextRec, text.field)
 
@@ -93,10 +98,16 @@ static XtResource resources[] =
 		 XtRImmediate,(XtPointer)20},
 	{XtNgraphicRendition, XtCRendition, XtRString, sizeof(String),
 		 offset(graphic_rendition), XtRImmediate, (XtPointer)NULL},
-	{XtNindentation, XtCIndentation, XtRDimension, sizeof(Dimension),
-		 offset(indentation), XtRImmediate, (Dimension)0 },
-	{XtNcolumnWidth, XtCColumnWidth, XtRDimension, sizeof(Dimension),
-		 offset(column_width), XtRImmediate, (Dimension)0 },
+	{XtNindentation, XtCIndentation, XtRInt, sizeof(int),
+		 offset(indentation), XtRImmediate, 0 },
+	{XtNcolumnHeight, XtCColumnHeight, XtRInt, sizeof(int),
+		 offset(column_height), XtRImmediate, 0 },
+	{XtNcolumns, XtCColumns, XtRInt, sizeof(int),
+		 offset(columns), XtRInt, (XtPointer)&init_Columns },
+	{XtNcolumnSeparation, XtCColumnSeparation, XtRInt, sizeof(int),
+		 offset(column_separation), XtRImmediate, 0 },
+	{XtNcolumnWidth, XtCColumnWidth, XtRInt, sizeof(int),
+		 offset(column_width), XtRImmediate, 0 },
 	{XtNinitialState, XtCSequence,  XtRString, sizeof(String),
 		 offset(initial_state), XtRImmediate, (XtPointer)NULL},
 	{XtNitemization,XtCItemization, XtRXeItemization,
@@ -110,8 +121,10 @@ static XtResource resources[] =
 	{XtNlineLayoutTable, XtCLineLayoutTable, XtRXeLineLayoutTable,
 		 sizeof(XeTabStop *),
 		 offset(line_layout_table), XtRImmediate, (XtPointer)NULL},
-	{XtNlineSpacing, XtCLineSpacing, XtRDimension, sizeof(Dimension),
-		 offset(line_spacing), XtRImmediate, (Dimension)0},
+	{XtNlineSpacing, XtCLineSpacing, XtRInt, sizeof(int),
+		 offset(line_spacing), XtRImmediate, 0},
+	{XtNpageSeparation, XtCPageSeparation, XtRInt, sizeof(int),
+		 offset(page_separation), XtRImmediate, 0},
 	{XtNproportional, XtCProportional, XtRBoolean, sizeof(Boolean),
 	     offset(proportional), XtRImmediate, (XtPointer)0},
     };
@@ -188,6 +201,11 @@ static XtGeometryResult QueryGeometry();
 static XtGeometryResult GeometryManager();
 static void InsertChild(), DeleteChild(), ConstraintDestroy(), ChangeManaged();
 
+static void PrintInitialize();
+static void PrintCancel(), PrintDestroy();
+static XtPointer Print();
+static void PrintConstraintInitialize(), PrintConstraintDestroy();
+
 /*
 ** REDISPLAY
 **	Call widget instance expose method. As Text may be subclassed, this
@@ -270,35 +288,41 @@ XeTextClassRec xeTextClassRec = {
     /* extension		*/	NULL,
 #endif
   },
-  { /* composite class fields */
-    /* geometry_manager		*/   GeometryManager,
-    /* change_managed		*/   ChangeManaged,
-    /* insert_child		*/   InsertChild,
-    /* delete_child		*/   DeleteChild,
-    /* extension		*/   NULL
+  { /* composite class fields	*/
+    /* geometry_manager		*/	GeometryManager,
+    /* change_managed		*/	ChangeManaged,
+    /* insert_child		*/	InsertChild,
+    /* delete_child		*/	DeleteChild,
+    /* extension		*/	NULL
+  },
+  { /* constraint class fields	*/
+    /* subresources		*/	subresources,
+    /* subresource_count	*/	XtNumber(subresources),
+    /* constraint_size		*/	sizeof(XeTextConstraintsRec),
+    /* initialize		*/	_XeFrameInitialize,
+    /* destroy			*/	ConstraintDestroy,
+    /* set_values		*/	FrameSetValues,
+    /* extension		*/	NULL
   },
-  { /* constraint class fields */
-    /* subresources		*/   subresources,
-    /* subresource_count	*/   XtNumber(subresources),
-    /* constraint_size		*/   sizeof(XeTextConstraintsRec),
-    /* initialize		*/   _XeFrameInitialize,
-    /* destroy			*/   ConstraintDestroy,
-    /* set_values		*/   FrameSetValues,
-    /* extension		*/   NULL
-   },
 #ifdef USING_MOTIF_122
-    {  /* manager class record */
-       /* translations */                 XtInheritTranslations,
-       /* syn_resources */                NULL,
-       /* num_syn_resources */            0,
-       /* syn_constraint_resources */     NULL,
-       /* num_syn_constraint_resources */ 0,
-       /* parent_process */               XmInheritParentProcess,
-       /* extension */                    NULL  
-    },
+  { /* manager class record	*/
+    /* translations		*/	XtInheritTranslations,
+    /* syn_resources		*/	NULL,
+    /* num_syn_resources	*/	0,
+    /* syn_constraint_resources	*/	NULL,
+    /* num_syn_constraint_resources */	0,
+    /* parent_process		*/	XmInheritParentProcess,
+    /* extension		*/	NULL  
+  },
 #endif
-  { /* XeBasic fields */
-    /* not used			*/	0
+  { /* basic fields */
+    /* print_initialize		*/	PrintInitialize,
+    /* print			*/	Print,
+    /* print_cancel		*/	PrintCancel,
+    /* print_destroy		*/	PrintDestroy,
+    /* print_constraint_initialize */	PrintConstraintInitialize,
+    /* print_constraint_destroy */	PrintConstraintDestroy
+
   },
   { /* XeText fields */
     /* not used			*/	0
@@ -351,20 +375,6 @@ static CvtEnumInfo cvtExportFormat[] =
 	return True;					\
     } while (0)
 
-
-/*
-** XeFontsDestructor
-*/
-static void XeFontsDestructor(app, to, converter_data, args, num_args)
-XtAppContext app;
-XrmValue *to;
-XtPointer converter_data;
-XrmValue *args;
-Cardinal *num_args;
-    {
-	XtFree((char *)converter_data);
-    }
-
 /*
 ** XeCvtStringToFonts
 **	Convert a string into Fonts (XeFont *)
@@ -857,6 +867,8 @@ XtWidgetGeometry *current;
 		x = inset->x + inset->offset;
 		y = inset->y - inset->ascent;
 	    }
+	x += t->basic.x_translation;
+	y += t->basic.y_translation;
 	/*
 	** A quick fix: guess a correction for y, if the snip
 	** appears to be framed and inline. A very specific
@@ -893,31 +905,168 @@ XeTextWidget t;
 
 	for (i = 0; i < t->composite.num_children; i++)
 	    {
-
 		w = t->composite.children[i];
-		if (!XtIsManaged(w))
+		if (!XtIsManaged(w) || !GetInsetGeometry(t, w, &set))
 			continue;
-		if (GetInsetGeometry(t, w, &set))
+		if (t->text.printing)
+		    {
+			/*
+			** If printing is set, assume we are really
+			** working with the cloned tree. As currently
+			** each cloned widget really has the same
+			** window, doing reconfigure to it is not
+			** the best thing... (especially, when
+			** there might not be any real window!!)
+			*/
+			w->core.x = set.x;
+			w->core.y = set.y;
+			if (w->core.width != set.width ||
+			    w->core.height != set.height)
+			    {
+				w->core.width = set.width;
+				w->core.height = set.height;
+				/*
+				** Call resize method directly if size changed
+				** (only if Xew widget, unknown widget resize
+				** cannot be called, because it might cause
+				** expose method being executed too..)
+				*/
+#ifndef USING_MOTIF_122
+				/*
+				** MOTIF has some strange 'ResizeWrapper'
+				** around the resize method, and this seems
+				** require 'per display information', which
+				** 'printer display' does not have! Cannot
+				** do anything else, but skip the resize
+				** totally for MOTIF subclassed Xew...
+				*/
+				if (w->core.widget_class->core_class.resize &&
+				    XtIsSubclass(w, xeBasicWidgetClass))
+					(*w->core.widget_class->
+					 core_class.resize)(w);
+#endif
+			    }
+		    }
+		else
 			XtConfigureWidget
-				(w, set.x, set.y, set.width, set.height,
+				(w, set.x, set.y,
+				 set.width, set.height,
 				 w->core.border_width);
 	    }
     }
 
+/*
+** GetTextLayoutGeometry
+**	compute a column/paging attributes for the text layout assuming
+**	the given width and height of the widget window.
+**
+**	Return non-zero, if the computed text layout geometry differs
+**	from the current setting.
+*/
+static int GetTextLayoutGeometry(w, set, width, height)
+XeTextWidget w;
+XeTextLayoutGeometry *set;
+int width, height;
+    {
+	set->columns = w->text.columns;
+	set->column_width = w->text.column_width;
+	set->column_height = w->text.column_height;
+
+	switch ((set->column_height > 0) +
+		(set->column_width > 0) * 2 +
+		(set->columns > 0) * 4)
+	    {
+	    case 0:	/* None specified, no paging (default) */
+		set->column_height = INT_MAX;
+		/* FALL TRHOUGH */
+	    case 1:	/* Only column height: "Simple Vertical Paging" */
+		set->columns = 1;
+		set->column_width = width;
+		break;
+
+	    case 2:	/* Only Column width: "Page to window" */
+		set->column_height = height;
+		/* FALL THROUGH */
+	    case 3:	/* Only Column width and height */
+		set->columns = width / set->column_width;
+		break;
+
+	    case 4:	/* Only columns */
+		set->column_height = height;
+		/* FALL THROUGH */
+	    case 5:	/* Columns and height */
+		set->column_width = width / set->columns;
+		break;
+
+	    case 6:	/* Columns and width */
+		set->column_height = height;
+		break;
+	    default:
+		break;
+	    }
+	/*
+	** Above may result invalid values, adjust some.
+	*/
+	if (set->columns < 1)
+		set->columns = 1;
+	else if (set->columns > MAX_TEXT_COLUMNS)
+		set->columns = MAX_TEXT_COLUMNS;
+
+	if (set->column_width == 0)
+	    {
+		set->columns = 1;
+		set->column_width = INT_MAX;
+	    }
+	if (set->column_height == 0)
+		set->column_height = INT_MAX;
+
+	if (set->column_width != INT_MAX)
+		set->width = set->column_width;
+	else
+		set->width = 0;
+	if (set->column_height != INT_MAX)
+		set->height = set->column_height;
+	else
+		set->height = 0;
+
+	return	set->columns != w->text.set.columns ||
+		set->width != w->text.set.width ||
+		set->height != w->text.set.height ||
+		set->column_width != w->text.set.column_width ||
+		set->column_height != w->text.set.column_height;
+    }
+
 /*
 ** RedoTextLayout
 **	totally redo the text layout and configure children
 */
-static void RedoTextLayout(t, set_width, commit)
+#if NeedFunctionPrototypes
+static void RedoTextLayout(XeTextWidget, XeTextLayoutGeometry *, int);
+#endif
+
+static void RedoTextLayout(t, set, commit)
 XeTextWidget t;
-int set_width;
+XeTextLayoutGeometry *set;
 int commit;
     {
-	_XeTextLayout(t, set_width, (Snip *)NULL, (Region)NULL);
+	_XeTextLayout(t, set, (Snip *)NULL, (Region)NULL);
 	if (commit)
 		_XeTextConfigureChildren(t);
     }
 
+static int LimitDimension(d, limit)
+int d;
+XeFrameDimension *limit;
+    {
+	if (limit->value > 0)
+		return limit->value;
+	if (limit->min > 0 && d < limit->min)
+		d = limit->min;
+	if (limit->max > 0 && d > limit->max)
+		d = limit->max;
+	return d;
+    }
+
 /*
 ** QueryGeometry
 **
@@ -931,29 +1080,38 @@ XtWidgetGeometry *request;
 XtWidgetGeometry *preferred;
     {
 	XeTextWidget t = (XeTextWidget)w;
-	int fixed_width;
-
-	fixed_width = (t->text.column_width > 0 || !t->text.format) ?
-		(t->text.w > 0 ? t->text.w : t->text.column_width) : 0;
+	XeTextLayoutGeometry set;
 
 	preferred->request_mode = CWWidth | CWHeight;
-	preferred->width =
-		(request->request_mode & CWWidth) ? request->width : 0;
-	if (preferred->width == 0 || fixed_width > 0)
-		preferred->width = fixed_width > 0 ? fixed_width:t->core.width;
-	if (fixed_width == 0 && t->text.set_width != preferred->width)
+	preferred->width = (request->request_mode & CWWidth) ?
+		request->width : 0;
+	preferred->height = (request->request_mode & CWHeight) ?
+		request->height : 0;
+	preferred->width = LimitDimension
+		(preferred->width, &t->basic.dimension[0]);
+	preferred->height= LimitDimension
+		(preferred->height,&t->basic.dimension[1]);
+	if (GetTextLayoutGeometry(t,&set,preferred->width,preferred->height))
 	    {
 		/* ...this is something I may yet regret... --msa */
-		int old_width = t->text.set_width;
-		
-		_XeTextLayout(t, preferred->width, (Snip *)NULL, (Region)0);
-		/* preferred->width = t->text.w; --msa */
+
+		XeTextLayoutGeometry old_set = t->text.set;
+
+		_XeTextLayout(t, &set, (Snip *)NULL, (Region)0);
+		preferred->width = t->text.w;
 		preferred->height = t->text.h;
-		if (old_width > 0)
-			_XeTextLayout(t, old_width, (Snip *)NULL, (Region)0);
+		_XeTextLayout(t, &old_set, (Snip *)NULL, (Region)0);
 	    }
 	else
+	    {
+		preferred->width = t->text.w;
 		preferred->height = t->text.h;
+	    }
+	preferred->width = LimitDimension
+		(preferred->width, &t->basic.dimension[0]);
+	preferred->height= LimitDimension
+		(preferred->height,&t->basic.dimension[1]);
+
 	if (((request->request_mode&(CWWidth|CWHeight))==(CWWidth|CWHeight)) &&
 	    request->width == preferred->width &&
 	    request->height == preferred->height)
@@ -1017,11 +1175,18 @@ int dolayout;
 	if (w->text.enable_display < 0)
 		dolayout = 0;
 	if (dolayout)
-		RedoTextLayout(w, TEXT_SET_WIDTH(w), True);
+	    {
+		XeTextLayoutGeometry set;
+		(void)GetTextLayoutGeometry(w, &set,
+					    w->core.width, w->core.height);
+		RedoTextLayout(w, &set, True);
+	    }
 	if ((dolayout && w->basic.resize) || w->core.width == 0)
-		w->core.width =  w->text.w;
+		w->core.width = w->text.w;
 	if ((dolayout && w->basic.resize) || w->core.height == 0)
 		w->core.height = w->text.h;
+	w->core.width = LimitDimension(w->core.width,&w->basic.dimension[0]);
+	w->core.height = LimitDimension(w->core.height,&w->basic.dimension[1]);
 	if (w->core.height < TEXT_MIN_HEIGHT)
 		w->core.height = TEXT_MIN_HEIGHT;
 	if (w->core.width < TEXT_MIN_WIDTH)
@@ -1062,7 +1227,8 @@ Widget request, init;
 	w->text.first = NULL;
 	w->text.colors = NULL;
 	w->text.inserting = NULL;
-	w->text.set_width = -1;	/* Initialize to impossible value */
+	(void)GetTextLayoutGeometry
+		(w, &w->text.set, w->core.width, w->core.height);
 	if (w->text.initial_state)
 		w->text.initial_state = XtNewString(w->text.initial_state);
 	if (w->text.itemization)
@@ -1073,6 +1239,7 @@ Widget request, init;
 	    }
        	w->text.highlight = False;
 	w->text.enable_display = 0;
+	w->text.printing = False;
 	SetupContent(w, True);
     }
 
@@ -1222,8 +1389,8 @@ int x, y;
 	    }
 	if (s == NULL)
 		return;
-	x = s->x;
-	y = s->y;
+	x = s->x + cx->w->basic.x_translation;
+	y = s->y + cx->w->basic.y_translation;
 	switch (s->mode.bits & Underline_MASK)
 	    {
 	    case Underline_DOUBLE:
@@ -1245,8 +1412,8 @@ int x, y;
 		MarkArea(cx, s, MARK_Crossed, x, y, 0);
 	if (s == NULL)
 		return;
-	x = s->x;
-	y = s->y;
+	x = s->x + cx->w->basic.x_translation;
+	y = s->y + cx->w->basic.y_translation;
 	if (s->mode.bits & CrossedOut_MASK)
 		MarkArea(cx, s, MARK_Crossed, x, y, 1);
     }
@@ -1262,8 +1429,8 @@ int x, y;
 		MarkArea(cx, s, MARK_Framed, x, y, 0);
 	if (s == NULL)
 		return;
-	x = s->x;
-	y = s->y;
+	x = s->x + cx->w->basic.x_translation;
+	y = s->y + cx->w->basic.y_translation;
 	if (s->mode.bits & Framing_MASK)
 		MarkArea(cx, s, MARK_Framed, x, y, 1);
     }
@@ -1280,42 +1447,31 @@ int x, y;
 static void GetColors(w)
 XeTextWidget w;
     {
-	int i;
-	XColor exact, color;
-
-	static char *colorname[] =
+	static XeSample rgb[24] =
 	    {
-		"black", "red", "green", "yellow",
-		"blue", "magenta", "cyan", "white",
+		/* black,red,green,yellow,blue,magenta,cyan,white */
+
+		0, ~0,  0, ~0,  0, ~0,  0, ~0,	/* red */
+		0,  0, ~0, ~0,  0,  0, ~0, ~0,	/* green */
+		0,  0,  0,  0, ~0, ~0, ~0, ~0,	/* blue */
 	    };
+	
+	XeSample tmp[24];
 
+	memcpy((void *)tmp, (void *)rgb, sizeof(tmp));
 	w->text.colors = (unsigned long *)XtMalloc(8 * sizeof(unsigned long));
-	for (i = 0; i < 8; i++)
-	    {
-		if (!XAllocNamedColor(XtDisplay(w), w->core.colormap,
-				      colorname[i], &color, &exact))
-		    {
-			XeWidgetWarningMsg((Widget)w,
-					   "colorAllocation",
-					   "Failed allocating color '%s'",
-					   (String *)&colorname[i], 1);
-			color.pixel = FRAME_NO_PIXEL;
-		    }
-		w->text.colors[i] = color.pixel;
-	    }
+	(void)XeAllocColors(XtDisplay(w), XtWindow(w), w->core.colormap,
+			    XeColormapUse_SHARED, 8, 8,
+			    &tmp[0], &tmp[8], &tmp[16],
+			    w->text.colors);
     }
 
 static void FreeColors(w)
 XeTextWidget w;
     {
-	int i;
-
 	if (w->text.colors == NULL)
 		return;
-	for (i = 0; i < 8; i++)
-		if (w->text.colors[i] != FRAME_NO_PIXEL)
-			XFreeColors(XtDisplay(w), w->core.colormap,
-			    &w->text.colors[i], 1, (unsigned long)0);
+	XeFreeColors(XtDisplay(w), w->core.colormap, 8, w->text.colors);
 	XtFree((void *)w->text.colors);
 	w->text.colors = NULL;
     }
@@ -1405,7 +1561,8 @@ Region r;
 #else
 	int (*f_draw16)();
 	int (*f_draw)();
-#endif	
+#endif
+
 	memset((void *)&cx, 0, sizeof(cx));
 	cx.my_r = r;
 	cx.w = t;
@@ -1456,6 +1613,7 @@ Region r;
 		XSetClipMask(XtDisplay(w), t->text.mygcXOR, None);
 		XSetClipMask(XtDisplay(w), t->text.mygcBorder, None);
 	    }
+
 	XSetBackground(XtDisplay(w), t->text.mygc, bg_pixel);
 	if (e == NULL || bg_pixel != t->core.background_pixel)
 	    {
@@ -1478,13 +1636,14 @@ Region r;
 		    }
 		else
 			XFillRectangle(XtDisplay(w),XtWindow(w),t->text.mygc,
-				       0, 0, t->core.width, t->core.height);
+				       t->basic.x_translation,
+				       t->basic.y_translation,
+				       t->core.width, t->core.height);
 	    }
 	XSetForeground(XtDisplay(w), t->text.mygc, t->basic.foreground_pixel);
 	if (t->composite.num_children > 0)
 	    {
 		int i;
-
 		
 		for (i = 0; i < t->composite.num_children; ++i)
 		    {
@@ -1506,8 +1665,8 @@ Region r;
 	    }
 	f_draw16 = XDrawString16;
 	f_draw = XDrawString;
-	y = t->text.y;
-	x = t->text.x;
+	y = t->text.y + t->basic.y_translation;
+	x = t->text.x + t->basic.x_translation;
 	for (s = t->text.first; s; s = s->next)
 	    {
 		SnipData *h = s->content.head;
@@ -1537,15 +1696,16 @@ Region r;
 			    }
 		    }
 		cx.p = s->mode;
-		x = s->x;
-		y = s->y;
+		x = s->x + t->basic.x_translation;
+		y = s->y + t->basic.y_translation;
+
 		if (!s->space && s->xWidth > 0 &&
 		    (cx.my_r == NULL ||
 		     XRectInRegion(cx.my_r, x, y - s->ascent, s->xWidth,
-				  s->ascent + s->descent) != RectangleOut) &&
+				   s->ascent + s->descent) != RectangleOut) &&
 		    s->data != NULL)
 		    {
-			XSetFont(XtDisplay(w), t->text.mygc, h->font->fid);
+                        XSetFont(XtDisplay(w), t->text.mygc, h->font->fid);
 			if (h->bytes == 2)
 				(*f_draw16)
 					(XtDisplay(w), XtWindow(w),
@@ -1558,6 +1718,8 @@ Region r;
 					 t->text.mygc,
 					 (int)x + s->offset, (int)y,
 					 s->data, s->length);
+
+                        
 		    }
 		if (!s->floating)
 			x += s->xWidth;
@@ -1584,6 +1746,230 @@ Region r;
 			       0, 0, t->core.width - 1, t->core.height - 1);
     }
 /*
+**	****************
+**	Printing support
+**	****************
+*/
+static void _ResetSnips(w, pp)
+XeTextWidget w;
+XePrintProcess pp;
+    {
+	Snip *t;
+
+	for (t = w->text.first; t; t= t->next)
+		if (!t->widget && t->content.head)
+			t->content.head->font = NULL;
+    }
+
+static GC MakePrintGC(printer, page, display, gc)
+Display *printer;
+Drawable page;
+Display *display;
+GC gc;
+    {
+	unsigned long value_mask;
+	XGCValues values;
+
+	/*
+	** create a mask that will retrieve/set all members of XGCValues
+	** (except those that are server/display specific, and need to
+	** to be transferred explicitly, if needed!)
+	*/
+	value_mask = (1L << (GCLastBit+1)) - 1;
+	value_mask &= ~(GCFont | GCTile | GCStipple | GCClipMask |
+			GCDashList | GCForeground | GCBackground);
+	
+	XGetGCValues(display, gc, value_mask, &values);
+	return XCreateGC(printer, page, value_mask, &values);
+    }
+
+static void PrintInitialize(w, c, pp)
+Widget w, c;
+XePrintProcess pp;
+    {
+	XeTextWidget clone = (XeTextWidget)c;
+	XeTextWidget t = (XeTextWidget)w;
+	Pixel pixels[XtNumber(clone->text.framed_rendition)];
+	XeSample r[XtNumber(clone->text.framed_rendition)];
+	XeSample g[XtNumber(clone->text.framed_rendition)];
+	XeSample b[XtNumber(clone->text.framed_rendition)];
+	
+	int i, j;
+
+	/*
+	** *NOTE*
+	**	Currently, nothing special is done if print_initialize is
+	**	called while another print is active. The current relayout
+	**	really does not work properly, with simultaenous multiple
+	**	printings.. Waiting for ideas what should be done about
+	**	this, but for now, just let it pass (one may get X protocol
+	**	error due to bad font id, if printers are different, or
+	**	just plain garbled output....) -- msa
+	*/
+	t->text.printing = True;
+	t->text.enable_display -= 1;
+	/*
+	** The right thing would be make a copy of all content to the
+	** clone, but trying for a faster method first: move the content
+	** to the clone widget, and leave original empty for the duration
+	** of the printing.
+	*/
+	t->text.first = NULL;
+	t->text.inserting = NULL;
+	_ResetSnips(clone, pp);
+	clone->text.printing = True;
+	clone->text.mygc = MakePrintGC
+		(XtDisplay(c), XtWindow(c), XtDisplay(w), t->text.mygc);
+	clone->text.mygcXOR = MakePrintGC
+		(XtDisplay(c), XtWindow(c), XtDisplay(w), t->text.mygcXOR);
+	/*
+	** Translate frame_rendition. First, one need to translate the
+	** pixel values, if any..
+	*/
+	for (i = j = 0; i < XtNumber(t->text.framed_rendition); ++i)
+		if (t->text.framed_rendition[i].pixel != FRAME_NO_PIXEL)
+			pixels[j++] = t->text.framed_rendition[i].pixel;
+	if (j > 0)
+	    {
+		XeQueryColors(XtDisplay(w),t->core.colormap, j, pixels,r,g,b);
+		clone->core.colormap = XeAllocColors
+			(XtDisplay(c), XtWindow(c), c->core.colormap,
+			 XeColormapUse_SHARED, j, j, r, g, b, pixels);
+	    }
+	for (i = j = 0; i < XtNumber(t->text.framed_rendition); ++i)
+		if (t->text.framed_rendition[i].pixel != FRAME_NO_PIXEL)
+			clone->text.framed_rendition[i].pixel = pixels[j++];
+	
+	clone->text.mygcBorder = _XeFrameCreateGC
+		((Widget)clone, &clone->text.line,
+		 clone->text.framed_rendition,
+		 XtNumber(clone->text.framed_rendition),
+		 clone->text.shadow_contrast);
+	clone->text.colors = NULL;
+	/*
+	** Do paging of the text content by adjusting some resources on
+	** clone structure (this may take too much control away from the
+	** application, and it might be reconsidered whether this is the
+	** right place to do such adjustments)
+	*/
+	if (!XtParent(clone))
+	    {
+		clone->text.column_height = pp->printer->page_height;
+		(void)GetTextLayoutGeometry
+			(clone, &clone->text.set,
+			 pp->printer->page_width, pp->printer->page_height);
+	    }
+	else
+		(void)GetTextLayoutGeometry(clone, &clone->text.set, 0, 0);
+	if (clone->composite.num_children > 0)
+	    {
+		/*
+		** Ugly and very "brittle" fix to a problem: the layout
+		** below needs to find the preferred geometry of the
+		** insets, but PrintConstraint Initialize methods are
+		** not called yet and the widget pointers in Snips are
+		** not correct. The fix... patch up the widget pointers
+		** here...
+		*/
+		Widget inset;
+		int i;
+		XeTextConstraints tc;
+
+		for (i = 0; i < clone->composite.num_children; i++)
+		    {
+			inset = clone->composite.children[i];
+			if (!inset)
+				continue;
+			tc = (XeTextConstraints)inset->core.constraints;
+			if (tc && tc->text.inset)
+				tc->text.inset->content.widget = inset;
+		    }
+	    }
+	RedoTextLayout(clone, &clone->text.set, True);
+	if (!XtParent(clone))
+	    {
+		clone->core.width = clone->text.w;
+		clone->core.height = clone->text.h;
+	    }
+    }
+
+static XtPointer Print(w, c, x, y, pp)
+Widget w, c;
+int x, y;
+XePrintProcess pp;
+    {
+	XeTextWidget clone = (XeTextWidget)c;
+
+	clone->basic.x_translation = x;
+	clone->basic.y_translation = y;
+	REDISPLAY(c, (XEvent *)NULL, (Region)0);
+	return NULL;
+    }
+
+static void PrintCancel(w, clone, pp, client_data)
+Widget w, clone;
+XePrintProcess pp;
+XtPointer client_data;
+    {
+    }
+
+static void PrintDestroy(w, c, pp)
+Widget w, c;
+XePrintProcess pp;
+    {
+	XeTextWidget t = (XeTextWidget)w;
+	XeTextWidget clone = (XeTextWidget)c;
+
+	XFreeGC(XtDisplay(clone), clone->text.mygc);
+	XFreeGC(XtDisplay(clone), clone->text.mygcXOR);
+	_XeFrameFreeGC(c, clone->text.mygcBorder,
+		       clone->text.framed_rendition,
+		       XtNumber(clone->text.framed_rendition));
+	FreeColors(clone);
+	t->text.printing = False;
+	/*
+	** Restore content back to the original widget. NOTE! Need to
+	** copy back the 'first' from clone to the original widget, because
+	** the original first might have been pointing to a layout generated
+	** content, which has been deleted by now...
+	*/
+	t->text.first = clone->text.first;
+	t->text.inserting = clone->text.inserting;
+	_ResetSnips(t, pp);
+	RedoTextLayout(t, &t->text.set, True);
+	if ((t->text.enable_display += 1) == 0)
+		REDISPLAY(t, NULL, 0);
+    }
+
+static void PrintConstraintInitialize(w, c, pp)
+Widget w, c;
+XePrintProcess pp;
+    {
+	XeTextConstraints tc = (XeTextConstraints)w->core.constraints;
+
+	_XeFrameInitialize(w, c, (ArgList)0, (Cardinal)0);
+	/*
+	** Replace the widget pointer in the Snip
+	*/
+	if (tc && tc->text.inset)
+		tc->text.inset->content.widget = c;
+    }
+
+static void PrintConstraintDestroy(w, c, pp)
+Widget w, c;
+XePrintProcess pp;
+    {
+	XeTextConstraints tc = (XeTextConstraints)c->core.constraints;
+
+	_XeFrameDestroy(c);
+	/*
+	** Restore the original widget pointer into the Snip
+	*/
+	if (tc && tc->text.inset)
+		tc->text.inset->content.widget = w;
+    }
+
+/*
 **	*****************
 **	Actions functions
 **	*****************
@@ -1648,6 +2034,10 @@ Cardinal *num_args;
 		   if column_width is defined */
 		(!nw->text.column_width && cw->core.width != nw->core.width) ||
 		 cw->text.column_width != nw->text.column_width ||
+		 cw->text.columns != nw->text.columns ||
+		 cw->text.column_height != nw->text.column_height ||
+		 cw->text.column_separation != nw->text.column_separation ||
+		 cw->text.page_separation != nw->text.page_separation ||
 		 /* Other parameters affecting layout */
 		 cw->text.alignment != nw->text.alignment ||
 		 cw->text.first_line_offset!=nw->text.first_line_offset ||
@@ -1695,25 +2085,44 @@ Cardinal *num_args;
 		    }
 	    }
 	redisplay |= SetupContent(nw, dolayout);
-	return nw->text.enable_display < 0 ? 0 : redisplay;
+#if 0
+	/* This piece of "optimization" has the problem that the core
+	   width and height may have been changed in SetupContent and
+	   the clearing of the window in Redisplay based on these values
+	   is not usually correct => garbage is left on screen...
+	   The code left here as a reminder of what not to do!!!
+
+	   -- msa */
+
+	if (redisplay && nw->text.enable_display >= 0)
+		/*
+		** To avoid flickering, when widget is used to display
+		** frequently changing text, just do the redisplay here
+		** now!
+		*/
+		REDISPLAY(nw, NULL, 0);
+	return False;
+#else
+	return redisplay && nw->text.enable_display >= 0;
+#endif
     }
 
 static void Resize(w)
 Widget w;
     {
 	XeTextWidget t = (XeTextWidget)w;
-	int width;
+	XeTextLayoutGeometry layout;
 
 	if (t->text.enable_display < 0)
 		return;
 	/*
 	** Ignore full re-layout, if set_width doesn't change.
 	*/
-	width = TEXT_SET_WIDTH(t);
-	if (t->text.set_width != width)
+	if (GetTextLayoutGeometry(w, &layout, w->core.width, w->core.height))
 	    {
-		RedoTextLayout(t, width, True);
-		REDISPLAY(t, NULL, 0);
+		RedoTextLayout(t, &layout, True);
+		if (!t->text.printing)
+			REDISPLAY(t, NULL, 0);
 	    }
     }
 
@@ -1737,7 +2146,7 @@ Cardinal *num_args;
 		*/
 		XtWidgetGeometry current;
 
-		_XeTextLayout(t,TEXT_SET_WIDTH(t),(Snip *)NULL,(Region)NULL);
+		_XeTextLayout(t, &t->text.set, (Snip *)NULL, (Region)NULL);
 		/*
 		** This is icky.. changing core fields here only to get
 		** XtSetValues to call GeometryManager.. which again has
@@ -1889,6 +2298,9 @@ XtWidgetGeometry *request, *reply;
 	XeTextWidget tw = (XeTextWidget)XtParent(w);
 	XtWidgetGeometry rq, current;
 	XtGeometryMask rm = request->request_mode;
+	int width, height;
+	XeTextLayoutGeometry set, old;
+	XtGeometryResult my_result = XtGeometryNo;
 
 	rq.request_mode = rm;
 	rq.x = (rm & CWX) ? request->x : w->core.x;
@@ -1899,9 +2311,24 @@ XtWidgetGeometry *request, *reply;
 	if (rq.x == w->core.x && rq.y == w->core.y &&
 	    rq.width == w->core.width && rq.height == w->core.height)
 		return XtGeometryNo;
-	_XeTextLayout(tw, TEXT_SET_WIDTH(tw), (Snip *)NULL, (Region)NULL);
+	/*
+	** Do some ad hoc heuristics here: if the dimension requests
+	** MAX (== 0), then prefer the current size (should at
+	** least work with XeFrame parent, in which case the core fields
+	** already may include the effect of the MAX contraint
+	*/
+	width = tw->basic.dimension[0].value ? 0 : tw->core.width;
+	height = tw->basic.dimension[1].value ? 0 : tw->core.height;
+
+	(void)GetTextLayoutGeometry(tw, &set,
+			      LimitDimension(width, &tw->basic.dimension[0]),
+			      LimitDimension(height, &tw->basic.dimension[1]));
+	old = tw->text.set;
+	_XeTextLayout(tw, &set, (Snip *)NULL, (Region)NULL);
+	width = LimitDimension(tw->text.w, &tw->basic.dimension[0]);
+	height = LimitDimension(tw->text.h, &tw->basic.dimension[1]);
 	if (tw->basic.resize &&
-	    (tw->text.h != tw->core.height || tw->text.w != tw->core.width))
+	    (height != tw->core.height || width != tw->core.width))
 	    {
 		/*
 		** The new dimensions for parent differ from the current
@@ -1917,11 +2344,14 @@ XtWidgetGeometry *request, *reply;
 		my_req.request_mode = CWWidth | CWHeight;
 		if (XtCWQueryOnly & rm)
 			my_req.request_mode |= XtCWQueryOnly;
-		my_req.width = tw->text.w;
-		my_req.height = tw->text.h;
+		my_req.width = width;
+		my_req.height = height;
 		result = XtMakeGeometryRequest((Widget)tw, &my_req, &my_rep);
 		if (result == XtGeometryNo)
-			return XtGeometryNo;
+		    {
+			my_result = XtGeometryNo;
+			goto restore_layout;
+		    }
 		else if (result == XtGeometryAlmost)
 		    {
 			if ((rm & XtCWQueryOnly) == 0)
@@ -1933,14 +2363,20 @@ XtWidgetGeometry *request, *reply;
 	if (!reply)
 		reply = &current;
 	if (!GetInsetGeometry(tw, w, reply))
-		return XtGeometryNo; /* Probaly being destroyed... */
+	    {
+		my_result = XtGeometryNo; /* Probaly being destroyed... */
+		goto restore_layout;
+	    }
 	if ((!(rm & CWWidth) || rq.width == reply->width) &&
 	    (!(rm & CWHeight) || rq.height == reply->height) &&
 	    (!(rm & CWX) || rq.x == reply->x) &&
 	    (!(rm & CWY) || rq.y == reply->y))
 	    {
 		if (rm & XtCWQueryOnly)
-			return XtGeometryYes;
+		    {
+			my_result = XtGeometryYes;
+			goto restore_layout;
+		    }
 		if (tw->text.enable_display >= 0)
 		    {
 			_XeTextConfigureChildren(tw);
@@ -1949,33 +2385,54 @@ XtWidgetGeometry *request, *reply;
 		    }
 		return XtGeometryDone;
 	    }
-        return reply ? XtGeometryAlmost : XtGeometryNo;
+        my_result = reply ? XtGeometryAlmost : XtGeometryNo;
+    restore_layout:
+	_XeTextLayout(tw, &old, (Snip *)NULL, (Region)NULL);
+	return my_result;
     }
 
 /*
-** Progamming Interface functions. For documentation, see Text.h
+** User callable convenience function support
+** ==========================================
 */
-void XeTextLayout(w)
+
+/*
+** IsTextWidget
+**	Returns True, if the widget is of class xeText and
+**	False otherwise (also, a warning message is generated).
+*/
+static int IsTextWidget(w)
 Widget w;
     {
 	String params[2];
 
 	if (XtIsSubclass(w, xeTextWidgetClass))
+		return True;
+	params[0] = w->core.name;
+	params[1] = w->core.widget_class->core_class.class_name;
+	XeWidgetWarningMsg
+		(w, "bugTextEdNotText",
+		 "Widget %s is class %s, and not xeTextWidget",
+		 (String *)params, (Cardinal)2);
+	return False;
+    }
+
+/*
+** XeTextLayout
+*/
+void XeTextLayout(w)
+Widget w;
+    {
+	if (IsTextWidget(w))
 	    {
 		XeTextWidget t = (XeTextWidget)w;
+		XeTextLayoutGeometry layout;
 
-		RedoTextLayout(t, TEXT_SET_WIDTH(t), True);
+		(void)GetTextLayoutGeometry
+			(w, &layout, t->core.width, t->core.height);
+		RedoTextLayout(t, &layout, True);
 		REDISPLAY(w, NULL, 0);
 	    }
-	else
-	    {
-		params[0] = w->core.name;
-		params[1] = w->core.widget_class->core_class.class_name;
-		XeWidgetWarningMsg
-			(w, "bugTextNotText",
-			 "Widget %s is class %s, and not xeTextWidget",
-			 (String *)params, (Cardinal)2);
-	    }
     }
 
 
