@@ -1,16 +1,18 @@
-#ifndef _TextFont_h
-#define _TextFont_h
+#ifndef _XewTextFont_h
+#define _XewTextFont_h
 /*
-** Copyright 1992 by Markku Savela and
+** Copyright 1992-1995 by Markku Savela and
 **	Technical Research Centre of Finland
 */
-#include <X11/Xew/TextP.h>
+#include <X11/Xew/Basic.h>
+
+typedef struct XeFontListRec *XeFontList; /* Defined in TextFont.c */
 
 _XFUNCPROTOBEGIN
 XFontStruct *XeFindFont
 	(
 #if NeedFunctionPrototypes
-	 XeTextWidget,	/* Associated Text Widget */
+	 Display *,	/* Display */
 	 char *,	/* Pointer to constant character set string */
 	 XeFontList,	/* FontList pointer or NULL */
 	 int,		/* Weigth (_NORMAL, _FAINT, _BOLD) */
