@@ -1647,7 +1647,11 @@ int n, *p, F, I;
 			align = XeAlignment_JUSTIFIED;
 			break;
 		    }
-		t->quad = align;
+		if (t->quad != align)
+		    {
+			t->quad = align;
+			t->valid = False;
+		    }
 		break;
 	    case SLS:
 		/* Fix later --msa */
@@ -2038,7 +2042,11 @@ Snip *_XeTextInsertWidget(cx, w)
 XeTextInsertContext cx;		/* Context Pointer */
 Widget w;			/* Widget to insert */
     {
-	Snip *t = AppendSnip(cx);
+	Snip *t;
+
+	if (IgnoreFormattedContent(cx))
+		return NULL;	/* Ignore Formatted content for now */
+	t = AppendSnip(cx);
 	if (t != NULL)
 	    {
 		t->widget = True;
@@ -2154,6 +2162,8 @@ int x, y, width, height;
     {
 	if (width == 0 || height == 0)
 		return;
+	x += cx->w->basic.x_translation;
+	y += cx->w->basic.y_translation;
 	if (cx->expose.width == 0)
 	    {
 		cx->expose.x = x;
@@ -2353,7 +2363,7 @@ int mode;
 /*
 ** SplitSnip
 **	Split Snip t into two at offset, and return the pointer to the
-**	*NEW* snip, which represents the fisrt part of the split.
+**	*NEW* snip, which represents the first part of the split.
 **
 **	*** No checks is made. Function assumes the offset <= length.
 */
@@ -2565,6 +2575,135 @@ long amount;
 	return t;
     }
 
+static void CopyLockedModes(cx, t)
+Context *cx;
+Snip *t;
+    {
+	unsigned long old_font, new_font;
+
+	old_font = FontSelectInfo(t->mode);
+
+	t->mode.bits &= ~cx->lock.bits;
+	t->mode.bits |= cx->lock.bits & cx->mode.bits;
+	if (cx->lock.tag)
+		t->mode.tag = cx->mode.tag;
+	if (cx->lock.size_modification)
+		t->mode.size_modification = cx->mode.size_modification;
+	new_font = FontSelectInfo(t->mode);
+	/*
+	** If mode change affected the font selection, then we must
+	** move the snip content into a matching SnipData chunk.
+	** (argh.. this mode change function proves to be nastier than
+	** I expected... --msa )
+	*/
+	if (!t->widget && old_font != new_font && t->content.head)
+	    {
+		int n = t->content.head->bytes * t->length;
+
+		if (cx->used + n > cx->size ||
+		    cx->head->character_set != t->content.head->character_set||
+		    new_font != cx->fontinfo)
+		    {
+			FlushAppend(cx);
+			if (cx->size < n)
+			    {
+				/* Need to reallocate larger head block */
+				DerefDataHeader(cx->head);
+				cx->size = n>HEAD_MIN_SIZE ? n : HEAD_MIN_SIZE;
+				cx->head = (SnipData *)
+					XtCalloc(1, XtOffsetOf(SnipData, s[0])
+						 + cx->size);
+				cx->head->refs = 1; /* cx ref is counted ! */
+			    }	
+			cx->fontinfo = new_font;
+			cx->head->bytes = t->content.head->bytes;
+			cx->head->font = NULL;
+			cx->head->character_set=t->content.head->character_set;
+			cx->first = t;
+		    }
+		memcpy(&cx->head->s[cx->used], t->data, n);
+		cx->head->refs += 1;
+		DerefDataHeader(t->content.head);
+		t->content.head = cx->head;
+		t->data = &cx->head->s[cx->used];
+		cx->used += n;
+	    }
+	t->valid = FALSE;	/* This snip requires new layout */
+    }
+
+/*
+** _XeTextChangeMode
+**	Change the mode of the next 'amount' characters. Only the 'locked'
+**	modes defined in the current insert context will be changed.
+**
+**	amount > 0, change forward from current point,
+**	amount < 0, change backward from current point
+*/
+void _XeTextChangeMode(context, amount)
+XeTextInsertContext context;
+long amount;
+    {
+	Snip *t;
+	Snip **h;
+	int offset, vlength;
+	Context *cx = (Context *)context;
+
+	if (cx == NULL)
+		return;
+	h = cx->last ? &cx->last->next : cx->list;
+	if (amount > 0)
+	    {
+		/*
+		** Change forward
+		*/
+		for (t = *h; t != NULL; t = t->next)
+			if (IsEditableContent(t->mode.bits))
+			    {
+				amount -= (vlength = VirtualLength(t));
+				if (amount < 0)
+				    {
+					offset = amount + vlength;
+					if (offset > 0 && offset < vlength)
+					    {
+						t = SplitSnip(cx, t, offset);
+						CopyLockedModes(cx, t);
+					    }
+					break;
+				    }
+				CopyLockedModes(cx, t);
+			    }
+	    }
+	else if (amount < 0 && *h != NULL)
+	    {
+		/*
+		** Change backward
+		*/
+		for (t = *h; ; t = PreviousSnip(t))
+		    {
+			if (IsEditableContent(t->mode.bits))
+			    {
+				vlength = VirtualLength(t);
+				amount += vlength;
+				if (amount == 0)
+					break;
+				else if (amount > 0)
+				    {
+					if (amount < vlength)
+					    {
+						(void)SplitSnip(cx, t, amount);
+						CopyLockedModes(cx, t);
+					    }
+					break;
+				    }
+				CopyLockedModes(cx, t);
+			    }
+			else if (t->back == cx->list)
+				break;
+		    }
+	    }
+    }
+
+
 /*
 ** _XeTextInsertLocation
 **	Return the TextLocation corresponding the current insert
@@ -2653,6 +2792,8 @@ XeTextTag tag;
     {
 	XeTextTag old = cx->mode.tag;
 	cx->mode.tag = tag;
+	if (!cx->locked)
+		cx->lock.tag = (XeTextTag)~0;
 	return old;
     }
 /*
