@@ -8,13 +8,15 @@
 #define TextLayout_h
 
 #include <X11/Xew/TextP.h>
+
+
 _XFUNCPROTOBEGIN
 int _XeTextLayout(
 #if NeedFunctionPrototypes
-		 XeTextWidget, int, Snip *, Region
+		 XeTextWidget, XeTextLayoutGeometry *, Snip *, Region
 #endif
 );
-Dimension XeTextSnipWidth
+int XeTextSnipWidth
 	(
 #if NeedFunctionPrototypes
 	 XeTextWidget, Snip *, char *, int
