@@ -23,10 +23,12 @@ static char *fallback_resources[] =
 	"*FrameSeparation:1",
 	"*FrameOffset:1",
 	"*FrameDimension:rule-b",
-	"*Panel.middle.FrameBorder:invisible",
-	"*Panel.middle.pane.vertical:max",
-	"*Panel.top.horizontal:max",
-	"*Panel.bottom.horizontal:max",
+	"*panel.middle.FrameBorder:invisible",
+	"*panel.middle.pane.vertical:max",
+	"*panel.top.horizontal:max",
+	"*panel.bottom.horizontal:max",
+	"*panel.middle.layoutPath:0",
+	"*panel.middle.menuPanel.vertical:max",
 	"*labelBox.FrameBorder:invisible",
 	"*inputBox.FrameBorder:invisible",
 	"*inputBox.horizontal:max",
@@ -40,3 +42,7 @@ static char *fallback_resources[] =
     };
 
 
+
+
+
+
