@@ -5,8 +5,10 @@
 
 #include <stdio.h>
 #include <ctype.h>
+#include <string.h>
 #if SYSV_INCLUDES
 #	include <memory.h>
+#	include <malloc.h>
 #endif
 #include <string.h>
 #include <X11/Intrinsic.h>
@@ -450,7 +452,7 @@ Widget w;
 char *s, *e;
     {
 	TagDef *t;
-	int endtag, i;
+	int endtag;
 
 	*e = 0;
 	if (*s == '/')
@@ -662,15 +664,14 @@ int start, end;
 	    }
     }
 
-static int SaveContentSegment(s, n, tag, client_data)
+static int SaveContentSegment(s, len, tag, client_data)
 char *s;
-int n;
-int tag;
+int len;
+XeTextTag tag;
 XtPointer client_data;
     {
 	SaveState *save = (SaveState *)client_data;
-	int t, p;
-	char *q;
+	int t, p, n = len;
 	/*
 	** Locate the first common ancestor of the current tag
 	** (save->tag) and the new tag. When found, generate
@@ -678,7 +679,7 @@ XtPointer client_data;
 	** then generate opening tags up to the new tag.
 	*/
 	for (t = save->tag; t; t = tags[t].down)
-		for (p = tag; p; p = tags[p].down)
+		for (p = (int)tag; p; p = tags[p].down)
 			if (t == p)
 				goto found;
 	t = 0;
@@ -686,8 +687,8 @@ XtPointer client_data;
 	/* t is the common ancestor (or ZERO) */
 	for (p = save->tag; p != t; p = tags[p].down)
 		WriteTag(save, tags[p].tag, 1);
-	OpenTags(save, t, tag); /* ..generate in reverse order */
-	save->tag = tag;
+	OpenTags(save, t, (int)tag); /* ..generate in reverse order */
+	save->tag = (int)tag;
 	while (n > 0)
 	    {
 		while ((save->ptr - save->line) >= (LINE_WIDTH-1))
@@ -702,6 +703,7 @@ XtPointer client_data;
 		    }
 		--n;
 	    }
+	return len;
     }
 
 /*
@@ -739,7 +741,7 @@ XtPointer client_data, call_data;
     }
 
 static Position AdjustCoord(x, y, a, b)
-int x, a, b;
+int x, y, a, b;
     {
 	if (x < a)
 		a = x;
@@ -783,7 +785,7 @@ XtPointer client_data, call_data;
 #endif
 		break;
 	    default:
-		printf("UNKNOWN: XeRC=%d\n");
+		printf("UNKNOWN: XeRC=%d\n", (int)notify->reason);
 		break;
 	    }
     }
@@ -791,10 +793,11 @@ XtPointer client_data, call_data;
 extern UI_PanelDef popup_dialog_saveas;
 extern UI_PanelDef popup_dialog_loadfile;
 
+static char input_name[] = "inputBox";
 
 static UI_ElementDef file_name_input[] =
     {
-	UI_INPUT("Name: "),
+	UI_INPUT(input_name, "Name: ", NULL),
 	UI_END(),
     };
 
@@ -1022,7 +1025,8 @@ typedef struct ChangeState
 
 static int ChangeContentSegment(s, n, tag, client_data)
 char *s;
-int n, tag;
+int n;
+XeTextTag tag;
 XtPointer client_data;
     {
 	ChangeState *change = (ChangeState *)client_data;
@@ -1061,14 +1065,14 @@ XtPointer client_data;
 		change->last = new;
 		change->seg_ptr = &new->data[size];
 		change->data_ptr = (char *)&new->data[0];
-		change->seg_ptr->tag = tag;
+		change->seg_ptr->tag = (int)tag;
 		change->seg_ptr->length = 0;
 		change->available = size;
 	    }
-	if (change->seg_ptr->tag != tag)
+	if (change->seg_ptr->tag != (int)tag)
 	    {
 		change->seg_ptr -= 1;
-		change->seg_ptr->tag = tag;
+		change->seg_ptr->tag = (int)tag;
 		change->seg_ptr->length = 0;
 	    }
 	memcpy((void *)change->data_ptr, (void *)s, n);
@@ -1079,19 +1083,20 @@ XtPointer client_data;
 	/*
 	** Maintain the common ancestor
 	*/
-	if (change->ancestor < 0 || tag == 0)
-		change->ancestor = tag;
-	else for (i = tag; change->ancestor;)
+	if (change->ancestor < 0 || (int)tag == 0)
+		change->ancestor = (int)tag;
+	else for (i = (int)tag; change->ancestor;)
 	    {
 		if (i == 0)
 		    {
-			i = tag;
+			i = (int)tag;
 			change->ancestor = tags[change->ancestor].down;
 		    }
 		else if (i == change->ancestor)
 			break;	/* Common found */
 		i = tags[i].down;
 	    }
+	return n;
     }
 
 static int MakeTags(base, start, end)
@@ -1443,7 +1448,7 @@ char *name;
 	    }
 	while (fgets(buf, sizeof(buf), fp) != NULL)
 	    {
-		int f, c;
+		int f;
 		char *p;
 
 		int n = strlen(buf);
@@ -1461,16 +1466,16 @@ char *name;
 
 #include "fallback.c"
 
-main(argc, argv)
+int main(argc, argv)
 int argc; char *argv[];
     {
 	XtAppContext demo_application;
-	Widget root_widget, view, widget, pane, box, button;
+	Widget root_widget, view, pane, box;
 	Arg args[10];
 	int n;
 	char *tags_def;
 
-	root_widget = XtAppInitialize(&demo_application, "Enriched",
+	root_widget = XtAppInitialize(&demo_application, "Demo",
 				      NULL,0, &argc, argv,
 				      fallback_resources,
 				      NULL, 0);
