@@ -7,7 +7,8 @@
 **	TextEd features individually at low level.
 */
 #include <stdio.h>
-#include <sys/file.h>
+#include <stdlib.h> /* is this available always? --msa */
+#include <fcntl.h>
 #include <sys/types.h> /* Some <sys/stat.h>'s need this, not all */
 #include <sys/stat.h>
 #if SYSV_INCLUDES
@@ -27,6 +28,7 @@
 #include <X11/Xew/Frame.h>
 #include <X11/Xew/Support.h>
 #include <X11/Xmu/Atoms.h>	/* needed for cut/paste export/import */
+#include "printing.h"
 
 #if SYSV_INCLUDES || ANSI_INCLUDES
 #	include <limits.h>
@@ -50,7 +52,7 @@ static Widget edit_title;	/* Label Widget for filename */
 static int text_modified;	/* True if text modified */
 
 static Position AdjustCoord(x, y, a, b)
-int x, a, b;
+int x, y, a, b;
     {
 	if (x < a)
 		a = x;
@@ -199,6 +201,12 @@ static void SetFilenameToTitle()
 **	insert a widget at current insert point in the text stream.
 */
 
+static char command_name[] = "commandButton";
+static char panel_name[] = "menuPanel";
+static char toggle_name[] = "toggleBox";
+static char input_name[] = "inputBox";
+static char label_name[] = "labelBox";
+
 #include "image_panel.c"
 #include "video_panel.c"
 #include "text_panel.c"
@@ -206,93 +214,107 @@ static void SetFilenameToTitle()
 extern UI_PanelDef popup_paragraph_control;
 extern UI_PanelDef popup_insert_state;
 
-
 static UI_ElementDef insert_state[] =
     {
 	/* *NOTE* The control sequences in data assume the C compiler is
 	   is using ASCII code set. --msa */
-	UI_PANE(	" "),
-	UI_TOGGLE(	"Normal intensity",	"\033[22m"),
-	UI_RADIO(	"Bold",			"\033[1m"),
-	UI_TOGGLE(	"Italic",		"\033[3m"),
-	UI_RADIO(	"Not italic",		"\033[23m"),
-	UI_TOGGLE(	"Underlined",		"\033[4m"),
-	UI_RADIO(	"Double underlined",	"\033[21m"),
-	UI_RADIO(	"Not underlined",	"\033[24m"),
-	UI_TOGGLE(	"Inverse", 		"\033[7m"),
-	UI_RADIO(	"Not inversed",		"\033[27m"),
-	UI_TOGGLE(	"Crossed out", 		"\033[9m"),
-	UI_RADIO(	"Not crossed out", 	"\033[29m"),
-	UI_TOGGLE(	"Framed", 		"\033[51m"),
-	UI_RADIO(	"Not framed",	 	"\033[54m"),
-	UI_TOGGLE(	"Superscript",		"\033L\033L"),
-	UI_RADIO(	"Baseline",		"\033K\033K\033L"),
-	UI_RADIO(	"Subscript",		"\033K\033K"),
-	UI_TOGGLE(	"Default aligned",	"\033[0 F"),
-	UI_RADIO(	"Justified",		"\033[2 F"),
-	UI_RADIO(	"Start aligned",	"\033[5 F"),
-	UI_RADIO(	"Centered",		"\033[6 F"),
-	UI_RADIO(	"End aligned",		"\033[7 F"),
-	UI_PANE(	"Foreground:"),
-	UI_TOGGLE(	"Default",		"\033[39m"),
-	UI_RADIO(	"Black",		"\033[30m"),
-	UI_RADIO(	"Red",			"\033[31m"),
-	UI_RADIO(	"Green",		"\033[32m"),
-	UI_RADIO(	"Yellow",		"\033[33m"),
-	UI_RADIO(	"Blue",			"\033[34m"),
-	UI_RADIO(	"Magenta",		"\033[35m"),
-	UI_RADIO(	"Cyan",			"\033[36m"),
-	UI_RADIO(	"White",		"\033[37m"),
-	UI_LABEL(	"Background:"),
-	UI_TOGGLE(	"Default",		"\033[49m"),
-	UI_RADIO(	"Black",		"\033[40m"),
-	UI_RADIO(	"Red",			"\033[41m"),
-	UI_RADIO(	"Green",		"\033[42m"),
-	UI_RADIO(	"Yellow",		"\033[43m"),
-	UI_RADIO(	"Blue",			"\033[44m"),
-	UI_RADIO(	"Magenta",		"\033[45m"),
-	UI_RADIO(	"Cyan",			"\033[46m"),
-	UI_RADIO(	"White",		"\033[47m"),
-	UI_PANE(	"Font:"),
-	UI_TOGGLE(	"primary (default)",	"\033[10m"),
-	UI_RADIO(	"First alternative",	"\033[11m"),
-	UI_RADIO(	"Second alternative",	"\033[12m"),
-	UI_RADIO(	"Third alternative",	"\033[13m"),
-	UI_RADIO(	"Fourth alternative",	"\033[14m"),
-	UI_RADIO(	"Fifth alternative",	"\033[15m"),
-	UI_RADIO(	"Sixth alternative",	"\033[16m"),
-	UI_RADIO(	"Seventh alternative",	"\033[17m"),
-	UI_RADIO(	"Eight alternative",	"\033[18m"),
-	UI_RADIO(	"Ninth alternative",	"\033[19m"),
-	UI_LABEL(	"Size modification:"),
-	UI_TOGGLE(	" 25%",	"\033[25 B"),
-	UI_RADIO(	" 50%",	"\033[50 B"),
-	UI_RADIO(	" 75%",	"\033[75 B"),
-	UI_RADIO(	"100%",	"\033[100 B"),
-	UI_RADIO(	"125%",	"\033[125 B"),
-	UI_RADIO(	"150%",	"\033[150 B"),
-	UI_RADIO(	"175%",	"\033[175 B"),
-	UI_RADIO(	"200%",	"\033[200 B"),
-	UI_RADIO(	"225%",	"\033[225 B"),
-	UI_RADIO(	"250%",	"\033[250 B"),
-	UI_RADIO(	"275%",	"\033[275 B"),
-	UI_RADIO(	"300%",	"\033[300 B"),
-	UI_PANE(	"Character set:"),
-	UI_TOGGLE("iso8859-1 (Latin No.1)","\33(B\33)B\33.A\33/A\17\33\175"),
-	UI_RADIO("iso8859-2 (Latin No.2)","\33(B\33)B\33.B\33/B\17\33\175"),
-	UI_RADIO("iso8859-3 (Latin No.3)","\33(B\33)B\33.C\33/C\17\33\175"),
-	UI_RADIO("iso8859-4 (Latin No.4)","\33(B\33)B\33.D\33/D\17\33\175"),
-	UI_RADIO("iso8859-5 (Latin/Cyrillic)","\33(B\33)B\33.L\33/L\17\33\175"),
-	UI_RADIO("iso8859-6 (Latin/Arabic)","\33(B\33)B\33.G\33/G\17\33\175"),
-	UI_RADIO("iso8859-7 (Latin/Greek)","\33(B\33)B\33.F\33/F\17\33\175"),
-	UI_RADIO("iso8859-8 (Latin/Hebrew)","\33(B\33)B\33.H\33/H\17\33\175"),
-	UI_RADIO("iso8859-9 (Latin No.5)","\33(B\33)B\33.M\33/M\17\33\175"),
-	UI_RADIO("iso6937","\33(B\33)B\33.R\33/R\17\33\175"),
-	UI_RADIO("iso646-se",		"\33(G\33)G\33.A\33/A\17\33\175"),
-	UI_RADIO("iso646-se (names)",	"\33(H\33)H\33.A\33/A\17\33\175"),
-	UI_RADIO("EUC", "\33(J\33$)B\33*I\33$+C\17\33\176"),
-	UI_RADIO("KSC 5601-1987",	"\33(B\33)B\33$*C\33$+C\17\33\175"),
-	UI_RADIO("GB 2312-80",		"\33(B\33)B\33$*A\33$+A\17\33\175"),
+	UI_PANE(panel_name,	NULL),
+	UI_TOGGLE(toggle_name,	"Normal intensity",	"\033[22m", NULL),
+	UI_RADIO(toggle_name,	"Bold",			"\033[1m", NULL),
+	UI_TOGGLE(toggle_name,	"Italic",		"\033[3m", NULL),
+	UI_RADIO(toggle_name,	"Not italic",		"\033[23m", NULL),
+	UI_TOGGLE(toggle_name,	"Underlined",		"\033[4m", NULL),
+	UI_RADIO(toggle_name,	"Double underlined",	"\033[21m", NULL),
+	UI_RADIO(toggle_name,	"Not underlined",	"\033[24m", NULL),
+	UI_TOGGLE(toggle_name,	"Inverse", 		"\033[7m", NULL),
+	UI_RADIO(toggle_name,	"Not inversed",		"\033[27m", NULL),
+	UI_TOGGLE(toggle_name,	"Crossed out", 		"\033[9m", NULL),
+	UI_RADIO(toggle_name,	"Not crossed out", 	"\033[29m", NULL),
+	UI_TOGGLE(toggle_name,	"Framed", 		"\033[51m", NULL),
+	UI_RADIO(toggle_name,	"Not framed",	 	"\033[54m", NULL),
+	UI_TOGGLE(toggle_name,	"Superscript",		"\033L\033L", NULL),
+	UI_RADIO(toggle_name,	"Baseline",	"\033K\033K\033L", NULL),
+	UI_RADIO(toggle_name,	"Subscript",		"\033K\033K", NULL),
+	UI_TOGGLE(toggle_name,	"Default aligned",	"\033[0 F", NULL),
+	UI_RADIO(toggle_name,	"Justified",		"\033[2 F", NULL),
+	UI_RADIO(toggle_name,	"Start aligned",	"\033[5 F", NULL),
+	UI_RADIO(toggle_name,	"Centered",		"\033[6 F", NULL),
+	UI_RADIO(toggle_name,	"End aligned",		"\033[7 F", NULL),
+	UI_PANE(panel_name,	"Foreground:"),
+	UI_TOGGLE(toggle_name,	"Default",		"\033[39m", NULL),
+	UI_RADIO(toggle_name,	"Black",		"\033[30m", NULL),
+	UI_RADIO(toggle_name,	"Red",			"\033[31m", NULL),
+	UI_RADIO(toggle_name,	"Green",		"\033[32m", NULL),
+	UI_RADIO(toggle_name,	"Yellow",		"\033[33m", NULL),
+	UI_RADIO(toggle_name,	"Blue",			"\033[34m", NULL),
+	UI_RADIO(toggle_name,	"Magenta",		"\033[35m", NULL),
+	UI_RADIO(toggle_name,	"Cyan",			"\033[36m", NULL),
+	UI_RADIO(toggle_name,	"White",		"\033[37m", NULL),
+	UI_LABEL(label_name,	"Background:"),
+	UI_TOGGLE(toggle_name,	"Default",		"\033[49m", NULL),
+	UI_RADIO(toggle_name,	"Black",		"\033[40m", NULL),
+	UI_RADIO(toggle_name,	"Red",			"\033[41m", NULL),
+	UI_RADIO(toggle_name,	"Green",		"\033[42m", NULL),
+	UI_RADIO(toggle_name,	"Yellow",		"\033[43m", NULL),
+	UI_RADIO(toggle_name,	"Blue",			"\033[44m", NULL),
+	UI_RADIO(toggle_name,	"Magenta",		"\033[45m", NULL),
+	UI_RADIO(toggle_name,	"Cyan",			"\033[46m", NULL),
+	UI_RADIO(toggle_name,	"White",		"\033[47m", NULL),
+	UI_PANE(panel_name,	"Font:"),
+	UI_TOGGLE(toggle_name,	"primary (default)",	"\033[10m", NULL),
+	UI_RADIO(toggle_name,	"First alternative",	"\033[11m", NULL),
+	UI_RADIO(toggle_name,	"Second alternative",	"\033[12m", NULL),
+	UI_RADIO(toggle_name,	"Third alternative",	"\033[13m", NULL),
+	UI_RADIO(toggle_name,	"Fourth alternative",	"\033[14m", NULL),
+	UI_RADIO(toggle_name,	"Fifth alternative",	"\033[15m", NULL),
+	UI_RADIO(toggle_name,	"Sixth alternative",	"\033[16m", NULL),
+	UI_RADIO(toggle_name,	"Seventh alternative",	"\033[17m", NULL),
+	UI_RADIO(toggle_name,	"Eight alternative",	"\033[18m", NULL),
+	UI_RADIO(toggle_name,	"Ninth alternative",	"\033[19m", NULL),
+	UI_LABEL(label_name,	"Size modification:"),
+	UI_TOGGLE(toggle_name,	" 25%",	"\033[25 B", NULL),
+	UI_RADIO(toggle_name,	" 50%",	"\033[50 B", NULL),
+	UI_RADIO(toggle_name,	" 75%",	"\033[75 B", NULL),
+	UI_RADIO(toggle_name,	"100%",	"\033[100 B", NULL),
+	UI_RADIO(toggle_name,	"125%",	"\033[125 B", NULL),
+	UI_RADIO(toggle_name,	"150%",	"\033[150 B", NULL),
+	UI_RADIO(toggle_name,	"175%",	"\033[175 B", NULL),
+	UI_RADIO(toggle_name,	"200%",	"\033[200 B", NULL),
+	UI_RADIO(toggle_name,	"225%",	"\033[225 B", NULL),
+	UI_RADIO(toggle_name,	"250%",	"\033[250 B", NULL),
+	UI_RADIO(toggle_name,	"275%",	"\033[275 B", NULL),
+	UI_RADIO(toggle_name,	"300%",	"\033[300 B", NULL),
+	UI_PANE(panel_name,	"Character set:"),
+	UI_TOGGLE(toggle_name, "iso8859-1 (Latin No.1)",
+		  "\33(B\33)B\33.A\33/A\17\33\175", NULL),
+	UI_RADIO(toggle_name, "iso8859-2 (Latin No.2)",
+		 "\33(B\33)B\33.B\33/B\17\33\175", NULL),
+	UI_RADIO(toggle_name, "iso8859-3 (Latin No.3)",
+		 "\33(B\33)B\33.C\33/C\17\33\175", NULL),
+	UI_RADIO(toggle_name, "iso8859-4 (Latin No.4)",
+		 "\33(B\33)B\33.D\33/D\17\33\175", NULL),
+	UI_RADIO(toggle_name, "iso8859-5 (Latin/Cyrillic)",
+		 "\33(B\33)B\33.L\33/L\17\33\175", NULL),
+	UI_RADIO(toggle_name, "iso8859-6 (Latin/Arabic)",
+		 "\33(B\33)B\33.G\33/G\17\33\175", NULL),
+	UI_RADIO(toggle_name, "iso8859-7 (Latin/Greek)",
+		 "\33(B\33)B\33.F\33/F\17\33\175", NULL),
+	UI_RADIO(toggle_name, "iso8859-8 (Latin/Hebrew)",
+		 "\33(B\33)B\33.H\33/H\17\33\175", NULL),
+	UI_RADIO(toggle_name, "iso8859-9 (Latin No.5)",
+		 "\33(B\33)B\33.M\33/M\17\33\175", NULL),
+	UI_RADIO(toggle_name, "iso6937",
+		 "\33(B\33)B\33.R\33/R\17\33\175", NULL),
+	UI_RADIO(toggle_name, "iso646-se",
+		 "\33(G\33)G\33.A\33/A\17\33\175", NULL),
+	UI_RADIO(toggle_name, "iso646-se (names)",
+		 "\33(H\33)H\33.A\33/A\17\33\175", NULL),
+	UI_RADIO(toggle_name, "EUC",
+		 "\33(J\33$)B\33*I\33$+C\17\33\176", NULL),
+	UI_RADIO(toggle_name, "KSC 5601-1987",
+		 "\33(B\33)B\33$*C\33$+C\17\33\175", NULL),
+	UI_RADIO(toggle_name, "GB 2312-80",
+		 "\33(B\33)B\33$*A\33$+A\17\33\175", NULL),
 	UI_END(),
     };
 
@@ -301,6 +323,8 @@ static UI_ActionDef insert_state_actions[] =
     {
 	{"set", SetInsertState, (XtPointer)&popup_insert_state},
 	{"close"},
+	{"print", PrintToDisplay, (XtPointer)&popup_insert_state.shell},
+
 	{NULL},
     };
 
@@ -335,39 +359,39 @@ char *name;
 	*/
 	static UI_ElementDef inset_list[] =
 	    {
-		UI_PANE("Frame Position"),
-		UI_INPUT("positioning:"),
+		UI_PANE(panel_name, "Frame Position"),
+		UI_INPUT(input_name, "positioning:", UI_AcceptPanel),
 		UI_LOAD(XtNframePosition),
-		UI_INPUT("alignment:"),
+		UI_INPUT(input_name, "alignment:", UI_AcceptPanel),
 		UI_LOAD(XtNframeAlignment),
 		
-		UI_LABEL("Frame Dimension"),
-		UI_INPUT("Horizontal: "),
+		UI_LABEL(label_name, "Frame Dimension"),
+		UI_INPUT(input_name, "Horizontal: ", UI_AcceptPanel),
 		UI_LOAD(XtNhorizontal),
-		UI_INPUT("Vertical: "),
+		UI_INPUT(input_name, "Vertical: ", UI_AcceptPanel),
 		UI_LOAD(XtNvertical),
 		
-		UI_LABEL("Frame separation"),
-		UI_INPUT("trailing:"),
+		UI_LABEL(label_name, "Frame separation"),
+		UI_INPUT(input_name, "trailing:", UI_AcceptPanel),
 		UI_LOAD(XtNtrailingSeparation),
-		UI_INPUT("leading:"),
+		UI_INPUT(input_name, "leading:", UI_AcceptPanel),
 		UI_LOAD(XtNleadingSeparation),
-		UI_INPUT("left:"),
+		UI_INPUT(input_name, "left:", UI_AcceptPanel),
 		UI_LOAD(XtNleftSeparation),
-		UI_INPUT("right:"),
+		UI_INPUT(input_name, "right:", UI_AcceptPanel),
 		UI_LOAD(XtNrightSeparation),
 		
-		UI_PANE("Frame borders"),
-		UI_INPUT("left: "),
+		UI_PANE(panel_name, "Frame borders"),
+		UI_INPUT(input_name, "left: ", UI_AcceptPanel),
 		UI_LOAD(XtNleftBorder),
-		UI_INPUT("right: "),
+		UI_INPUT(input_name, "right: ", UI_AcceptPanel),
 		UI_LOAD(XtNrightBorder),
-		UI_INPUT("trailing: "),
+		UI_INPUT(input_name, "trailing: ", UI_AcceptPanel),
 		UI_LOAD(XtNtrailingBorder),
-		UI_INPUT("leading: "),
+		UI_INPUT(input_name, "leading: ", UI_AcceptPanel),
 		UI_LOAD(XtNleadingBorder),
 
-		UI_TOGGLE("invert shadow",	True),
+		UI_TOGGLE(toggle_name, "invert shadow",	True, NULL),
 		UI_LOAD(XtNinvertShadow),
 		UI_END(),
 	    };
@@ -489,7 +513,7 @@ XtPointer client_data;
 		*/
 		Widget inset = (Widget)data;
 		if (!inset)
-			return;
+			return 0;
 		length = strlen(XtName(inset)) + 2;
 		data = XtName(inset);
 		wrap = True;
@@ -606,7 +630,7 @@ XtPointer client_data, call_data;
     {
 	XeImportCallbackData *data = (XeImportCallbackData *)call_data;
 	Display *d = XtDisplay(w);
-	char *buf, *mark;
+	char *buf;
 	int length;
 
 	if (data == NULL || data->reason != XeCR_IMPORT)
@@ -643,26 +667,26 @@ extern UI_PanelDef popup_dialog_saveas;
 extern UI_PanelDef popup_dialog_loadfile;
 extern UI_PanelDef popup_dialog_insetfile;
 
+static void PopdownFileDialog();
+
 static UI_ElementDef file_save_input[] =
     {
-	UI_INPUT("Name: "),
+	UI_INPUT(input_name, "Name: ", PopdownFileDialog),
 	UI_END(),
     };
 
 static UI_ElementDef file_load_input[] =
     {
-	UI_INPUT("Name: "),
+	UI_INPUT(input_name, "Name: ", PopdownFileDialog),
 	UI_END(),
     };
 
 static UI_ElementDef file_inset_input[] =
     {
-	UI_INPUT("Name: "),
+	UI_INPUT(input_name, "Name: ", PopdownFileDialog),
 	UI_END(),
     };
 
-static void PopdownFileDialog();
-
 static UI_ActionDef saveas_actions[] =
     {
 	{"save", PopdownFileDialog, (XtPointer)&popup_dialog_saveas},
@@ -684,7 +708,7 @@ static UI_ActionDef insetfile_actions[] =
 	{NULL},
     };
 
-static UI_PanelDef popup_dialog_saveas =
+UI_PanelDef popup_dialog_saveas =
     {
 	"dialogPanel",
 	file_save_input,
@@ -692,7 +716,7 @@ static UI_PanelDef popup_dialog_saveas =
 	"Save File",
     };
 
-static UI_PanelDef popup_dialog_loadfile =
+UI_PanelDef popup_dialog_loadfile =
     {
 	"dialogPanel",
 	file_load_input,
@@ -700,7 +724,7 @@ static UI_PanelDef popup_dialog_loadfile =
 	"Load File",
     };
 
-static UI_PanelDef popup_dialog_insetfile =
+UI_PanelDef popup_dialog_insetfile =
     {
 	"dialogPanel",
 	file_inset_input,
@@ -881,7 +905,6 @@ UI_ElementDef *r;
 
 	Arg args[1];
 	Boolean state;
-	char *s;
 	int i;
 
 	for (i = 0, r = insert_state; r && r->op != UI_EndOp; ++r)
@@ -916,38 +939,23 @@ static void SetInsertState(w, client_data, call_data)
 Widget w;
 XtPointer client_data, call_data;
     {
-	Arg args[1];
 	long start, end;
 	UI_PanelDef *panel = (UI_PanelDef *)client_data;
 	char *prefix = GetCurrentInsertState(panel->elements);
 
 	XeTextInsertPrefix(panel->target, prefix, prefix ? strlen(prefix) : 0);
 	/*
-	** If a selection is currently active, then extract the selected
-	** text away from the widget, and then replace the text with the
-	** extracted text. This is rather inefficient way of getting the
-	** attributes of the selected text to change as defined by the
-	** input state.
-	**
-	** NOTE:As the selected content is effected by the currently
-	**	chosen export format, some information may be lost,
-	**	if the format is too simple (like string).
+	** If a selection is currently active, then change the current
+	** text mode in the selection.
 	*/
 	XeTextGetSelectionPosition(panel->target, &start, &end);
 	if (end > start)
 	    {
-		String value;
-		DataBuffer buffer;
-
-		buffer = init_DataBuffer;
-		XeTextExtract(panel->target, start, end, FeedData,
-			      (XtPointer)&buffer);
 		XeTextDisableDisplay(panel->target);
-		XeTextReplace(panel->target, start, end, (char *)NULL, 0);
 		XeTextUnsetSelection(panel->target);
-		PasteData(panel->target, buffer.data, buffer.length);
+		XeTextSetInsertionPosition(panel->target, start);
+		XeTextChangeMode(panel->target, (char *)NULL, 0L, end - start);
 		XeTextEnableDisplay(panel->target);
-		XtFree(buffer.data);
 	    }
     }
 
@@ -975,7 +983,7 @@ static UI_ActionDef query_reset_actions[] =
     };
 
 
-static UI_PanelDef query_reset_modified =
+UI_PanelDef query_reset_modified =
     {
 	"Text has been modified--clear flag?",
 	NULL,
@@ -1008,6 +1016,7 @@ static UI_MenuItemDef file_menu_items[] =
 	{"Save as", UI_PopupPanel, (XtPointer)&popup_dialog_saveas,},
 	{"Load File", CheckForModify, (XtPointer)&popup_dialog_loadfile,},
 	{"Add Inset", UI_PopupPanel, (XtPointer)&popup_dialog_insetfile,},
+	{"Print", PrintToDisplay, (XtPointer)&edit_widget,},
 	{"Quit", CheckModifyForQuit, NULL},
 	{NULL},
     };
@@ -1066,13 +1075,25 @@ String message;
 	UI_PopupPanel(main_window, (XtPointer)&warning, (XtPointer)NULL);
     }
 
+/*
+** ErrorHandler
+**	This exists only for as a place to put a breakpoint in case of
+**	need to track down Xt error...
+*/
+static void ErrorHandler(message)
+String message;
+    {
+	printf("\n%s\n", message);
+	exit(1);
+    }
+
 #include "fallback.c"
 
-main(argc, argv)
+int main(argc, argv)
 int argc; char *argv[];
     {
 	XtAppContext demo_application;
-	Widget view, widget, pane, box, button;
+	Widget view, pane, box;
 	Arg args[10];
 	int n;
 	XeContentFormat content_type;
@@ -1103,6 +1124,7 @@ int argc; char *argv[];
 		("fileName", labelWidgetClass, box, args, n);
 	SetFilenameToTitle();
 	XtAppSetWarningHandler(demo_application, WarningHandler);
+	XtAppSetErrorHandler(demo_application, ErrorHandler);
 	n = 0;
 	XtSetArg(args[n], XtNallowHoriz, True); ++n;
 	XtSetArg(args[n], XtNallowVert, True); ++n;
