@@ -1,6 +1,8 @@
 /*
 ** ui_tools.c
 **
+** Copyright 1993-1995 by Markku Savela and
+**	Technical Research Centre of Finland
 */
 #include <stdio.h>
 #if SYSV_INCLUDES
@@ -17,10 +19,20 @@
 #include <X11/Xaw/SimpleMenu.h>
 #include <X11/Xaw/Toggle.h>
 #include <X11/Xew/Frame.h>
+#include <X11/Xew/Text.h>
 #include <X11/Xew/TextEd.h>
-
+#include <X11/Xew/Raster.h>
+#include <X11/Xew/Audio.h>
+#include <X11/Xew/Video.h>
+#include <X11/Xew/Support.h>
+#if USING_MOTIF_122
+#include <Xm/Xm.h>
+#endif
+#include "filename.h"
 #include "ui_tools.h"
 
+char *UI_library_search_path;
+
 /*
 ** ***************************************************************
 ** For documentation of the global UI_ functions, see "ui_tools.h"
@@ -120,36 +132,205 @@ XtGrabKind mode;
 	XtPopup(w, mode);
     }
 
-void UI_WarningMessage(parent, panel,
-		     name, type, class, defaultp, params, num_params)
+void UI_WarningMessage(parent, format, params, num_params)
 Widget parent;
-UI_PanelDef *panel;
-String name, type, class, defaultp;
-String *params;
-Cardinal *num_params;
+char *format;
+char **params;
+int num_params;
     {
-	char buffer[1000];
-	static char message[1000];
-	
-	XtGetErrorDatabaseText(name,type,class,defaultp,buffer,sizeof(buffer));
-	if (params && num_params && *num_params > 0)
+	static char buffer[1000];
+
+	UI_PanelDef panel = {"warningPanel",};
+
+	char *r, *mark, *q;
+	int n = sizeof(buffer)-1, p = 0, code;
+	int length;
+
+	for (r = mark = format, q = buffer; (r = strchr(r, '%')) != NULL; )
 	    {
-		int i = *num_params, j;
-		String par[10];
-		if (i > 10)
-			i = 10;
-		for (j = 0; j < i; j++)
-			par[j] = params[j];
-		while (i < 10)
-			par[i++] = NULL;
-		sprintf(message, buffer,
-			par[0], par[1], par[2], par[3], par[4],
-			par[5], par[6], par[7], par[8], par[9]);
+		length = r - mark;
+		if (length > 0 && n > length)
+		    {
+			memcpy(q, mark, length);
+			q += length;
+			n -= length;
+		    }
+		mark = (code = *++r) ? ++r : r;
+		if (code == 's')
+		    {
+			/*
+			** The only formatting code recognised is just '%s'
+			*/
+			if (p < num_params && params[p])
+			    {
+				length = strlen(params[p]);
+				if (length < n)
+				    {
+					memcpy(q, params[p], length);
+					q += length;
+					n -= length;
+				    }
+				p += 1;
+			    }
+		    }
+		else
+			--mark;
 	    }
+	length = strlen(mark);
+	if (length > 0 && length < n)
+	    {
+		memcpy(q, mark, length);
+		q += length;
+		n -= length;
+	    }
+	*q = 0;
+	/*
+	** Force a totally new widget for each warning popup: destroy
+	** old popup widget if it exists...
+	*/
+	if (panel.shell)
+	    {
+		XtDestroyWidget(panel.shell);
+		panel.shell = NULL;
+	    }
+	panel.heading = buffer;
+	UI_PopupPanel(parent, (XtPointer)&panel, (XtPointer)NULL);
+    }
+
+/*
+** CreateWidgetFromFile
+**	(if there are any problems or the file content format is not
+**	recognized, just create a simple XeText containing the filename.
+*/
+static Widget CreateWidgetFromFile(parent, name, start, end)
+Widget parent;
+char *name, *start, *end;
+    {
+	int n;
+	Arg args[2];
+
+	Widget w;
+	WidgetClass class = NULL;
+	XeContentFormat fmt = XeContentFormat_UNKNOWN;
+	char *path;
+
+	if (end == NULL)
+		end = start + strlen(start);
+	n = end - start;
+	/*
+	** Set up the filename as a default content
+	*/
+	XtSetArg(args[0], XtNcontentString, start);
+	XtSetArg(args[1], XtNcontentLength, n);
+
+	path = FileNameFromPath(start, end, UI_library_search_path);
+	if (path != NULL)
+	    {
+		fmt = XeFindContentFormat(path);
+		switch (XeFindContentClass(fmt))
+		    {
+		    case XeContentClass_RASTER:
+			class = xeRasterWidgetClass;
+			break;
+		    case XeContentClass_AUDIO:
+			class = xeAudioWidgetClass;
+			break;
+		    case XeContentClass_VIDEO:
+			class = xeVideoWidgetClass;
+			break;
+		    default:
+			break;
+		    }
+	    }
+	if (class == NULL)
+		class = xeTextWidgetClass;
 	else
-		strncpy((char *)message, buffer, sizeof(message));
-	panel->heading = message;
-	UI_PopupPanel(parent, (XtPointer)panel, (XtPointer)NULL);
+	    {
+		XtSetArg(args[0], XtNcontentFile, path);
+		XtSetArg(args[1], XtNcontentFormat, fmt);
+	    }
+	w = XtCreateManagedWidget(name, class, parent, args, 2);
+	/*
+	** Remove resource refs to the strings (that will not be available)
+	*/
+	XtSetArg(args[0], XtNcontentString, 0);
+	XtSetArg(args[1], XtNcontentFile, 0);
+	XtSetValues(w, args, 2);
+	return w;
+    }
+
+/*
+** CreateLabelElement
+**	in basic form create a simplest XeText widget the content of which
+**	will be set to be the "label".
+**
+**	A special feature ("hack"):
+**		If the 'label' contains a special marker INSET_MARK,
+**		then the string following it up to the next INSET_MARK
+**		or to the end of the label is taken as a filename, the
+**		content of which is assumed to be the intended label.
+**
+**		If the 'label' contains INSERT_MARK and other text outside
+**		them, then XeTextEd widget is created to contain the outside
+**		text and insets.
+*/
+
+#define INSET_MARK '\1'
+
+static Widget CreateLabelElement(parent, name, label)
+Widget parent;
+char *name, *label;
+    {
+	char *start, *end;
+	Widget w = None;
+
+	if (label == NULL)
+		return w;
+	if ((start = strchr(label, INSET_MARK)) == NULL)
+	    {
+		/*
+		** The simple case. Just create a XeText with the
+		** 'label' as a content.
+		*/
+		w = XtVaCreateManagedWidget
+			(name, xeTextEdWidgetClass, parent,
+			 XtNallowEdit, False,
+			 XtNcontentString, label,
+			 XtNcontentLength, strlen(label),
+			 (XtPointer)NULL);
+		XtVaSetValues(w, XtNcontentString, (XtPointer)NULL,
+			      (XtPointer)NULL);
+		return w;
+	    }
+	end = strchr(start + 1, INSET_MARK);
+	if (start == label && (end == NULL || !end[1]))
+		/*
+		** The next simplest case: the label is just a reference
+		** to some file to be used.
+		*/
+		return CreateWidgetFromFile(parent, name, start+1, end);
+	/*
+	** Most complex case: need a XeTextEd with text and/or other widgets
+	*/
+	w = XtVaCreateManagedWidget
+		(name, xeTextEdWidgetClass, parent,
+		 XtNallowEdit, False,
+		 (XtPointer)NULL);
+	for (;;)
+	    {
+		XeTextInsert(w, label, start - label);
+		CreateWidgetFromFile(w, "labelInset", start+1, end);
+		if (end == NULL)
+			break;
+		label = ++end;
+		if ((start = strchr(label, INSET_MARK)) == NULL)
+		    {
+			XeTextInsert(w, label, strlen(label));
+			break;
+		    }
+		end = strchr(start+1, INSET_MARK);
+	    }
+	return w;
     }
 
 static void RelayAction(w, client_data, call_data)
@@ -170,52 +351,113 @@ Widget w;
 XtPointer client_data, call_data;
     {
        	Arg tmp[1];
-       	Boolean state;
 
 	XtSetArg(tmp[0], XtNinvertShadow, (call_data != NULL));
 	XtSetValues(w, tmp, 1);
     }
 	
-static Widget CreateToggleElement(parent, name, args, n)
+static Widget CreateToggleElement(parent, name, label, callback, client_data)
 Widget parent;
-char *name;
-Arg *args;
-int n;
+char *name, *label;
+XtCallbackProc callback;
+XtPointer client_data;
     {
+	static String translations = 
+"\
+<Btn1Down>,<Btn1Up>:notify(toggle) notify(notify)\n\
+<EnterWindow>:highlight() notify(highlight)\n\
+<LeaveWindow>:unhighlight() notify(unhighlight)\
+";
+	static XtTranslations box_translations;
+	
 	Widget box, toggle;
+	Arg args[2];
+	int n;
 
+	if (!box_translations)
+		box_translations = XtParseTranslationTable(translations);
+	n = 0;
 	XtSetArg(args[n], XtNallowEdit, False); ++n;
+	XtSetArg(args[n], XtNtranslations, box_translations); ++n;
 	box = XtCreateManagedWidget
-		("toggleBox", xeTextEdWidgetClass, parent, args, n);
+		(name, xeTextEdWidgetClass, parent, args, n);
 	XeTextDisableDisplay(box);      
 	n = 0;
 	XtSetArg(args[n], XtNlabel, " "); ++n;
 	toggle = XtCreateManagedWidget
 		("toggleButton", toggleWidgetClass, box, args, n);
-       	XtAddCallback(toggle, XtNcallback, ToggleNotify, (XtPointer)NULL);
-	XeTextInsert(box, name, strlen(name));       
+	if (callback == NULL)
+		callback = ToggleNotify;
+       	XtAddCallback(toggle, XtNcallback, callback, (XtPointer)client_data);
+	if (label)
+		XeTextInsert(box, label, strlen(label));
 	XtInstallAccelerators(box, toggle);
 	XtAddCallback(box,XtNnotifyCallback,RelayAction,(XtPointer)toggle);
 	XeTextEnableDisplay(box);      
 	return toggle;
     }
 
+static Widget CreateButtonElement(parent, name, label, callback, client_data)
+Widget parent;
+char *name, *label;
+XtCallbackProc callback;
+XtPointer client_data;
+    {
+	static String translations =
+"\
+<Btn1Down>,<Btn1Up>:activate()\n\
+<EnterWindow>:highlight()\n\
+<LeaveWindow>:unhighlight()\n\
+";
+	static XtTranslations box_translations;
+
+	Widget button;
+	Arg arg;
+
+	if (!box_translations)
+		box_translations = XtParseTranslationTable(translations);
+
+	button = CreateLabelElement(parent, name, label);
+	if (button && callback != NULL)
+		XtAddCallback(button,XtNactivateCallback,callback,client_data);
+	if (button)
+	    {
+		XtSetArg(arg, XtNtranslations, box_translations);
+		XtSetValues(button, &arg, 1);
+	    }
+	return button;
+    }
+
 /*
 ** CreateInputElement
 */
-static Widget CreateInputElement(parent, prompt, init, args, n)
+static Widget CreateInputElement
+	(parent, name, prompt, init, callback, client_data)
 Widget parent;
-char *prompt, *init;
-Arg *args;
-int n;
+char *name, *prompt, *init;
+XtCallbackProc callback;
+XtPointer client_data;
     {
+	static String translations =
+"\
+<EnterWindow>: highlight()\n\
+<LeaveWindow>: unhighlight()\
+";
+	static XtTranslations box_translations;
+
+	Arg args[4];
+	int n;
 	Widget box, input;
 
+	if (!box_translations)
+		box_translations = XtParseTranslationTable(translations);
+	n = 0;
 	XtSetArg(args[n], XtNallowEdit, False); ++n;
-	box = XtCreateManagedWidget
-		("inputBox", xeTextEdWidgetClass, parent, args, n);
-	XeTextDisableDisplay(box);      
-	XeTextInsert(box, prompt, strlen(prompt));      
+	XtSetArg(args[n], XtNtranslations, box_translations); ++n;
+	box = XtCreateManagedWidget(name,xeTextEdWidgetClass, parent, args,n);
+	XeTextDisableDisplay(box);
+	if (prompt)
+		XeTextInsert(box, prompt, strlen(prompt));      
 	n = 0;
 	XtSetArg(args[n], XtNresize, True); ++n;
 	XtSetArg(args[n], XtNexportFormat, XeTextExport_STRING); ++n;
@@ -232,6 +474,9 @@ int n;
 		("inputArea", xeTextEdWidgetClass, box, args, n);
 #endif
 	XtSetKeyboardFocus(box, input);
+	if (callback != NULL)
+		XtAddCallback(input,XtNactivateCallback,callback,client_data);
+	XtInstallAccelerators(box, input);
 	XeTextEnableDisplay(box);
 	return input;
     }
@@ -323,7 +568,7 @@ BigVal **big;
 	if (!*rlist)
 		GetResourceList(w, rlist, num);
 	if (*rlist == NULL || *num == 0)
-		return;
+		return False;
 	/*
 	** Locate the resource to find out the representation size
 	*/
@@ -367,12 +612,10 @@ XtPointer client_data, call_data;
 	int need_convert;
 	String s;
 	Arg args[20];
-	int n, i;
+	int n;
 	XtResourceList resources = NULL;
 	Cardinal num_resources = 0;
-	BigVal *big_values = NULL, *big;
-
-	static long null_value = 0;
+	BigVal *big_values = NULL;
 
 	if (!panel || !panel->target)
 		return;	/* Nothing to load */
@@ -432,7 +675,7 @@ XtPointer client_data, call_data;
 		    default:
 			break;
 		    }
-	if (n && panel->target)
+	if (n)
 		XtSetValues(panel->target, args, n);
 	while (big_values)
 	    {
@@ -480,7 +723,6 @@ Cardinal *num;
     {
 	XtPointer value;
 	Widget t = p->widget;
-	char buf[20];
 	XtResourceList rl, last;
 	XtArgVal argval;
 	XrmValue ref, to;
@@ -573,58 +815,70 @@ UI_PanelDef *panel;
 	XtFree((char *)resources);
     }
 
-static Widget CreateElementPanel(parent, name, definition)
+static Widget CreateElementPanel(parent, name, definition, client_data)
 Widget parent;
 char *name;
 UI_ElementDef *definition;
+XtPointer client_data;
     {
-	Arg args[20];
-	int n;
+	Arg args[4];
+	int layout_path;
 	Widget panel, pane, radio = 0;
-	char *s;
 	UI_ElementDef *r;
 
-	n = 0;
-	XtSetArg(args[n], XtNlayoutPath, 0); ++n;
 	pane = panel = XtCreateManagedWidget
-		(name, xeFrameWidgetClass, parent, args, n);
-	for (r = definition; r && r->name; r++)
+		(name, xeFrameWidgetClass, parent, args, 0);
+	XtVaGetValues(pane, XtNlayoutPath, &layout_path, (XtPointer)NULL);
+	if (layout_path == 0)
+		layout_path = 270;
+	else
+		layout_path = 0;
+	for (r = definition; r && r->op != UI_EndOp; r++)
+	    {
+		if (r->name == NULL)
+			continue;
 		switch (r->op)
 		    {
 		    case UI_PaneOp:
-			n = 0;
-			XtSetArg(args[n], XtNlayoutPath, 270); ++n;
+			XtSetArg(args[0], XtNlayoutPath, layout_path);
 			pane = XtCreateManagedWidget
-				("pane", xeFrameWidgetClass, panel, args, n);
+				(r->name, xeFrameWidgetClass, panel, args, 1);
+			if (r->label == NULL)
+				break;
 			/* FALL THROUGH TO LABEL Handling */
 		    case UI_LabelOp:
-			n = 0;
-			XtSetArg(args[n],XtNcontentString,r->name); ++n;
-			XtSetArg(args[n],XtNcontentLength,strlen(r->name));++n;
-			r->widget = XtCreateManagedWidget
-				("labelBox", xeTextWidgetClass, pane, args, n);
+			r->widget = CreateLabelElement(pane,r->name,r->label);
+			break;
+		    case UI_ButtonOp:
+			r->widget = CreateButtonElement
+				(pane, r->name, r->label,
+				 r->callback, client_data);
 			break;
 		    case UI_ToggleOp:
 		    case UI_RadioOp:
-			n = 0;
-			r->widget = CreateToggleElement(pane,r->name,args,n);
-			n = 0;
-			XtSetArg(args[n], XtNradioData, r); ++n;
-			XtSetValues(r->widget, args, n);
+			r->widget = CreateToggleElement
+				(pane, r->name, r->label,
+				 r->callback, client_data);
+			XtSetArg(args[0], XtNradioData, r);
+			XtSetValues(r->widget, args, 1);
 			XawToggleChangeRadioGroup
 				(r->widget,
 				 r->op == UI_RadioOp ? radio : r->widget);
 			radio = r->widget;
 			break;
 		    case UI_InputOp:
-			n = 0;
 			r->widget = CreateInputElement
-				(pane, r->name, NULL, args, n);
+				(pane, r->name, r->label, NULL,
+				 r->callback, client_data);
+			break;
+		    case UI_BlankOp:
+			if (r->callback)
+				(*r->callback)(pane, client_data, (XtPointer)r);
 			break;
-			
 		    default:
 			break;
 		    }
+	    }
 	return panel;
     }
 
@@ -635,18 +889,28 @@ UI_ActionDef *actions;
     {
 	int n;
 	Arg args[3];
-	Widget box, but;
-	int have_popdown = 0;
+	Widget box;
 	XtCallbackProc callback;
 
 	n = 0;
 	XtSetArg(args[n], XtNlayoutPath, 0); ++n;
 	box = XtCreateManagedWidget(name, xeFrameWidgetClass,parent,args,n);
-	for (;actions && actions->name; ++actions)
+	if (actions == NULL)
+	    {
+		/* If no actions are present, make default OK button */
+		Widget tmp;
+
+		n = 0;
+		XtSetArg(args[n], XtNlabel, "OK"); ++n;
+		tmp = XtCreateManagedWidget
+			("actionBox", commandWidgetClass, box, args, n);
+		XtAddCallback(tmp, XtNcallback, UI_Popdown, (XtPointer)NULL);
+	    }
+	else for (;actions->name; ++actions)
 	    {
 		n = 0;
 		XtSetArg(args[n], XtNlabel, actions->name); ++n;
-		but = XtCreateManagedWidget
+		actions->widget = XtCreateManagedWidget
 			("actionBox", commandWidgetClass, box, args, n);
 		callback = actions->callback;
 		/*
@@ -655,23 +919,11 @@ UI_ActionDef *actions;
 		** a way to override the default popdown label "ok".
 		*/
 		if (callback == NULL)
-		    {
 			callback = UI_Popdown;
-			have_popdown = 1;
-		    }
-		XtAddCallback(but,XtNcallback,callback,actions->client_data);
-	    }
-	/*
-	** Implicitly add "popdown" button, if not added by the user
-	*/
-	if (!have_popdown)
-	    {
-		n = 0;
-		XtSetArg(args[n], XtNlabel, "ok"); ++n;
-		but = XtCreateManagedWidget
-			("actionBox", commandWidgetClass, box, args, n);
-		XtAddCallback(but, XtNcallback, UI_Popdown, (XtPointer)parent);
+		XtAddCallback(actions->widget, XtNcallback,
+			      callback,actions->client_data);
 	    }
+	return box;
     }
 
 void UI_PopupPanel(w, client_data, call_data)
@@ -679,7 +931,7 @@ Widget w;
 XtPointer client_data, call_data;
     {
 	UI_PanelDef *panel = (UI_PanelDef *)client_data;
-	Widget close, pane;
+	Widget pane;
 	Cardinal n;
 	Arg args[2];
 	char *label;
@@ -693,9 +945,8 @@ XtPointer client_data, call_data;
 	if (!panel->shell)
 	    {
 		/*
-		** Some widgets just cannot work as popup parents for
-		** some reason.. just lookup for shell and hope it works
-		** ... (maybe someday I get to know why... --msa)
+		** Not all widgets can work as popup parents, just lookup
+		** the closest shell for a parent (to be safe (?)).
 		*/
 		while (XtParent(w) && !XtIsShell(w))
 			w = XtParent(w);
@@ -705,23 +956,48 @@ XtPointer client_data, call_data;
 		XtSetArg(args[n], XtNlayoutPath, 270); ++n;
 		pane = XtCreateManagedWidget
 			("panel", xeFrameWidgetClass, panel->shell, args, n);
-		n = 0;
-		XtSetArg(args[n], XtNlabel, label); ++n;
-		panel->label = XtCreateManagedWidget
-			("top",labelWidgetClass, pane, args, n);
-		CreateElementPanel(pane, "middle", panel->elements);
+		/*
+		** *WARNING*
+		**	Because XeText does not make a local copy of the
+		**	contentString resource, the 'label' given here
+		**	should exist in static store (if not, it will
+		**	be a disaster if someone resets XtNcontentLoaded
+		**	without resetting XtNcontentString at same time!
+		*/
+		panel->label = XtVaCreateManagedWidget
+			("top", xeTextWidgetClass, pane,
+			 XtNcontentString, label,
+			 XtNcontentLength, strlen(label),
+			 (XtPointer)NULL);
+		if (panel->elements)
+			CreateElementPanel(pane, "middle", panel->elements,
+					   (XtPointer)panel);
 		CreateBottomPanel(pane, "bottom", panel->actions);
 		UI_SetToggles(panel);
 	    }
 	else if (label && panel->label)
-	    {
-		n = 0;
-		XtSetArg(args[n], XtNlabel, label); ++n;
-		XtSetValues(panel->label, args, n);
-	    }
+		XtVaSetValues(panel->label,
+			      XtNcontentLoaded, False,
+			      XtNcontentString, label,
+			      XtNcontentLength, strlen(label),
+			      (XtPointer)NULL);
 	UI_PopupCentered(NULL, panel->shell, XtGrabNone);
     }
 
+/*
+** UI_CreateElementPanel
+**	create a simple XeFrame and elements inside it.
+*/
+Widget UI_CreateElementPanel(parent, name, elements, client_data)
+Widget parent;
+char *name;
+UI_ElementDef *elements;
+XtPointer client_data;
+    {
+	return CreateElementPanel
+		(parent, name ? name : "panel", elements, client_data);
+    }
+
 void UI_CreateMenu(parent, menu)
 Widget parent;
 UI_MenuDef *menu;
@@ -805,7 +1081,6 @@ XtPointer *converter_data;
 	return CvtStringDone(cvt_buf, to);
     }
 
-
 Boolean CvtPixel_String(display, args, num_args, from, to, converter_data)
 Display *display;
 XrmValue *args;
@@ -987,7 +1262,6 @@ XrmValue *from, *to;
 XtPointer *converter_data;
     {
 	XeFrameBorder *border = (XeFrameBorder *)from->addr;
-	XColor query;
 
 	if (border == NULL)
 		return CvtStringDone("", to);
@@ -1052,6 +1326,14 @@ XtAppContext app;
 	XtAppSetTypeConverter
 		(app, XeRFrameBorder, XtRString, CvtFrameBorder_String,
 		 (XtConvertArgList)NULL, 0, XtCacheNone, (XtDestructor)NULL);
+#if USING_MOTIF_122
+	XtAppSetTypeConverter
+		(app, XmRVerticalDimension, XtRString, CvtInt_String,
+		 (XtConvertArgList)NULL, 0, XtCacheNone, (XtDestructor)NULL);
+	XtAppSetTypeConverter
+		(app, XmRHorizontalDimension, XtRString, CvtInt_String,
+		 (XtConvertArgList)NULL, 0, XtCacheNone, (XtDestructor)NULL);
+#endif
     }
 
 
